{

                          CHAPTER 1




                      P-machine Overview





 This program provides an operational definition of  the  P-machine
 for executing object programs generated by the Model Compiler. The
 machine is characterised by its pure-stack  architecture  and  its
 large  repertoire of instructions tailored specifically towards an
 efficient and secure implementation of Standard Pascal.   As  with
 most high-level machine descriptions, no free-standing implementa-
 tion of arithmetic functions has been attempted, and this  version
 will  inherit  the real and integer arithmetic of its host.  More-
 over, no general  implementation  of  program  parameters  can  be
 expressed  in  Standard  Pascal, and accordingly all communication
 with  the  external  environment  is  channelled  through  the  P-
 machine's own program heading which is as follows:


                                                                     }

                          program

      Pmachine ( CodeFile, CorpseFile, Input, Output);

{


 These input and output files have the following  nature  and  pur-
 pose:

 CodeFile     is an input file containing the object program to  be
              executed.

 Input        is an input file providing an input  medium  for  the
              executing program.

 Output       is an output file providing an output medium for  the
              executing program.

 CorpseFile   is an output file that contains  the  'corpse'  of  a
              program  that terminated in error. The corpse file is
              passed to the Postmortem Generator program for subse-
              quent error analysis.

 The P-machine is modular in design, and comprises a  Loader,  Pro-
 cessor,  Memory  Manager,  and Input/Output Manager.  In addition,
 there are modules to provide support for file  creation,  instruc-
 tion  execution,  and overall P-machine control. As with the Model
 Compiler, the modular structure is represented by  organising  the
 program  as  a  series of code 'chapters'. However Standard Pascal
 constrains all declarations to be collected together within a sin-
 gle  chapter which must then be read in conjuction with those that
 follow.

 Chapter 3 provides overall P-machine control by reading the set of
 compile-time  options  from  the  header  on  the P-code file. The
 degree of runtime checking supported  follows  directly  from  the
 option  values  supplied.  This chapter also provides for the gen-
 eration of the CorpseFile  and  contains  hooks  for  the  system-
 dependent call to the Postmortem Generator.

 Chapter 4 provides 'micro-level' support  for  instruction  execu-
 tion.

 Chapter 5 contains the Memory Manager, and provides stack and heap
 management  for  all  program data and for I/O buffers. The module
 includes  a  simple  garbage  collector  allowing  effective   re-
 utilisation of discarded heap storage.

 Chapter 6 contains the System File Manager, and provides  a  simu-
 lated  filestore  which is sufficient to support the creation of a
 limited number of 'local files'. The module also arranges for sys-
 tem  communication  via the P-machine program parameters Input and
 Output.

 Chapter 7 contains the Pascal I/O Manager,  and  provides  support
 for  all  standard  I/O  functions.  The  I/O manager controls all
 transfers between the System File Manager  and  the  corresponding
 file variables by means of I/O buffers acquired from the heap.

 Chapter 8 contains a number of utilities supporting  integer,  set
 and string arithmetic.

 Chapter 9 contains the machine Loader that reads the  P-code  file
 generated by the compiler.

 Chapter 10 contains the Processor that controls program execution.
 A  program is run until normal termination, or an error occurs. In
 the latter case the Controller dumps the entire  P-machine  memory
 onto the CorpseFile, and calls the Postmortem Generator for subse-
 quent analysis.


 Copyright  (C)     University of Manchester
                    Institute of Science and Technology
                    January 1984.

                                                                     }
{

                          CHAPTER 2




                         Declarations





 As with the Model Compiler, this chapter contains all of the  glo-
 bal definitions and declarations on which the procedures in subse-
 quent chapters depend. In this chapter, only a brief indication of
 the  overall  nature  of global constants, types, and variables is
 given, together with a reference to the chapter in which they  are
 used.

                                                                     }

label

   13 ;                                 { catastrophic machine failure }


const


   {*************************************************************************}
   {*************************************************************************}
   {******                                                             ******}
   {******                                                             ******}
   {******                     Host Installation                       ******}
   {******                     -----------------                       ******}
   {******                                                             ******}
   {******                                                             ******}
   {****** 1. Investigate how your host Pascal processor implements    ******}
   {******    the variable:-                                           ******}
   {******                                                             ******}
   {******      HostWord : set of B0..Bn                               ******}
   {******                                                             ******}
   {******    If member B0 is mapped onto the most significant bit     ******}
   {******    position, then set the constant HostBitOrder to the      ******}
   {******    value LeftToRight.  Otherwise, set it to RightToLeft.    ******}
   {******                                                             ******}
   {****** 2. Investigate how your host Pascal processor implements    ******}
   {******    the packed variable:-                                    ******}
   {******                                                             ******}
   {******      HostWord : packed array [0..n] of char                 ******}
   {******                                                             ******}
   {******    If element HostWord[0] is mapped onto the most           ******}
   {******    significant byte position, then set the constant         ******}
   {******    HostByteOrder to the value LeftToRight. Otherwise, set   ******}
   {******    it to RightToLeft.                                       ******}
   {******                                                             ******}
   {****** 3. Set the value of FormFeed so that chr(FormFeed) will     ******}
   {******    throw a new printer page, or have an equivalent effect   ******}
   {******    on the chosen output device.                             ******}
   {******                                                             ******}
   {****** 4. Redefine the constants prefixed 'MC' for the host        ******}
   {******    machine.                                                 ******}
   {******                                                             ******}
   {*************************************************************************}
   {*************************************************************************}



   LeftToRight = 0 ;
   RightToLeft = 1 ;

   HostBitOrder  = RightToLeft ;         { for host Pascal compiler }
   HostByteOrder = LeftToRight ;
   FormFeed      = 12 ;                  { ASCII form-feed character }



   { constants describing the model P-machine }

   MCBitsPerWord  = 32 ;                { bits per memory-word }
   MCWordSetBits  = 31 ;                { bits per set-word }
   MCBitsPerByte  = 8  ;                { bits per memory-byte }
   MCBytesPerWord = 4  ;                { mcbitsperword div mcbitsperbyte }
   MCMaxBitnum    = 31 ;                { mcbitsperword-1 }
   MCMaxSetBit    = 30 ;                { mcwordsetbits-1 }
   MCMaxByteNum   = 3 ;                 { mcbytesperword-1 }

   MCRealSize     = 2 ;                 { real size in words }
   MCMaxByte      = 255 ;               { max mc-byte value }
   MCMaxInt       = 2147483647 ;        { max mc-integer value }
   MCMaxRealRoot  = 46340.95 ;          { max real such that              }
                                        { sqr(mcmaxrealroot)<=mcmaxint.0  }
   MCMaxMemWord   = 16383 ;        { max memory-word address }
   MCMaxMemByte   = 65532 ;        { max memory-byte address }


   AlfaSize       = 12 ;
   MaxVariantLevel= 9 ;                 { max variant nesting level }


   { P-machine class 1 instructions }


   cons0  =   0 ; cons1  =   1 ; cons2  =   2 ; cons3  =   3 ; cons4  =   4 ;
   cons5  =   5 ; cons6  =   6 ; cons7  =   7 ; cons8  =   8 ; cons9  =   9 ;
   cons10 =  10 ; cons11 =  11 ; cons12 =  12 ; cons13 =  13 ; cons14 =  14 ;
   cons15 =  15 ; cons16 =  16 ; cons17 =  17 ; cons18 =  18 ; cons19 =  19 ;
   cons20 =  20 ; cons21 =  21 ; cons22 =  22 ; cons23 =  23 ; cons24 =  24 ;
   cons25 =  25 ; cons26 =  26 ; cons27 =  27 ; cons28 =  28 ; cons29 =  29 ;
   cons30 =  30 ; cons31 =  31 ; consb  =  32 ; consw  =  33 ; conmb  =  34 ;
   conmw  =  35 ; crefb  =  36 ; crefw  =  37 ; csetb  =  38 ; csetw  =  39 ;
   lodl0  =  40 ; lodl1  =  41 ; lodl2  =  42 ; lodl3  =  43 ; lodl4  =  44 ;
   lodl5  =  45 ; lodl6  =  46 ; lodl7  =  47 ; lodlb  =  48 ; lodlw  =  49 ;
   lrf0   =  50 ; lrf1   =  51 ; lrf2   =  52 ; lrf3   =  53 ; lrf4   =  54 ;
   lrf5   =  55 ; lrf6   =  56 ; lrf7   =  57 ; lrflb  =  58 ; lrflw  =  59 ;
   stol0  =  60 ; stol1  =  61 ; stol2  =  62 ; stol3  =  63 ; stol4  =  64 ;
   stol5  =  65 ; stol6  =  66 ; stol7  =  67 ; stolb  =  68 ; stolw  =  69 ;
   lodgb  =  70 ; lodgw  =  71 ; lrfgb  =  72 ; lrfgw  =  73 ; stogb  =  74 ;
   stogw  =  75 ; lodeb  =  76 ; lodew  =  77 ; lrfeb  =  78 ; lrfew  =  79 ;
   stoeb  =  80 ; stoew  =  81 ; lodib  =  82 ; lodiw  =  83 ; lrfib  =  84 ;
   lrfiw  =  85 ; stoib  =  86 ; stoiw  =  87 ; lodin  =  88 ; indx1  =  89 ;
   indx2  =  90 ; indx3  =  91 ; indx4  =  92 ; indx5  =  93 ; indx6  =  94 ;
   indx7  =  95 ; indxb  =  96 ; indxw  =  97 ; ixrfb  =  98 ; ixrfw  =  99 ;
   ixcap  = 100 ; stoin  = 101 ; adjpb  = 102 ; adjpw  = 103 ; adjp1  = 104 ;
   adjmb  = 105 ; adjmw  = 106 ; adjm1  = 107 ; adadr  = 108 ; lodbt  = 109 ;
   lodpk  = 110 ; stobt  = 111 ; stopk  = 112 ; ixpkb  = 113 ; ixpcap = 114 ;
   ixswb  = 115 ; adjpk  = 116 ; lodby  = 117 ; stoby  = 118 ; lbyrf  = 119 ;
   ibyrf  = 120 ; lodmb  = 121 ; lodmw  = 122 ; stomb  = 123 ; stomw  = 124 ;
   storb  = 125 ; storw  = 126 ; moveb  = 127 ; movew  = 128 ; szcap  = 129 ;
   mvcap  = 130 ; dpcap  = 131 ; rlcap  = 132 ; andop  = 133 ; notop  = 134 ;
   incrb  = 135 ; incrw  = 136 ; incr1  = 137 ; decrb  = 138 ; decrw  = 139 ;
   decr1  = 140 ; absint = 141 ; negint = 142 ; sqrint = 143 ; addint = 144 ;
   subint = 145 ; mulint = 146 ; divint = 147 ; modint = 148 ; oddint = 149 ;
   eqint  = 150 ; neint  = 151 ; ltint  = 152 ; leint  = 153 ; chktv  = 154 ;
   chktd  = 155 ; chklm  = 156 ; chklw  = 157 ; chkup  = 158 ; chkcap = 159 ;
   float  = 160 ; absrl  = 161 ; negrl  = 162 ; sqrl   = 163 ; addrl  = 164 ;
   subrl  = 165 ; mulrl  = 166 ; divrl  = 167 ; trurl  = 168 ; rndrl  = 169 ;
   eqrl   = 170 ; nerl   = 171 ; ltrl   = 172 ; lerl   = 173 ; lodptr = 174 ;
   eqptr  = 175 ; neptr  = 176 ; eqstr  = 177 ; nestr  = 178 ; ltstr  = 179 ;
   lestr  = 180 ; lodset = 181 ; stoset = 182 ; sglset = 183 ; rngset = 184 ;
   setuni = 185 ; setint = 186 ; setdif = 187 ; setexp = 188 ; eqset  = 189 ;
   neset  = 190 ; leset  = 191 ; inset  = 192 ; chksl  = 193 ; chksu  = 194 ;
   jmpfw  = 195 ; jmpbk  = 196 ; jmpvia = 197 ; jmpout = 198 ; fjpfw  = 199 ;
   fjpbk  = 200 ; fjpvia = 201 ; fjpcnd = 202 ; tjpcnd = 203 ; casejp = 204 ;
   mark   = 205 ; calglo = 206 ; calloc = 207 ; callev = 208 ; calotr = 209 ;
   calfml = 210 ; endp   = 211 ; endf   = 212 ; endmf  = 213 ; newp1b = 214 ;
   newp1w = 215 ; dspp1b = 216 ; dspp1w = 217 ; chkn1  = 218 ; chkd1  = 219 ;
   chkd2  = 220 ; noop   = 221 ; preset = 222 ; presrb = 223 ; presrw = 224 ;
   ldstk  = 225 ; dpstk  = 226 ; popstk = 227 ; chkwd  = 228 ; chkrp  = 229 ;
   chkvf  = 230 ; trapp  = 231 ; trapt  = 232 ; trapf  = 233 ; setlk  = 234 ;
   dislk  = 235 ; pglk   = 236 ; smpglk = 237 ; haltp  = 238 ; chkptr = 241 ;
   chkflw = 242 ; chkfup = 243 ; escape = 255 ;


   { P-machine class 2 instructions }


   chkrd  =   0 ; chkwr  =   1 ; rewrtf =   2 ; resetf =   3 ; eolnf  =   4 ;
   eoff   =   5 ; getf   =   6 ; rdint  =   7 ; rdrl   =   8 ; rdln   =   9 ;
   putf   =  10 ; wrchar =  11 ; wrbool =  12 ; wrstr  =  13 ; wrint  =  14 ;
   wrflt  =  15 ; wrfix  =  16 ; wrln   =  17 ; pagef  =  18 ; closef =  19 ;
   pckpr  =  20 ; pckprc =  21 ; Upkpr  =  22 ; upkprc =  23 ; sinf   =  24 ;
   cosf   =  25 ; expf   =  26 ; logf   =  27 ; sqrtf  =  28 ; arctnf =  29 ;
   mapp   =  30 ; chkbuf =  31 ;



   { P-machine stack-frame layout }

   SLOffset       = 1 ;                 { [framebase-1]=staticlink }
   CBOffset       = 2 ;                 { [framebase-2]=codebase   }
   DLOffset       = 3 ;                 { [framebase-3]=dynamiclink }
   RAOffset       = 4 ;                 { [framebase-4]=returnaddress }
   LDOffset       = 5 ;                 { [framebase-5]=lock-stack depth }
   LTOffset       = 6 ;                 { [framebase-6]=lock-stack top   }
   FrameSize      = 6 ;


   { P-machine jump-table header }

   ParamsOffset   = 1 ;                 { [codebase-1] = parameter size }
   LocalsOffset   = 2 ;                 { [codebase-2] = stack size }
   LocksOffset    = 3 ;                 { [codebase-3] = lock-stack size }
   SerialOffset   = 4 ;                 { [codebase-4] = 0/pmd block serial }
   FinalsOffset   = 5 ;                 { [codebase-5] = finalization ep }


   { constants used by the machine-controller, chapter 3 }

   RuntimeBase    = 300 ;               { runtime errors:      301..400 }
   SystemBase     = 400 ;               { system limit errors: 401..420 }
   FatalBase      = 420 ;               { fatal errors:        421..450 }



   { constants used by the memory-manager, chapter 5 }

   MinBlock       = 64 ;                { lower bound on free-space }
   MinHeapBlock   = 2 ;                 { lower bound on heap block size }
   FreeList       = MCMaxMemWord ;      { [freelist] = next freeblock }
   Null           = 0 ;                 { null block pointer }
   PointerSize    = 2 ;                 { heap pointer size }
   LockSize       = 3 ;                 { lock-stack entry size }
   HeapLockOffset = 2 ;                 { [heapaddress-2] = lock/level word }
   HeapLockBit    = MCMaxBitnum ;       { heap lock bit in lock/level word }
   HeapLevelByte  = 0 ;                 { new1/new2 level in lock/level word }



   { global constants used by the system file manager and }
   { the Pascal I/O manager, chapters 6 & 7               }

   MaxFileIndex   = 16 ;                { max file-index value }
   UndefinedIndex = 0 ;                 { denotes 'undefined' file variable }


   { constant used  by the system file manager, chapter 6 }

   BlankName      = '            ' ;


   { constants used by the Pascal I/O manager, chapter 7 }

   TextCode       = 1 ;                 { textfile code }
   WordCode       = 2 ;                 { wordfile code }
   PackedCode     = 3 ;                 { packedfile code }
   FileLockBit    = 0 ;                 { set when buffer is referenced }
   EofBit         = 1 ;                 { save eof status }
   EolBit         = 2 ;                 { save eoln status }
   MaxTextLine    = 255 ;               { max textfile line length  }





type

   {-------------------------------------------------------------------------}
   {     My types                                                            }
   {-------------------------------------------------------------------------}
   TAMbit=( b0, b1, b2, b3, b4, b5, b6, b7, b8, b9,
            b10,b11,b12,b13,b14,b15,b16,b17,b18,b19,
            b20,b21,b22,b23,b24,b25,b26,b27,b28,b29,
            b30,b31 );

   TAMword= set of TAMbit;

   { types describing the model P-machine }

   MCBit          = 0..1 ;
   MCByte         = 0..MCMaxByte ;
   MCWord         = integer ;
   MCScalar       = 0..MCMaxInt ;
   MCInteger      = integer ;
   MCReal         = real ;
   MCBitRange     = 0..MCBitsPerWord ;
   MCBitIndex     = 0..MCMaxBitnum ;
   MCSetBits      = 0..MCMaxSetBit ;
   MCBitArray     = packed array[MCBitIndex] of MCBit ;
   MCByteIndex    = 0..MCMaxByteNum ;
   MCByteArray    = packed array[MCByteIndex] of MCByte ;
   MCCharArray    = packed array[MCByteIndex] of char ;
   MCWordSet      = set of MCSetBits ;

   TAMCastBit = record
                   case boolean of
                      false : (Index: MCBitIndex);
                      true  : (Bitt : TAMBit);
                end;

   WordCast       = ( AsSet,AsBits,AsBitSet,AsBytes,AsChars,AsReal,AsValue ) ;
   MCWordForm     = record
                      case WordCast of
                        AsSet   : ( WSet : MCWordSet ) ;
                        AsBits  : ( WBits : MCBitArray ) ;
                        AsBitSet: ( BitSet : TAMWord );
                        AsBytes : ( WBytes : MCByteArray ) ;
                        AsChars : ( WChars : MCCharArray ) ;
                        AsReal  : ( WReal : MCReal ) ;
                        AsValue : ( WValue : MCWord )
                     end ;

    RealMap       = record
                      case Boolean of
                        false : ( RealValue : MCReal ) ;
                        true  : ( Word : array[1..MCRealSize] of MCWord )
                    end ;

   WordRange      = 0..MCMaxMemWord ;
   ByteRange      = 0..MCMaxMemByte ;
   BitRange       = MCBitRange ;

   Scalar         = 0..maxint ;

   AlfaIndex      = 1..AlfaSize ;
   AlfaLength     = 0..AlfaSize ;
   Alfa           = packed array[AlfaIndex] of char ;



   { types required by the machine controller, chapter 3 }

   OptionType     = ( Listing,CodeDump,Checks,PMDump,
                      Level,Programs,Margin,Other ) ;

   Terminus       = ( Normal,RuntimeError,SystemLimit,Fatal ) ;


   { type used by the memory manager, chapter 5 }

   KeyValues      = -MCMaxInt..+MCMaxInt ;


   { global types used by the system file manager and }
   { the Pascal I/O manager, chapters 6 & 7           }

   IndexRange     = UndefinedIndex..MaxFileIndex ;
   FileStatus     = ( Detach,Retain ) ;
   IOmode         = ( ForReading,ForWriting ) ;
   IOTransfer     = record
                      ByteAmount : ByteRange ; BitAmount : BitRange
                    end ;


   { type used by the system file manager, chapter 6 }

   SysFile         = record
                       BlockList,CurrentBlock : WordRange ;
                       Padding : BitRange
                     end ;


   { types used by the Pascal I/O manager, chapter 7 }

   PascFileMode   = ( NotSet,Inspection,Generation ) ;

   PascFileDescriptor =
     record
       case Defined : Boolean of
         false :
           ( ) ;
         true :
           ( Mode : PascFileMode ;
             EndOfFile : Boolean ;
             BufferSize,FirstWord : WordRange ;
             case TextFile : Boolean of
               true : ( BufferWord : WordRange ;
                        BytePointer,LastByte : ByteRange ) ;
               false : ( PointerWord,LastWord : WordRange ;
                         case PackedFile : Boolean of
                           false : ( ComponentSize : WordRange ) ;
                           true  : ( WordPointer : WordRange ;
                                     ElsPerWord : 2..MCBitsPerWord)))
     end ;

   StringRange    = 1..MaxTextLine ;
   TextString     = record
                       Length : 0..MaxTextLine ;
                       String : packed array[StringRange] of char
                    end ;

   NumberForms    = ( IntForm,RealForm ) ;
   PrintedForms   = ( Floated,Fixed ) ;

   Number   = record
                case Form : NumberForms of
                  IntForm  : ( IntValue : MCInteger ) ;
                  RealForm : ( RealValue : MCReal ;
                               case Format : PrintedForms of
                                 Fixed   : ( FracDigits : MCWord ) ;
                                 Floated : ( TotalWidth : MCWord ) )
              end ;

   { types required by the arithmetic unit, chapter 8 }

   OpType         = ( Plus,Minus,Mul,Idiv,Imod ) ;
   Operands       = MCWord ;





var


   { The following global variable holds the representation of an  }
   { undefined P-machine memory word. This value cannot be repres- }
   { ented in Standard Pascal and is therefore constructed on a    }
   { bit by bit basis by the procedure SetMCUndefined when the     }
   { P-Machine is initialised.                                     }

   Undefined : MCWord ;

   { P-machine global memory }

   Memory : array[WordRange] of MCWord ;


   { P-machine global address registers }

   ProgramCounter : ByteRange ;
   CodeBase : WordRange ;
   NewFrame,EnclosingFrame,CurrentFrame,GlobalFrame : WordRange;
   StackTop,HeapTop : WordRange ;
   Address : MCWord ;


   { P-machine global arithmetic registers }

   RealResult,RealOperand,RealOp1,RealOp2 : MCReal ;
   Result,Operand,Operand1,Operand2 : Operands ;


   { P-machine global set registers }

   Element,Element1,Element2 : MCWordForm ;
   SetDetails : record
                  Size1,Size2,Subset,Stride : MCWord ;
                  Base1,Base2 : WordRange
                end ;


   { P-machine global condition code }

   Overflowed : Boolean ;


   { variables used by the machine controller, chapter 3 }

   Control : array[OptionType] of MCWord ;
   Termination : Terminus ;
   ErrorCode : Scalar ;
   CorpseFile : file of MCWord ;
   ChecksOn : Boolean ;


   { variable used by the memory manager, chapter 5 }

   KeyValue : KeyValues ;


   { global variable used by the system file manager and the  }
   { Pascal I/O manager, chapters 6 & 7                       }

   FileIndex : MCWord ;


   { variable used by the system file manager, chapter 6 }

   SystemFiles : array [IndexRange] of SysFile  ;


   { variables used by the Pascal I/O manager, chapter 7 }

   InputId,OutputId : IndexRange ;
   PascalFiles : array [IndexRange] of PascFileDescriptor ;
   NullTransfer : IOTransfer ;

   { variable used by the loader, chapter 9 }

   CodeFile: file of MCWord ;


{

                          CHAPTER 3




                      P-Machine Control





 Overall operation of the P-machine is controlled by reading values
 of  the  compile-time options from the first 8 words of the P-code
 file.  In particular, the global variable  ChecksOn  will  be  set
 true  if  the  program was compiled with the Boolean option CHECKS
 selected. Thereafter, this variable is used to control all aspects
 of the P-machine concerned with the detection of run-time errors.

 Errors detected by the P-machine  are  classified  as  'run-time',
 'system',  or  'fatal'  according  to whether they derive from the
 executing program, from an  internal  P-machine  system-limit,  or
 from a corruption or malfunction of the P-machine itself. The fol-
 lowing procedures select an error-code from an appropriate 'base',
 and  record the reason for termination in the global variable Ter-
 mination.


                                                                     }

procedure Error(CheckCode: Scalar);

  begin
    ErrorCode := CheckCode + RuntimeBase;
    Termination := RuntimeError;
    goto 13
  end { error };

procedure SystemError(SystemCode: Scalar);

  begin
    ErrorCode := SystemCode + SystemBase;
    Termination := SystemLimit;
    goto 13
  end { systemerror };

procedure FatalError(FatalCode: Scalar);

  begin
    ErrorCode := FatalCode + FatalBase;
    Termination := Fatal;
    goto 13
  end { fatalerror };

procedure InitController;

  var Option: OptionType;

  begin

    { initialise important variables }
    ErrorCode := 0;
    ProgramCounter := 0;
    CurrentFrame := 0;
    GlobalFrame := 0;
    StackTop := 0;
    HeapTop := 0;
    for Option := Listing to Other do Control[Option] := 0;

    { now open the CodeFile }
    reset(CodeFile);
    if eof(CodeFile) then FatalError(4);
    for Option := Listing to Other do
      begin
        Control[Option] := CodeFile^;
        get(CodeFile)
      end;
    ChecksOn := Control[Checks] <> 0
  end { initcontroller };

{

 Errors detected by the  P-machine  are  analysed  by  passing  all
 necessary  information  in a 'corpse-file' to the Post-mortem Gen-
 erator. This includes:

 (a)  The error-code for subsequent error identification.

 (b)  The values of global P-machine variables  necessary  for  the
      production of a variable-dump.

 (c)  Control  information  which  indicates  the  values  of   the
      compile-time options CHECKS and PMD.

 (d)  The contents of the P-machine memory.

 Control is then passed serially to the Post-mortem  Generator  for
 all  subsequent  error  analysis.   The  means  by  which  this is
 achieved is necessarily system-dependent.

                                                                     }

procedure CallPMD;

  var Index: WordRange;

  procedure PutCorpseWord(Word: MCWord);

    begin
      CorpseFile^ := Word;
      put(CorpseFile)
    end { putcorpseword };

  begin
    rewrite(CorpseFile);

    { the 8-word control block }
    PutCorpseWord(ErrorCode);
    PutCorpseWord(ProgramCounter);
    PutCorpseWord(CurrentFrame);
    PutCorpseWord(GlobalFrame);
    PutCorpseWord(StackTop);
    PutCorpseWord(HeapTop);
    PutCorpseWord(Control[Checks]);
    PutCorpseWord(Control[PMDump]);

    { dump the program corpse }
    for Index := 0 to MCMaxMemWord do
      begin
        CorpseFile^ := Memory[Index];
        put(CorpseFile)
      end;

    {******************************************************}
    {* make system-dependent call to Postmortem Generator *}
    {******************************************************}
    writeln;
    writeln(' Runtime error ', ErrorCode: 3, ' occurred.');
    writeln(' Call Postmortem Generator.')
  end { callpmd };

{

 The procedure EndController calls the procedure  CallPMD  for  all
 programs that terminate abnormally.


                                                                     }

procedure EndController;

  begin
    if Termination <> Normal then CallPMD
  end { endcontroller };


{

                          CHAPTER 4




                          P-Machine
                       Micro Operations





 The procedures in this  chapter  implement  micro-operations  upon
 which  many  of  the  higher-level  instructions depend.  They are
 grouped into three sections.

 4.1  Part-Word Access

 The following procedures are used to access and update memory on a
 bit  or  byte basis.  Since the shift and mask operations normally
 associated with such accesses are not available in  Standard  Pas-
 cal, a familiar technique of employing a variant record MCWordForm
 as a type-recasting mechanism is used.  Note the technique  itself
 is  non-standard  since  the  variant-reselection  implied  by the
 recast strictly renders the  variant-field  undefined.   For  this
 reason, notional recast statements have been retained as comments.


                                                                     }

procedure MCSetBit ( var Word : MCWordForm ; BitIndex : MCBitIndex ) ;

   var
      TheBit : TAMCastBit;

   begin
      if HostBitOrder = LeftToRight
      then
         TheBit.Index := MCMaxBitNum-BitIndex
      else
         TheBit.Index := BitIndex;
      Word.BitSet:=Word.BitSet + [TheBit.Bitt];
   end { mcsetbit } ;

procedure MCClearBit ( var Word : MCWordForm ; BitIndex : MCBitIndex ) ;

   var
      TheBit : TAMCastBit;

   begin
      if HostBitOrder = LeftToRight
      then
         TheBit.Index := MCMaxBitNum-BitIndex
      else
         TheBit.Index := BitIndex;
      Word.BitSet:=Word.BitSet - [TheBit.Bitt];
   end { mcsetbit } ;

procedure MCGetBit ( Word : MCWordForm ; BitIndex : MCBitIndex ;
                     var ResultBit : MCBit ) ;

   var
      TheBit : TAMCastBit;

   begin
      if HostBitOrder = LeftToRight
      then
         TheBit.Index := MCMaxBitNum-BitIndex
      else
         TheBit.Index := BitIndex;
      ResultBit := ord( TheBit.Bitt in Word.BitSet);
   end { mcgetbit } ;


procedure MCSetByte(var MachineWord: MCWordForm;
                    ByteIndex: MCByteIndex; DataByte: MCByte);

  begin
    { recast(machineword,asbytes) }
    if HostByteOrder = LeftToRight
    then MachineWord.WChars[MCMaxByteNum - ByteIndex] := chr(DataByte)
    else MachineWord.WChars[ByteIndex] := chr(DataByte)
  end { mcsetbyte };

procedure MCGetByte(MachineWord: MCWordForm; ByteIndex: MCByteIndex;
                    var ResultByte: MCByte);

  begin
    { recast(machineword,asbytes) }
    if HostByteOrder = LeftToRight
    then ResultByte := ord(MachineWord.WChars[MCMaxByteNum - ByteIndex])
    else ResultByte := ord(MachineWord.WChars[ByteIndex])
  end { resultbyte };

{

 The function MCEmptyField  checks  a  bit-field  on  a  bit-by-bit
 basis,  and is used in the implementation of run-time checks asso-
 ciated with set-assignment.

                                                                     }

function MCEmptyField(MachineWord: MCWordForm;
                      FirstBit, LastBit: MCBitRange): Boolean;

  var Bit: MCBit;

  begin
    Bit := 0;
    while (FirstBit <= LastBit) and (Bit = 0) do
      begin
        MCGetBit(MachineWord, FirstBit, Bit);
        FirstBit := FirstBit + 1
      end;
    MCEmptyField := (Bit = 0)
  end { mcemptyfield };

{

 4.2  Undefined Value Manipulation

 The 'undefined' P-machine word  value  cannot  be  represented  in
 Standard Pascal and is therefore constructed on a bit-by-bit basis
 and held in the global variable Undefined.  Thereafter,  the  test
 for  undefined  values,  is  performed  by  function  MCUndefined,
 although the comparison of  non  integer  values  implied  is  not
 necessarily  operational.   The procedures PresetMemory and Memor-
 yUndefined provide multi-word memory initialisation and checking.

                                                                     }

procedure SetMCUndefined;

  var MachineWord: MCWordForm;

  begin
    MachineWord.WValue := 0;
    MCSetBit(MachineWord, MCMaxBitnum);
    Undefined := MachineWord.WValue
  end { setmcundefined };

function MCUndefined(Word: MCWord): Boolean;

  begin MCUndefined := (Word = Undefined) end;

procedure PresetMemory(Base, WordCount: WordRange);

  var i: WordRange;

  begin
    for i := Base to Base + WordCount - 1 do Memory[i] := Undefined
  end { presetmemory };

function MemoryUndefined(Base, WordCount: WordRange): Boolean;

  label 9;

  var Result: Boolean;
      i: WordRange;

  begin
    Result := false;
    for i := Base to Base + WordCount - 1 do
      if MCUndefined(Memory[i])
      then
        begin
          Result := true;
          goto 9
        end;
    9: MemoryUndefined := Result
  end { memoryundefined };

{

 4.3  Packing and Unpacking

 The following procedures provide packing and unpacking required by
 the  instructions  implementing  the  standard procedures pack and
 unpack, and by the implementation of packed-files.  The procedures
 Packpc  and  Unpackpc additionally check each source value against
 the appropriate undefined value and  terminate  execution  with  a
 run-time error if required.

                                                                     }

procedure Packp(UnpackedBase, Elements, PackedBase: WordRange;
                ElsPerWord: BitRange);

  var RightMost, FieldSize, BitIndex: BitRange;
      UpIndex, PkIndex: WordRange;
      OpBuffer, IpBuffer: MCWordForm;
      Bit: MCBit;

  begin
    FieldSize := MCBitsPerWord div ElsPerWord;
    RightMost := 0;
    UpIndex := 0;
    PkIndex := 0;
    OpBuffer.WValue := 0;
    while UpIndex <> Elements do
      begin
        IpBuffer.WValue := Memory[UnpackedBase + UpIndex];
        for BitIndex := 0 to FieldSize - 1 do
          begin
            MCGetBit(IpBuffer, BitIndex, Bit);
            if Bit <> 0 then MCSetBit(OpBuffer, RightMost + BitIndex)
          end;
        RightMost := RightMost + FieldSize;
        UpIndex := UpIndex + 1;
        if UpIndex mod ElsPerWord = 0
        then
          begin
            Memory[PackedBase + PkIndex] := OpBuffer.WValue;
            PkIndex := PkIndex + 1;
            RightMost := 0;
            OpBuffer.WValue := 0
          end
      end;
    if UpIndex mod ElsPerWord <> 0
    then Memory[PackedBase + PkIndex] := OpBuffer.WValue
  end { packp };

procedure Unpackp(PackedBase, UnpackedBase, Elements: WordRange;
                  ElsPerWord: BitRange);

  var RightMost, FieldSize, BitIndex: BitRange;
      UpIndex, PkIndex: WordRange;
      IpBuffer, OpBuffer: MCWordForm;
      Bit: MCBit;

  begin
    FieldSize := MCBitsPerWord div ElsPerWord;
    UpIndex := 0;
    PkIndex := 0;
    while UpIndex <> Elements do
      begin
        if UpIndex mod ElsPerWord = 0
        then
          begin
            IpBuffer.WValue := Memory[PackedBase + PkIndex];
            PkIndex := PkIndex + 1;
            RightMost := 0
          end;
        OpBuffer.WValue := 0;
        for BitIndex := 0 to FieldSize - 1 do
          begin
            MCGetBit(IpBuffer, BitIndex + RightMost, Bit);
            if Bit <> 0 then MCSetBit(OpBuffer, BitIndex)
          end;
        Memory[UnpackedBase + UpIndex] := OpBuffer.WValue;
        RightMost := RightMost + FieldSize;
        UpIndex := UpIndex + 1
      end
  end { unpackp };

procedure Packpc(UnpackedBase, Elements, PackedBase: WordRange;
                 ElsPerWord: BitRange);

  var RightMost, FieldSize, BitIndex: BitRange;
      UpIndex, PkIndex: WordRange;
      IpBuffer, OpBuffer: MCWordForm;
      Bit: MCBit;

  begin
    FieldSize := MCBitsPerWord div ElsPerWord;
    RightMost := 0;
    UpIndex := 0;
    PkIndex := 0;
    OpBuffer.WValue := 0;
    while UpIndex <> Elements do
      begin
        IpBuffer.WValue := Memory[UnpackedBase + UpIndex];
        if MCUndefined(IpBuffer.WValue) then Error(27);
        for BitIndex := 0 to FieldSize - 1 do
          begin
            MCGetBit(IpBuffer, BitIndex, Bit);
            if Bit <> 0 then MCSetBit(OpBuffer, RightMost + BitIndex)
          end;
        RightMost := RightMost + FieldSize;
        UpIndex := UpIndex + 1;
        if UpIndex mod ElsPerWord = 0
        then
          begin
            Memory[PackedBase + PkIndex] := OpBuffer.WValue;
            PkIndex := PkIndex + 1;
            RightMost := 0;
            OpBuffer.WValue := 0
          end
      end;
    if UpIndex mod ElsPerWord <> 0
    then Memory[PackedBase + PkIndex] := OpBuffer.WValue
  end { packpc };

procedure Unpackpc(PackedBase, UnpackedBase, Elements: WordRange;
                   CheckValue: MCWord; ElsPerWord: BitRange);

  var RightMost, FieldSize, BitIndex: BitRange;
      UpIndex, PkIndex: WordRange;
      IpBuffer, OpBuffer: MCWordForm;
      Bit: MCBit;

  begin
    FieldSize := MCBitsPerWord div ElsPerWord;
    UpIndex := 0;
    PkIndex := 0;
    while UpIndex <> Elements do
      begin
        if UpIndex mod ElsPerWord = 0
        then
          begin
            IpBuffer.WValue := Memory[PackedBase + PkIndex];
            PkIndex := PkIndex + 1;
            RightMost := 0
          end;
        OpBuffer.WValue := 0;
        for BitIndex := 0 to FieldSize - 1 do
          begin
            MCGetBit(IpBuffer, BitIndex + RightMost, Bit);
            if Bit <> 0 then MCSetBit(OpBuffer, BitIndex)
          end;
        if OpBuffer.WValue = CheckValue then Error(30);
        Memory[UnpackedBase + UpIndex] := OpBuffer.WValue;
        RightMost := RightMost + FieldSize;
        UpIndex := UpIndex + 1
      end
  end { unpackpc };


{

                          CHAPTER 5




                      Memory Management





 The P-machine possesses a linear, unstructured, memory implemented
 as an array of words whose capacity is defined by the configurable
 constant MCMaxMemWord.  In addition to direct  indexing,  bit  and
 byte  accessing  is supported by the micro-functions in Chapter 4.
 Bits are numbered from 0 at the least  significant end  to  MCMax-
 BitNum. Bytes are numbered correspondingly from 0 to MCMaxByteNum.
 Word capacity is given by MCBitsPerWord and MCBytesPerWord.

 The loader reads the contents of the P-code file into low  address
 memory.   The remaining 'free memory' is available for use as both
 a stack and heap.  The following memory-map indicates the  general
 layout of these three areas:


                   .--------------. MCMaxMemWord
                   |              |
                   |              |
                   |              |
                   |    Heap      |
                   |              |
                   |              |
                   |--------------|
                   |      |       |
                   |      |       |
                   |      v       |
                   |     Free     |
                   |    Memory    |
                   |              |
                   |      ^       |
                   |      |       |
                   |      |       |
                   |--------------|
                   |              |
                   |              |
                   |              |
                   |    Stack     |
                   |              |
                   |              |
                   |              |
                   |--------------|
                   |              |
                   |              |
                   |              |
                   |    Code      |
                   |              |
                   |              |
                   !______________! 0



 The remaining sections in  this  chapter  deal  with  the  dynamic
 management of the stack and heap.

 5.1  Acquisition and Control

 Throughout program execution, the stack  and  heap  acquire  space
 from  free  memory.  The following utilities monitor the growth of
 these structures and ensure that neither area will encroach on the
 other  by  an amount less than the configurable constant MinBlock.
 Should violation of this condition be anticipated, stack  or  heap
 overflow errors are reported as appropriate.


                                                                     }

function Safety(Request: WordRange): MCScalar;

  begin Safety := Request + MinBlock end;

procedure ClaimStack(Amount: WordRange);

  begin
    if StackTop + Safety(Amount) >= HeapTop then SystemError(11);
    StackTop := StackTop + Amount
  end { claimstack };

procedure FreeStack(Amount: WordRange);

  begin
    if StackTop - Amount < CurrentFrame - 1 then FatalError(2);
    StackTop := StackTop - Amount
  end { freestack };

procedure ClaimHeap(Amount: WordRange);

  begin
    if HeapTop - Safety(Amount) <= StackTop then SystemError(12);
    HeapTop := HeapTop - Amount
  end { claimheap };

procedure FreeHeap(Amount: WordRange);

  begin HeapTop := HeapTop + Amount end;

{

 5.2  The Runtime Stack

 All directly referenceable program data  is  held  on  a  run-time
 stack.   The stack is built from a series of stack-frames, each of
 which holds the local data of an  active  procedure,  function  or
 program  block.   The  stack-frames  are  linked using the classic
 Dijkstra method of static and dynamic chains to provide  non-local
 addressability and dynamic recovery of storage. The layout of each
 stack-frame is summarised in the diagram:


                   |      .       |
                   |      .       |
                   |     Free     |
                   |    Memory    |
                   |      .       |
                   |      .       |
                   |  Expression  |<------ Stack Top
                   |--------------|
                   |  Lock Stack  |
                   |--------------|
                   | Temporaries  |
                   |--------------|
                   |              |
                   |              |
                   |    Locals    |
                   |              |
                   |--------------|
                   |  Parameters  |
                   |--------------|<------ Frame Base
                   | Frame Header |
                   |--------------|


 Each stack-frame has a six-word  frame-header  indexed  negatively
 from the frame base.  The contents of the header are as follows:

 (1)  Word 1 holds the  static  link  to  the  stack-frame  of  the
      enclosing block.

 (2)  Word 2 holds the 'code-base' of the code for the given block.

 (3)  Word 3 holds the dynamic link to the stack-frame of the  cal-
      ling block.

 (4)  Word 4 holds the return address to the code  of  the  calling
      block.

 (5)  Word 5 holds the current depth of the lock-stack measured  in
      terms of the number of entries.

 (6)  Word 6 holds the address of the top-most lock-stack entry.

 The static chain is supplemented by variables CurrentFrame, Enclo-
 singFrame,  and  GlobalFrame  which in effect operate as a limited
 display by providing fast  access  to  variables  of  the  implied
 block.   Variables belonging to blocks nested at intermediate lev-
 els are located using the static chain.

 The following procedures  provide  utilities  for  establishing  a
 stack-frame  header, and for locating the stack-frame of an enclos-
 ing block by following the static chain.

                                                                     }

procedure SetFrameHeader;

  begin
    Memory[NewFrame - SLOffset] := EnclosingFrame;
    Memory[NewFrame - CBOffset] := CodeBase;
    Memory[NewFrame - DLOffset] := CurrentFrame;
    Memory[NewFrame - RAOffset] := ProgramCounter;
    Memory[NewFrame - LDOffset] := 0;
    Memory[NewFrame - LTOffset] := StackTop
  end { setframeheader };

function OuterFrame(FrameOffset: MCScalar): WordRange;

  var Index: MCScalar;
      FrameBase: WordRange;

  begin
    FrameBase := CurrentFrame;
    for Index := 1 to FrameOffset do
      FrameBase := Memory[FrameBase - SLOffset];
    OuterFrame := FrameBase
  end { outerframe };

{

 The static part of each stack-frame contains storage  for  parame-
 ters,  function  results, local variables, and temporaries used by
 the generated code.  Provision is also  made  for  a  'lock-stack'
 which  is  used  to  provide  a general mechanism for applying and
 releasing reference locks.  Each entry  on  the  lock-stack  is  a
 three-word  component  comprising  a  lock-bit  value,  and a lock
 word-address/bit-offset pair.  A lock is applied  by  pushing  the
 current  setting  and  lock location onto the lock-stack, and then
 setting the lock.  A lock is released by  popping  the  lock-stack
 entry and restoring the lock to the unstacked setting.

 Both the lock-stack and local storage requirements  that  comprise
 the  static part of a stack frame are computed at compile time and
 embedded in the  generated  code.   During  the  creation  of  the
 stack-frame  word 5 of the header is initialised to zero, and word
 6 is set to the base of the lock-stack  area.   Locks  are  pushed
 onto  the  stack individually by the procedure PushLock and popped
 by the procedure PopLock.

                                                                     }

procedure PushLock(LValue: MCBit; LWord: WordRange; LBit: BitRange);

  var LockTop, LockDepth: MCWord;

  begin
    LockDepth := Memory[CurrentFrame - LDOffset];
    LockTop := Memory[CurrentFrame - LTOffset];
    Memory[LockTop + 1] := LValue;
    Memory[LockTop + 2] := LWord;
    Memory[LockTop + 3] := LBit;
    Memory[CurrentFrame - LDOffset] := LockDepth + 1;
    Memory[CurrentFrame - LTOffset] := LockTop + LockSize
  end { pushlock };

procedure PopLock(var LValue: MCBit; var LWord, LBit: MCWord);

  var LockTop, LockDepth: WordRange;

  begin
    LockDepth := Memory[CurrentFrame - LDOffset];
    LockTop := Memory[CurrentFrame - LTOffset];
    LBit := Memory[LockTop];
    LWord := Memory[LockTop - 1];
    LValue := Memory[LockTop - 2];
    Memory[CurrentFrame - LDOffset] := LockDepth - 1;
    Memory[CurrentFrame - LTOffset] := LockTop - LockSize
  end { poplock };

{

 The top-most part of every  stack-frame  is  used  for  expression
 evaluation,  and  operands  are pushed and popped using the global
 variable StackTop.  During  function  evaluation,  the  new  stack
 frame  is  pushed on top of the current evaluation-stack.  On exit
 from the function, the stack-frame is  removed  and  the  function
 result  pushed  as an expression operand.  The following utilities
 provide for the pushing and popping of the evaluation stack.

                                                                     }

procedure Push(Item: MCWord);

  begin
    StackTop := StackTop + 1;
    Memory[StackTop] := Item
  end { push };

procedure Pop(var Item: MCWord);

  begin
    Item := Memory[StackTop];
    StackTop := StackTop - 1
  end { pop };

procedure PopOperands;

  begin
    Pop(Operand1);
    Pop(Operand2)
  end { popoperands };

procedure PushLogical(Item: MCWordSet);

  var Gate: MCWordForm;

  begin
    Gate.WSet := Item;
    { recast(gate,asvalue) }
    Push(Gate.WValue)
  end { pushlogical };

procedure PopLogical(var Item: MCWordSet);

  var Gate: MCWordForm;

  begin
    Pop(Gate.WValue);
    { recast(gate,asset) }
    Item := Gate.WSet
  end { poplogical };

procedure PushReal(Item: MCReal);

  var Gate: RealMap;
      i: 1..MCRealSize;

  begin
    Gate.RealValue := Item;
    for i := 1 to MCRealSize do Push(Gate.Word[i])
  end { pushreal };

procedure PopReal(var Item: MCReal);

  var Gate: RealMap;
      i: 1..MCRealSize;

  begin
    for i := MCRealSize downto 1 do Pop(Gate.Word[i]);
    Item := Gate.RealValue
  end { popreal };

procedure PopRealOperands;

  begin
    PopReal(RealOp1);
    PopReal(RealOp2)
  end { poprealoperands };

procedure Spray(Pattern, WordBase: MCWord; WordCount: Scalar);

  begin
    while WordCount > 0 do
      begin
        Memory[WordBase + WordCount - 1] := Pattern;
        WordCount := WordCount - 1
      end
  end { spray };

procedure MoveWords(FromBase, ToBase, Size: WordRange);

  var Index: WordRange;

  begin
    if FromBase < ToBase
    then
      for Index := Size - 1 downto 0 do
        Memory[ToBase + Index] := Memory[FromBase + Index]
    else
      for Index := 0 to Size - 1 do
        Memory[ToBase + Index] := Memory[FromBase + Index]
  end { move words };

procedure LoadWords(Base, Size: WordRange);

  var Too: WordRange;

  begin
    Too := StackTop + 1;
    ClaimStack(Size);
    MoveWords(Base, Too, Size)
  end { loadwords };

procedure StoreWords(Base, Size: WordRange);

  begin
    FreeStack(Size);
    MoveWords(StackTop + 1, Base, Size)
  end { storewords };

{

 5.3  Runtime Heap

 Storage for  file  buffers  and  dynamic  variables  accessed  via
 pointers  is  allocated on a run-time heap.  Initially the heap is
 empty and grows by claiming 'free' memory.  As storage is  dynami-
 cally  released however, it is chained on a 'free-list' for subse-
 quent reallocation.  The free-list contains linked blocks of inac-
 tive  memory  of known size.  The diagram below shows three active
 (shaded) areas of memory interspersed with  two  chained  inactive
 blocks.  The highest memory word at address MaxMemWord is reserved
 for the head of the free-list.


   Free List ----->.--------------.
                   |--------------|---.
                   |//////////////|   |
                   |//////////////|   |
                   |////  A  /////|   |
                   |//////////////|   |
                   |//////////////|   |
                   |--------------|   |
                   |              |   |
                   |      B       |   |
                   |              |   |
                   |              |   |
                   |--------------|-------.
                   |--------------|<--!   |
                   |//////////////|       |
                   |//////////////|       |
                   |//////////////|       |
                   |///// C //////|       |
                   |//////////////|       |
                   |//////////////|       |
                   |--------------|       |
                   |              |       |
                   |              |       |
                   |              |       |
                   |      D       |       |
                   |              |       |
                   |              |       |
                   |--------------|<------!
                   |--------------|
                   |//////////////|
                   |//////////////|
                   |///// E //////|
                   |//////////////|
                   |//////////////|
                   |--------------|<------ HeapTop
                   |      .       |
                   |      .       |
                   |    Free      |
                   |   Memory     |
                   |      .       |
                   |      .       |


 The procedure Acquire will allocate heap storage  from  the  free-
 list  if  possible,   but will call the procedure ClaimHeap in the
 event of the free-list being unable to satisfy the request.   The
 global variable HeapTop defines the low address boundary of active
 heap storage.

                                                                     }

procedure Acquire(Amount: WordRange; var Base: WordRange);

  var Current, Previous, Next, CurrentSize, SurplusBlock, Surplus:
        WordRange;
      Found: Boolean;

  begin
    if Amount < MinHeapBlock then Amount := MinHeapBlock;
    Current := FreeList;
    Found := false;
    repeat
      Previous := Current;
      Current := Memory[Current];
      if Current <> Null
      then
        begin
          CurrentSize := Memory[Current + 1];
          if CurrentSize >= Amount
          then
            begin
              Found := true;
              Next := Memory[Current];
              Surplus := CurrentSize - Amount;
              if Surplus = 0
              then Memory[Previous] := Next
              else
                begin
                  SurplusBlock := Current + Amount;
                  Memory[Previous] := SurplusBlock;
                  Memory[SurplusBlock + 1] := Surplus;
                  Memory[SurplusBlock] := Next
                end
            end
        end
    until (Current = Null) or Found;
    if not Found
    then
      begin
        ClaimHeap(Amount);
        Current := HeapTop
      end;
    Base := Current
  end { acquire };

{

 Storage that is released is chained onto the free-list or returned
 to  free  memory  depending  upon  its position in relation to the
 heap-top and the free-list.  The procedure Release  also  performs
 simple  garbage collection to ensure that whenever an active block
 is released, it is merged with any adjacent inactive blocks.  Thus
 the  allocated block C will be merged on release with both B and D
 to form a single aggregate chunk  of  inactive  memory.   Alterna-
 tively, if E were released it will be merged with D and the aggre-
 gate returned to free-memory.

                                                                     }

procedure Release(Amount, Base: WordRange);

  var StartGarbage, EndGarbage, Current, EndCurrent, Previous:
        WordRange;
      Adjacent: Boolean;

  begin
    if Amount < MinHeapBlock then Amount := MinHeapBlock;
    StartGarbage := Base;
    if (StartGarbage < HeapTop) or (Base + Amount > MCMaxMemWord) or
       (Amount <= 0)
    then FatalError(3);
    EndGarbage := Base + Amount;
    Current := FreeList;
    repeat
      Previous := Current;
      Current := Memory[Current]
    until (Current = Null) or (Current <= EndGarbage);
    Adjacent := true;
    while (Current <> Null) and Adjacent do
      begin
        EndCurrent := Current + Memory[Current + 1];
        if (Current = EndGarbage) or (EndCurrent = StartGarbage)
        then
          begin
            Adjacent := true;
            if Current = EndGarbage
            then EndGarbage := EndCurrent
            else
              if EndCurrent = StartGarbage
              then StartGarbage := Current;
            Current := Memory[Current]
          end
        else Adjacent := false
      end;
    if StartGarbage > HeapTop
    then
      begin
        Memory[StartGarbage + 1] := EndGarbage - StartGarbage;
        Memory[StartGarbage] := Current;
        Memory[Previous] := StartGarbage
      end
    else
      begin
        FreeHeap(EndGarbage - StartGarbage);
        Memory[Previous] := Null
      end
  end { release };

{

 The procedures Newp and Disposep provide a  secure  implementation
 of  dynamic  variables  accessed  via pointers by allocating extra
 storage for the implementation of run-time checks.   Each  dynamic
 variable  is  associated with a two-word header, the first word of
 which  holds  a   unique   key   allocated   in   the   range   [-
 MCMaxint..+MCMaxint].   The  second  word is preset to zero but is
 subsequently used to hold a lock-bit and a monitoring count of the
 number of variant levels (if any) preselected by the extended form
 of new.  The diagram below shows how the pointer  contents  relate
 to the  heap-storage.

                .---------------.
 2-word pointer | address | key |
                !_______________!    |--------------|  ^
                     |               |              |  |
                     |               |              |  |
                     |               |              |  |
                     |               |              |  | dynamic
                     |               |              |  | variable
                     |               |              |  |
                     |               |              |  |
                     |               |              |  |
                     |               |              |  v
                     .-------------->|--------------|  ^
                                     |lock|  |levels|  |
                                     |--------------|  | header
                                     |     key      |  |
                                     |--------------|  v



 When checks are on, code is generated by  the  Model  Compiler  to
 verify  any  pointer  dereference  by  comparing the keys embedded
 within the pointer and the  dynamic  variable  header.   When  the
 variable  is disposed code is generated to check the lock bit, and
 to check the levels count for consistency.  The procedure Disposep
 subsequently  resets  the  header  words to the 'undefined' value,
 enabling dereferences via dangling pointers to be caught by subse-
 quent  key  mismatch.  Note that total security is not guaranteed,
 for subsequent repartitioning of  heap-storage  coupled  with  the
 coincidental reconstruction of a previous key value as valid data,
 could fool the dangling pointer check.  However, such a situation,
 though not impossible is very unlikely.

 The P-machine allocates key-values in the range of integers there-
 bye  leading  to  a system-limit on the number of unique keys that
 can be generated.  Since this limit is 2*MCMaxint+1, it is  not  a
 serious implementation restriction.

                                                                     }

procedure NewKeyValue;

  begin
    if KeyValue = MCMaxInt
    then SystemError(13) else KeyValue := KeyValue + 1
  end { newkeyvalue };

procedure Newp(Size: WordRange);

  var Amount, HeapAddress: WordRange;
      PtrAddress: MCWord;

  begin
    Amount := Size + 2 * ord(ChecksOn);
    Pop(PtrAddress);
    Acquire(Amount, HeapAddress);
    if ChecksOn
    then
      begin
        NewKeyValue;
        Memory[HeapAddress] := 0;
        Memory[HeapAddress + 1] := KeyValue;
        HeapAddress := HeapAddress + 2;
        PresetMemory(HeapAddress, Size);
        Memory[PtrAddress + 1] := KeyValue
      end;
    Memory[PtrAddress] := HeapAddress
  end { newp };

procedure Disposep(Size: WordRange);

  var Amount, HeapAddress: WordRange;
      PtrAddress: MCWord;

  begin
    Pop(PtrAddress);
    HeapAddress := Memory[PtrAddress];
    if ChecksOn
    then
      begin
        Amount := Size + 2;
        HeapAddress := HeapAddress - 2;
        PresetMemory(HeapAddress, 2)
      end
    else Amount := Size;
    Release(Amount, HeapAddress)
  end { disposep };

{


 The following procedure initialises the memory manager by creating
 a  null  free-list  and  positioning  the  HeapTop  pointer at the
 highest memory word.  The first available  key  value  is  set  to
 -MCMaxint.


                                                                     }

procedure InitMemory;

  begin
    Memory[FreeList] := Null;
    HeapTop := MCMaxMemWord;
    KeyValue := -MCMaxInt
  end { initmemory };


{

                          CHAPTER 6




                    System File Management





 The implementation of file-variables requires some form of mapping
 between  the  variable  and  the  corresponding physical file.  In
 practice, such a mapping is system-dependent and cannot be given a
 generalised  expression  in Standard Pascal.  To enable the Pascal
 version of the P-machine to run however, a simulated filestore has
 been  implemented  which  allows limited creation of 'local files'
 together with communication with the external environment by means
 of  the  program parameters Input and Output.  This implementation
 is sufficient to run all programs in the Pascal Validation  Suite,
 and can be summarised as follows:

 (1)  A local filestore is simulated by the array variable  System-
      Files.   Each  element  of  the  array represents a P-machine
      local file implemented as a linear  list  of  memory  blocks.
      Each  block has a two-word header containing the link and the
      block capacity in bytes.  The fields of the record-type  Sys-
      File  contain a pointer BlockList to the memory blocks, and a
      pointer CurrentBlock identifying the  block  currently  being
      read  or written.  For packed files, a field Padding contains
      the number of unused bits in the last word of the  file.  The
      maximum number of files that can be opened simultaneously  is
      limited by the configurable constant MaxFileIndex.

 (2)  The System File Manager provides an interface for opening and
      closing  files,  and  for file-block transfers defined by the
      procedures Open, Close, Getbytes, and PutBytes.   A  call  to
      each  of  these  procedures  assumes that the global variable
      FileIndex identifies the  file  in  question.   If  FileIndex
      identifies  either  Input  or Output, calls to Open and Close
      are ignored, and calls to GetBytes and  PutBytes  ensure  the
      block  transfer occurs using the P-machine program parameters
      Input and Output.

 The following procedures are used internally by  the  System  File
 Manager.


                                                                     }

function WordsFor(ByteAmount: ByteRange): WordRange;

  var WordAmount: WordRange;

  begin
    WordAmount := ByteAmount div MCBytesPerWord;
    if ByteAmount mod MCBytesPerWord <> 0
    then WordAmount := WordAmount + 1;
    WordsFor := WordAmount
  end { wordsfor };

procedure AppendBlock(var Fyle: SysFile; Block: WordRange);

  begin
    with Fyle do
      begin
        if BlockList = Null
        then BlockList := Block else Memory[CurrentBlock] := Block;
        CurrentBlock := Block
      end
  end { appendblock };

procedure Destroy(var Fyle: SysFile);

  var Amount: WordRange;

  begin
    with Fyle do
      while BlockList <> Null do
        begin
          CurrentBlock := BlockList;
          BlockList := Memory[BlockList];
          Amount := WordsFor(Memory[CurrentBlock + 1]);
          Release(Amount + 2, CurrentBlock)
        end
  end { destroy };

{

 The procedures comprising the System File Manager interface are as
 follows:
                                                                     }

procedure Open(Mode: IOmode);

  begin
    if (FileIndex <> InputId) and (FileIndex <> OutputId)
    then
      with SystemFiles[FileIndex] do
        if Mode = ForReading
        then CurrentBlock := BlockList
        else
          begin
            CurrentBlock := Null;
            BlockList := Null;
            Padding := 0
          end
  end { open };

procedure GetBytes(Buffer: WordRange; var Amount: IOTransfer;
                   var EndFile: Boolean);

  var Base: WordRange;
      ByteBuffer: MCWordForm;

  begin
    Amount := NullTransfer;
    if FileIndex = InputId
    then
      begin
        EndFile := eof(Input);
        if not EndFile
        then
          with Amount do
            begin
              Base := Buffer;
              ByteBuffer.WValue := 0;
              while not eoln(Input) do
                begin
                  MCSetByte
                    (ByteBuffer, ByteAmount mod MCBytesPerWord,
                       ord(Input^));
                  ByteAmount := ByteAmount + 1;
                  if ByteAmount mod MCBytesPerWord = 0
                  then
                    begin
                      Memory[Base] := ByteBuffer.WValue;
                      ByteBuffer.WValue := 0;
                      Base := Base + 1
                    end;
                  get(Input)
                end;
              if ByteAmount mod MCBytesPerWord <> 0
              then Memory[Base] := ByteBuffer.WValue;
              get(Input)
            end
      end
    else
      with SystemFiles[FileIndex] do
        if CurrentBlock = Null
        then EndFile := true
        else
          with Amount do
            begin
              ByteAmount := Memory[CurrentBlock + 1];
              MoveWords
                (CurrentBlock + 2, Buffer, WordsFor(ByteAmount));
              EndFile := false;
              CurrentBlock := Memory[CurrentBlock];
              if CurrentBlock = Null then BitAmount := Padding
            end
  end { getbytes };

procedure PutBytes(Buffer: WordRange; Amount: IOTransfer);

  var Base, NewBlock, Words: WordRange;
      Index: ByteRange;
      ByteIndex: MCByteIndex;
      CharByte: MCByte;
      ByteBuffer: MCWordForm;

  begin
    with Amount do
      if FileIndex = OutputId
      then
        begin
          Base := Buffer;
          for Index := 0 to ByteAmount - 1 do
            begin
              ByteIndex := Index mod MCBytesPerWord;
              if ByteIndex = 0
              then
                begin
                  ByteBuffer.WValue := Memory[Base];
                  Base := Base + 1
                end;
              MCGetByte(ByteBuffer, ByteIndex, CharByte);
              Output^ := chr(CharByte);
              put(Output)
            end;
          writeln(Output)
        end
      else
        begin
          Words := WordsFor(ByteAmount);
          Acquire(Words + 2, NewBlock);
          MoveWords(Buffer, NewBlock + 2, Words);
          Memory[NewBlock] := Null;
          Memory[NewBlock + 1] := ByteAmount;
          AppendBlock(SystemFiles[FileIndex], NewBlock);
          SystemFiles[FileIndex].Padding := BitAmount
        end
  end { putbytes };

procedure Close(Status: FileStatus);

  begin
    if (FileIndex <> InputId) and (FileIndex <> OutputId)
    then
      if Status = Detach then Destroy(SystemFiles[FileIndex])
  end { close };

procedure InitSysFiles;

  begin { nothing required }  end;


{

                          CHAPTER 7




                    Pascal I/O Management





 The Pascal I/O Manager provides both high  and  low-level  support
 for the instructions implementing the standard I/O functions.

 Low-level support for I/O is provided  by  a  buffering  mechanism
 accompanied  by  the  allocation of file-descriptors.  Every file-
 variable in Pascal must be implicitly  or  explicitly  initialised
 for  I/O  by  a  call  to  one of the standard procedures reset or
 rewrite. The first such call for a local file results in the allo-
 cation  of  a  file  descriptor,  a file buffer, and a file index.
 Descriptors are allocated by a linear  search  of  the  descriptor
 array  PascalFiles,  which records the index-position of the first
 unallocated descriptor in the variable FileIndex.  This  value  is
 embedded  in  word  0  of  the  file  variable  block, and is used
 thereafter to identify the file.  The following utilities  provide
 for the allocation and de-allocation of descriptors.


                                                                     }

procedure AllocatePascalFile;

  var TypeCode, ElementSize: MCWord;

  begin
    with PascalFiles[FileIndex] do
      begin
        Defined := true;
        Mode := NotSet;
        EndOfFile := true;
        TypeCode := Memory[Address + 2];
        BufferSize := Memory[Address + 3];
        ElementSize := Memory[Address + 4];
        TextFile := (TypeCode = TextCode);
        if TextFile
        then BufferWord := Address + 2
        else
          begin
            PointerWord := Address + 2;
            PackedFile := (TypeCode = PackedCode);
            if PackedFile
            then ElsPerWord := ElementSize
            else ComponentSize := ElementSize
          end
      end
  end { allocatepascalfile };

procedure DeAllocatePascalFile;

  begin PascalFiles[FileIndex].Defined := false
  end { deallocatepascalfile };

procedure NextFileIndex;

  begin
    FileIndex := 1;
    while PascalFiles[FileIndex].Defined do
      FileIndex := succ(FileIndex) mod (MaxFileIndex + 1)
  end { nextfileindex };

{

 When a file is opened, a buffer is allocated  from  the  heap  and
 used  in  all  I/O  transfers between the file and the addressable
 file buffer variable.  The size of the buffer required is obtained
 from  the  file-variable  block,  where  word 3 is preset with the
 storage requirement when the file  variable  is  initialised.   In
 addition word 2 of the block is preset with a code identifying the
 nature of the file, and word 4 with the  element  size  in  words.
 The  descriptor  fields BufferSize and FirstWord respectively hold
 the size of the allocated buffer and its start word address.

 Three different  buffer  organizations  are  used  depending  upon
 whether  the  file variable is a text-file, an unpacked word file,
 or a packed file.  These can be summarised as follows:

 (1)  For text files, the fields BytePointer and  LastByte  of  the
      descriptor   respectively   hold  the  byte  address  of  the
      current-component of the file,  and  the  last  byte  of  the
      buffer.   Word 2 of the file variable memory block is used as
      the (unpacked) buffer variable, and its word-address is  held
      in the descriptor field BufferWord.

 (2)  For an unpacked word file, the buffer variable is implemented
      by  a  'window'  that slides over the file-buffer.  Word 2 of
      the file variable holds the word-address of  the  window  and
      therefore functions as a pointer to the buffer-variable.  The
      fields PointerWord and LastWord  of  the  descriptor  respec-
      tively  hold  the  word-addresses of word 2 of the file vari-
      able, and the last  buffer  word.   The  field  ComponentSize
      holds the word size of the current component of the file.

 (3)  Packed files use a two-level buffering scheme  in  which  the
      primary  file  buffer  is  supplemented by a secondary buffer
      used for packing and unpacking components of the  file.   The
      capacity  of  the  secondary  buffer  is  determined from the
      number of file components that can be  packed  per  word.   A
      secondary  pointer held in the memory word addressed by Last-
      Word is used to mark undefined  secondary  buffer  locations,
      arising  when  the  last  word of a packed-file contains less
      than elements-per-word packed components.  This  allows  true
      end-of-file  to  be deduced correctly.  The field WordPointer
      of the descriptor is the address of the primary  buffer  word
      from   or  into  which  transfer  of  packed components takes
      place.

 Given these buffer/file organisations, the primitive operations of
 reset,  rewrite, get, and put are performed by the following seven
 procedures.

                                                                     }

procedure FillBuffer;

  var Amount: IOTransfer;
      Padding,FieldWidth: BitRange;

  begin
    with PascalFiles[FileIndex] do
      if TextFile
      then
        begin
          GetBytes(FirstWord, Amount, EndOfFile);
          BytePointer := FirstWord * MCBytesPerWord;
          LastByte := BytePointer + Amount.ByteAmount
        end
      else
        begin
          GetBytes(FirstWord, Amount, EndOfFile);
          LastWord := FirstWord + WordsFor(Amount.ByteAmount);
          if PackedFile
          then
            begin
              Padding := Amount.BitAmount;
              FieldWidth := MCBitsPerWord div ElsPerWord;
              Memory[LastWord] :=
                LastWord + Padding div FieldWidth;
              WordPointer := FirstWord;
              Memory[PointerWord] := LastWord
            end
          else Memory[PointerWord] := FirstWord
        end
  end { fillbuffer };

procedure FlushBuffer;

  var Amount: IOTransfer;

  begin
    Amount := NullTransfer;
    with PascalFiles[FileIndex] do
      if TextFile
      then
        begin
          Amount.ByteAmount :=
            BytePointer - FirstWord * MCBytesPerWord;
          PutBytes(FirstWord, Amount);
          BytePointer := FirstWord * MCBytesPerWord
        end
      else
        begin
          with Amount do
            if PackedFile
            then
              begin
                ByteAmount :=
                  (WordPointer - FirstWord) * MCBytesPerWord;
                BitAmount := Memory[LastWord]
              end
            else
              ByteAmount :=
                (Memory[PointerWord] - FirstWord) * MCBytesPerWord;
          PutBytes(FirstWord, Amount);
          if PackedFile
          then WordPointer := FirstWord
          else Memory[PointerWord] := FirstWord
        end
  end { flushbuffer };

procedure GetFile;

  var ComponentPointer: WordRange;
      CharByte: MCByte;
      GetGate: MCWordForm;

  begin
    with PascalFiles[FileIndex] do
      if TextFile
      then
        begin
          if BytePointer = LastByte
          then FillBuffer else BytePointer := BytePointer + 1;
          if EndOfFile
          then
            { buffervar := undefined }
          else
            if BytePointer = LastByte
            then Memory[BufferWord] := ord(' ')
            else
              begin
                GetGate.WValue :=
                  Memory[BytePointer div MCBytesPerWord];
                MCGetByte
                  (GetGate, BytePointer mod MCBytesPerWord, CharByte);
                Memory[BufferWord] := CharByte
              end
        end
      else
        begin
          ComponentPointer := Memory[PointerWord];
          if PackedFile
          then
            begin
              ComponentPointer := ComponentPointer - 1;
              Memory[PointerWord] := ComponentPointer;
              if ComponentPointer = Memory[LastWord]
              then
                begin
                  if WordPointer = LastWord then FillBuffer;
                  if EndOfFile
                  then
                    { buffervar := undefined }
                  else
                    begin
                      Unpackp
                        (WordPointer, LastWord + 1, ElsPerWord,
                           ElsPerWord);
                      WordPointer := WordPointer + 1;
                      Memory[PointerWord] := LastWord + ElsPerWord
                    end
                end
            end
          else
            begin
              ComponentPointer := ComponentPointer + ComponentSize;
              Memory[PointerWord] := ComponentPointer;
              if ComponentPointer = LastWord
              then
                begin
                  FillBuffer;
                  if EndOfFile
                  then
                    { buffervar := undefined }
                end
            end
        end
  end { getfile };

procedure PutFile;

  var WordAddress, ComponentPointer: WordRange;
      PutGate: MCWordForm;

  begin
    with PascalFiles[FileIndex] do
      if TextFile
      then
        begin
          if BytePointer = LastByte then FlushBuffer;
          WordAddress := BytePointer div MCBytesPerWord;
          PutGate.WValue := Memory[WordAddress];
          MCSetByte
            (PutGate, BytePointer mod MCBytesPerWord,
               Memory[BufferWord]);
          Memory[WordAddress] := PutGate.WValue;
          BytePointer := BytePointer + 1
        end
      else
        begin
          ComponentPointer := Memory[PointerWord];
          if PackedFile
          then
            begin
              ComponentPointer := ComponentPointer - 1;
              Memory[PointerWord] := ComponentPointer;
              if ComponentPointer = LastWord
              then
                begin
                  if WordPointer = LastWord then FlushBuffer;
                  Packp
                    (LastWord + 1, ElsPerWord, WordPointer,
                       ElsPerWord);
                  WordPointer := WordPointer + 1;
                  Memory[PointerWord] := LastWord + ElsPerWord
                end
            end
          else
            begin
              ComponentPointer := ComponentPointer + ComponentSize;
              Memory[PointerWord] := ComponentPointer;
              if ComponentPointer = LastWord then FlushBuffer
            end
        end
  end { putfile };

procedure InitializeFile;

  var Limit: WordRange;

  begin
    with PascalFiles[FileIndex] do
      if TextFile
      then
        begin
          LastByte := (FirstWord + BufferSize) * MCBytesPerWord;
          if Mode = Generation
          then BytePointer := FirstWord * MCBytesPerWord
          else
            begin
              BytePointer := LastByte;
              GetFile
            end
        end
      else
        begin
          Limit := FirstWord + BufferSize;
          if PackedFile
          then LastWord := Limit - ElsPerWord - 1
          else LastWord := Limit;
          if Mode = Generation
          then
            begin
              if PackedFile
              then
                begin
                  Memory[LastWord] := 0;
                  Memory[PointerWord] := LastWord + ElsPerWord;
                  WordPointer := FirstWord
                end
              else Memory[PointerWord] := FirstWord
            end
          else
            begin
              if PackedFile
              then
                begin
                  Memory[LastWord] := LastWord;
                  Memory[PointerWord] := LastWord + 1;
                  WordPointer := LastWord
                end
              else Memory[PointerWord] := LastWord - ComponentSize;
              GetFile
            end
        end
  end { initializefile };

procedure FinalizeFile;

  var ComponentPointer: WordRange;
      Padding, FieldWidth: BitRange;

  begin
    with PascalFiles[FileIndex] do
      if TextFile
      then
        begin
          if BytePointer <> FirstWord * MCBytesPerWord
          then FlushBuffer
        end
      else
        begin
          ComponentPointer := Memory[PointerWord];
          if PackedFile
          then
            begin
              if WordPointer > FirstWord then FlushBuffer;
              if ComponentPointer < (LastWord + ElsPerWord)
              then
                begin
                  Padding := 0;
                  FieldWidth := MCBitsPerWord div ElsPerWord;
                  while ComponentPointer > LastWord do
                    begin
                      Memory[ComponentPointer] := 0;
                      Padding := Padding + FieldWidth;
                      ComponentPointer := ComponentPointer - 1
                    end;
                  Memory[LastWord] := Padding;
                  Packp
                    (LastWord + 1, ElsPerWord, WordPointer,
                       ElsPerWord);
                  WordPointer := WordPointer + 1;
                  FlushBuffer
                end
            end
          else
            if ComponentPointer > FirstWord then FlushBuffer
        end
  end { finalizefile };

procedure RewriteFile;

  begin
    FileIndex := Memory[Address];
    if FileIndex = UndefinedIndex
    then
      begin
        NextFileIndex;
        if FileIndex = UndefinedIndex
        then SystemError(14);
        Memory[Address] := FileIndex;
        AllocatePascalFile
      end;
    with PascalFiles[FileIndex] do
      begin
        case Mode of
          NotSet : Acquire(BufferSize, FirstWord);
          Generation :
            begin
              FinalizeFile;
              Close(Detach)
            end;
          Inspection : Close(Detach)
        end;
        Open(ForWriting);
        Mode := Generation;
        EndOfFile := true;
        InitializeFile
      end
  end { rewritefile };

procedure ResetFile;

  begin
    FileIndex := Memory[Address];
    with PascalFiles[FileIndex] do
      begin
        case Mode of
          NotSet : Acquire(BufferSize, FirstWord);
          Generation :
            begin
              FinalizeFile;
              Close(Retain)
            end;
          Inspection : Close(Retain)
        end;
        Open(ForReading);
        Mode := Inspection;
        InitializeFile
      end
  end { resetfile };

{

 The  Model  Compiler  generates  code  to  ensure  every  file  is
 correctly  closed.   In practice this need only happen if the file
 has already been opened.  If this  is  the  case,  the  buffer  is
 flushed if the file was in generation-mode, prior to releasing the
 buffer-space.   The  descriptor  is  deallocated  by  setting  the
 defined  field  to  false,  and  a call is made to the System File
 Manager to close and detach the corresponding physical file.

                                                                     }

procedure CloseFile;

  begin
    FileIndex := Memory[Address];
    if FileIndex <> UndefinedIndex
    then
      begin
        with PascalFiles[FileIndex] do
          case Mode of
            NotSet :
              { external file never used }
              ;
            Generation :
              begin
                FinalizeFile;
                Release(BufferSize, FirstWord)
              end;
            Inspection : Release(BufferSize, FirstWord)
          end;
        DeAllocatePascalFile;
        Close(Detach)
      end
  end { closefile };

{

 The Pascal I/O Manager also provides higher-level support for  the
 instructions  implementing  text I/O.  No attempt has been made to
 provide free-standing binary to decimal conversion, and  thus  the
 formatted  input/output  of real and integer values is host depen-
 dent.  With this proviso, the following procedures provide textual
 input/output  of  real,  integer, Boolean, and character data, and
 the layout functions for writeln, and page.

                                                                     }

procedure DeriveValue(var Digits: TextString; var Result: Number);

  var f: text;
      i: StringRange;

  begin
    rewrite(f);
    with Digits do
      for i := 1 to Length do
        begin
          f^ := String[i];
          put(f)
        end;
    reset(f);
    with Result do
      case Form of
        IntForm : read(f, IntValue);
        RealForm : read(f, RealValue)
      end
  end { derivevalue };

procedure DeriveDigits(Velue: Number; var Digits: TextString);

  var f: text;

  begin
    rewrite(f);
    with Velue do
      case Form of
        IntForm : writeln(f, IntValue: 1);
        RealForm :
          case Format of
            Fixed : writeln(f, RealValue: 1: FracDigits);
            Floated : writeln(f, RealValue: TotalWidth)
          end
      end;
    reset(f);
    with Digits do
      begin
        Length := 0;
        while not eoln(f) do
          begin
            Length := Length + 1;
            String[Length] := f^;
            get(f)
          end;
        while String[Length] = ' ' do Length := Length - 1
      end
  end { derivedigits };

procedure ReadNumber(FormRequired: NumberForms);

  var Digits: TextString;
      Result: Number;

  procedure BadDigitSequence;

    begin
      case FormRequired of
        IntForm : Error(54);
        RealForm : Error(56)
      end
    end { baddigitsequence };

  procedure CopyBufferWord;

    begin
      with Digits do
        begin
          Length := Length + 1;
          with PascalFiles[FileIndex] do
            String[Length] := chr(Memory[BufferWord])
        end;
      GetFile
    end { copybufferword };

  procedure DigitSequence;

    begin
      with PascalFiles[FileIndex] do
        if not (chr(Memory[BufferWord]) in ['0'..'9'])
        then BadDigitSequence
        else
          repeat CopyBufferWord
          until not (chr(Memory[BufferWord]) in ['0'..'9'])
    end { digitsequence };

  function Signed(cvalue: MCByte): Boolean;

    begin Signed := (cvalue = ord('+')) or (cvalue = ord('-'))
    end { signed };

  begin { readnumber }
    Digits.Length := 0;
    with PascalFiles[FileIndex] do
      begin
        while (Memory[BufferWord] = ord(' ')) and not EndOfFile do
          GetFile;
        if EndOfFile
        then BadDigitSequence
        else
          begin
            if Signed(Memory[BufferWord]) then CopyBufferWord;
            DigitSequence;
            if FormRequired = RealForm
            then
              begin
                if Memory[BufferWord] = ord('.')
                then
                  begin
                    CopyBufferWord;
                    DigitSequence
                  end;
                if (Memory[BufferWord] = ord('e')) or
                   (Memory[BufferWord] = ord('E'))
                then
                  begin
                    CopyBufferWord;
                    if Signed(Memory[BufferWord]) then CopyBufferWord;
                    DigitSequence
                  end
              end;
            Result.Form := FormRequired;
            DeriveValue(Digits, Result);
            case FormRequired of
              IntForm : Push(Result.IntValue);
              RealForm : PushReal(Result.RealValue)
            end
          end
      end
  end { readnumber };

procedure ReadLine;

  begin
    with PascalFiles[FileIndex] do
      begin
        BytePointer := LastByte;
        GetFile
      end
  end { readline };

procedure PutCh(CharByte: MCByte);

  begin
    with PascalFiles[FileIndex] do
      begin
        Memory[BufferWord] := CharByte;
        PutFile
      end
  end { putch };

procedure LeadingBlanks(TotalWidth, FieldWidth: StringRange);

  begin
    while TotalWidth > FieldWidth do
      begin
        PutCh(ord(' '));
        TotalWidth := TotalWidth - 1
      end
  end { leadingblanks };

procedure WriteChar;

  var TotalWidth, CharByte: MCWord;

  begin
    Pop(TotalWidth);
    Pop(CharByte);
    LeadingBlanks(TotalWidth, 1);
    PutCh(CharByte)
  end { writechar };

procedure WriteBoolean;

  var Logical, TotalWidth: MCWord;

  begin
    Pop(TotalWidth);
    Pop(Logical);
    LeadingBlanks(TotalWidth, 5);
    case Logical of
      0 :
        begin
          PutCh(ord('f'));
          PutCh(ord('a'));
          PutCh(ord('l'));
          PutCh(ord('s'));
          PutCh(ord('e'))
        end;
      1 :
        begin
          PutCh(ord(' '));
          PutCh(ord('t'));
          PutCh(ord('r'));
          PutCh(ord('u'));
          PutCh(ord('e'))
        end
    end
  end { writeboolean };

procedure WriteString;

  var TotalWidth, StringLength, Base, Index: MCWord;
      ByteIndex: MCByteIndex;
      CharByte: MCByte;
      ByteBuffer: MCWordForm;

  begin
    Pop(TotalWidth);
    Pop(StringLength);
    if StringLength <= MCBytesPerWord
    then
      begin
        Base := StackTop;
        StackTop := StackTop - 1
      end
    else Pop(Base);
    if TotalWidth > StringLength
    then
      begin
        LeadingBlanks(TotalWidth, StringLength);
        TotalWidth := StringLength
      end;
    for Index := 0 to TotalWidth - 1 do
      begin
        ByteIndex := Index mod MCBytesPerWord;
        if ByteIndex = 0
        then
          begin
            ByteBuffer.WValue := Memory[Base];
            Base := Base + 1
          end;
        MCGetByte(ByteBuffer, ByteIndex, CharByte);
        PutCh(CharByte)
      end
  end { writestring };

procedure WriteInteger;

  var Datum: Number;
      Digits: TextString;
      FieldWidth: MCWord;
      i: StringRange;

  begin
    Pop(FieldWidth);
    with Datum do
      begin
        Form := IntForm;
        Pop(IntValue)
      end;
    DeriveDigits(Datum, Digits);
    with Digits do
      begin
        LeadingBlanks(FieldWidth, Length);
        for i := 1 to Length do PutCh(ord(String[i]))
      end
  end { writeinteger };

procedure WriteReal(FormatRequired: PrintedForms);

  var Datum: Number;
      Digits: TextString;
      FieldWidth: MCWord;
      i: StringRange;

  begin
    with Datum do
      begin
        Form := RealForm;
        Format := FormatRequired;
        if FormatRequired = Fixed then Pop(FracDigits);
        Pop(FieldWidth);
        if FormatRequired = Floated then TotalWidth := FieldWidth;
        PopReal(RealValue)
      end;
    DeriveDigits(Datum, Digits);
    with Digits do
      begin
        if FormatRequired = Fixed
        then LeadingBlanks(FieldWidth, Length);
        for i := 1 to Length do PutCh(ord(String[i]))
      end
  end { writereal };

procedure WriteLine;

  begin FlushBuffer end;

procedure PageFile;

  begin
    with PascalFiles[FileIndex] do
      if BytePointer <> FirstWord * MCBytesPerWord then WriteLine;
    PutCh(FormFeed)
  end { pagefile };

{

 The procedure InitPascIO initialises global variables used by  the
 System  File  Manager  and  the Pascal I/O Manager.  The procedure
 EndPascIO is called prior to P-machine termination to  record  the
 state of all currently opened files in word 1 of the file variable
 block.  This enables subsequent analysis by the Post  Mortem  Gen-
 erator   to report the file mode and position as well as the value
 of the buffer-variable.

                                                                     }

procedure InitPascIO;

  var Index: IndexRange;

  begin
    for Index := UndefinedIndex to MaxFileIndex do
      PascalFiles[Index].Defined := false;
    with NullTransfer do
      begin
        ByteAmount := 0;
        BitAmount := 0
      end;
    InputId := UndefinedIndex;
    OutputId := UndefinedIndex
  end { initpascio };

procedure EndPascIO;

  var Index: IndexRange;
      FileBlock: WordRange;
      OpBuffer: MCWordForm;

  begin
    for Index := UndefinedIndex to MaxFileIndex do
      with PascalFiles[Index] do
        if Defined
        then
          begin
            if TextFile
            then FileBlock := BufferWord - 2
            else FileBlock := PointerWord - 2;
            Memory[FileBlock + 3] := ord(Mode);
            OpBuffer.WValue := Memory[FileBlock + 1];
            if EndOfFile then MCSetBit(OpBuffer, EofBit);
            if TextFile
            then
              if BytePointer = LastByte
              then MCSetBit(OpBuffer, EolBit);
            Memory[FileBlock + 1] := OpBuffer.WValue
          end
  end { endpascio };


{

                          CHAPTER 8




                     Data Representations
                             and
                     Arithmetic Utilities





 The P-machine instruction-set provides  for  the  manipulation  of
 integer, real, Boolean, string, pointer, and set data according to
 the rules of the Pascal Standard.  The machine  representation  of
 each  of  these data types is given below in terms of configurable
 constants.

 (1)  Integer data occupy a single P-machine word and  hold  values
      in the range [-MCMaxint..MCMaxint].  The value of MCMaxint is
      given by 2**(MCBitsPerWord-1)-1

 (2)  Real data occupy a number of P-machine words given by MCReal-
      Size.   The  range  and  accuracy of real values available is
      dependent upon the hardware hosting the P-machine.

 (3)  Boolean data occupy a single P-machine word.  The value false
      is  denoted by the integer value 0, and the value true by the
      integer value 1.

 (4)  String data occupy one or more P-machine  words.   Each  word
      contains  up to MCBytesPerWord character bytes, and each byte
      occupies MCBitsPerByte machine bits.  The bytes in each  word
      are  numbered  from  0 at the least significant end to MCMax-
      ByteNum at the most significant end.  The  maximum  word-size
      for string data is given by the maximum byte value MCMaxByte.

 (5)  Pointer data occupy two  P-machine  words.   The  low-address
      word  contains a P-machine word-address, and the high-address
      word contains the  check key.

 (6)  Set data have a dual representation  depending  upon  whether
      the  data is held in the static or dynamic part of the stack-
      frame.  Sets are implemented as bitmaps in  which  MCWordSet-
      Bits in every word are available to represent a member of the
      set.  To enable an 'undefined' set-value to  be  constructed,
      the  most  significant  bit is never used and MCWordSetBits =
      MCBitsPerWord-1.  For a given  base  type  whose  bounds  are
      BaseMin  and  BaseMax, sets are allocated storage as if their
      actual base-type were the  subrange  0..BaseMax.  Base  types
      with negative lower bounds are not permitted.

      When a set is pushed onto the stack, it is held  in  'reduced
      format'  by truncating any high-order empty words and replac-
      ing them with  a  negative  length-word.   If  the  truncated
      length  is one, the length-word is omitted.  The empty set is
      represented by a single zero word on top of the stack.   When
      a  set  result is popped, it is expanded to its static length
      prior to the store.  The maximum word size of a set  is  lim-
      ited to the maximum byte value MCMaxByte.

 Integer and Real arithmetic are performed by host hardware.   How-
 ever  all  integer operations are pre-checked to determine whether
 overflow would  occur.   If  runtime  checking  is  requested  the
 appropriate  error is flagged.  Otherwise the result is undefined.
 The following utilities check all integer operations and  real-to-
 integer conversions against overflow.


                                                                     }

procedure IntArith(Operator: OpType; Left, Right: Operands;
                   var Result: MCWord);

  var Temp: MCWord;
      Overflows: Boolean;

  begin
    case Operator of
      Plus, Minus :
        begin
          Overflows := false;
          if Operator = Minus then Right := -Right;
          if (Left > 0) and (Right > 0)
          then Overflows := (Left > MCMaxInt - Right);
          if (Left < 0) and (Right < 0)
          then Overflows := (Left < -MCMaxInt - Right);
          if Overflows
          then
            begin
              if ChecksOn then Error(47) else Result := Undefined
            end
          else Result := Left + Right
        end;
      Mul :
        if Right = 0
        then Result := 0
        else
          if abs(Left) > MCMaxInt div abs(Right)
          then
            begin
              if ChecksOn then Error(47) else Result := Undefined
            end
          else Result := Left * Right;
      Idiv :
        if Right = 0
        then
          begin
            if ChecksOn then Error(45) else Result := Undefined
          end
        else
          if Right > 0
          then Result := Left div Right
          else Result := (-Left) div (-Right);
      Imod :
        if Right <= 0
        then
          begin
            if ChecksOn then Error(46) else Result := Undefined
          end
        else
          begin
            Temp := abs(Left) mod Right;
            if (Temp = 0) or (Left > 0)
            then Result := Temp else Result := Right - Temp
          end
    end
  end { intarith };

procedure Fix(Argument: MCReal; var Result: MCWord;
              var Overflow: Boolean);

  var Epsilon: MCReal;

  begin
    if abs(Argument) > MCMaxInt
    then Overflow := true
    else
      begin
        Result := trunc(Argument);
        Epsilon := Argument - Result;
        if Argument >= 0
        then Overflow := (Epsilon < 0) or (Epsilon > 1)
        else Overflow := (Epsilon <= -1) or (Epsilon > 0)
      end
  end { fix };

{

 String comparisons are made with  the  following  utilities  which
 cater  for  strings  whose length is not a multiple of MCBytesPer-
 Word.

                                                                     }

function StringLess(StringLength: ByteRange): Boolean;

  label 1;

  var WordSize, Base1, Base2, Index: WordRange;
      ByteIndex: MCByteIndex;
      ChByte1, ChByte2: MCByte;

  begin
    WordSize := WordsFor(StringLength);
    Base2 := StackTop - WordSize + 1;
    Base1 := Base2 - WordSize;
    for Index := 0 to StringLength - 1 do
      begin
        ByteIndex := Index mod MCBytesPerWord;
        if ByteIndex = 0
        then
          begin
            Element1.WValue := Memory[Base1];
            Base1 := Base1 + 1;
            Element2.WValue := Memory[Base2];
            Base2 := Base2 + 1
          end;
        MCGetByte(Element1, ByteIndex, ChByte1);
        MCGetByte(Element2, ByteIndex, ChByte2);
        if ChByte1 <> ChByte2
        then
          begin
            StringLess := ChByte1 < ChByte2;
            goto 1
          end
      end;
    StringLess := false;
    1:
  end { stringless };

function StringEqual(StringLength: ByteRange): Boolean;

  label 1;

  var WordSize, Base1, Base2, Index: WordRange;
      ByteIndex: MCByteIndex;
      ChByte1, ChByte2: MCByte;

  begin
    WordSize := WordsFor(StringLength);
    Base2 := StackTop - WordSize + 1;
    Base1 := Base2 - WordSize;
    for Index := 0 to StringLength - 1 do
      begin
        ByteIndex := Index mod MCBytesPerWord;
        if ByteIndex = 0
        then
          begin
            Element1.WValue := Memory[Base1];
            Base1 := Base1 + 1;
            Element2.WValue := Memory[Base2];
            Base2 := Base2 + 1
          end;
        MCGetByte(Element1, ByteIndex, ChByte1);
        MCGetByte(Element2, ByteIndex, ChByte2);
        if ChByte1 <> ChByte2
        then
          begin
            StringEqual := false;
            goto 1
          end
      end;
    StringEqual := true;
    1:
  end { stringequal };

{

 The following utilities manipulate set values according  to  their
 truncated  representations  on  the  evaluation stack.  The global
 variable SetDetails is used as a repository of information for two
 set operands, prior to executing a set-operation.

                                                                     }

procedure SetElements(Base, Stride: WordRange);

  begin
    Element2.WValue := Memory[Base];
    Element1.WValue :=
      Memory[Base + Stride]
        { recast(element1,asset) }
        { recast(element2,asset) }
  end { setelements };

procedure ExpandSet(StaticSize: WordRange);

  var CurrentSize: MCWord;

  begin
    Pop(CurrentSize);
    if CurrentSize >= 0
    then
      begin
        Push(CurrentSize);
        CurrentSize := 1
      end
    else CurrentSize := -CurrentSize;
    if StackTop + Safety(StaticSize - CurrentSize) >= HeapTop
    then SystemError(11);
    while CurrentSize < StaticSize do
      begin
        Push(0);
        CurrentSize := CurrentSize + 1
      end
  end { expandset };

procedure GetSetDetails;

  begin
    with SetDetails do
      begin
        Size1 := Memory[StackTop];
        if Size1 < 0
        then
          begin
            Size1 := -Size1;
            FreeStack(Size1)
          end
        else
          if Size1 > 0 then Size1 := 1;
        Base1 := StackTop;
        StackTop := StackTop - 1;
        Size2 := Memory[StackTop];
        if Size2 < 0
        then
          begin
            Size2 := -Size2;
            FreeStack(Size2)
          end
        else
          if Size2 > 0 then Size2 := 1;
        Base2 := StackTop;
        StackTop := StackTop - 1;
        if Size1 > Size2 then Subset := Size2 else Subset := Size1;
        if Subset = 0
        then
          if Size1 + Size2 > 0 then Subset := 1;
        Stride := Base1 - Base2
      end
  end { getsetdetails };

procedure PushSetSize(Size: WordRange);

  begin
    if Size = 0
    then Push(0)
    else
      if Size > 1 then Push(-Size)
  end { pushsetsize };


{

                          CHAPTER 9




                    Object Program Format





 The object program comprises a number of segments  each  of  which
 derives  from  a  compiled  procedure, function, or program block.
 Each segment consists of a word-aligned jump-table followed  by  a
 word-aligned  code-table.   The  boundary  between these tables is
 referred to as the 'code base' and marks the  entry-point  of  the
 block.   The  code-table  contains  a  stream of executable P-code
 bytes together with their in-line arguments.  The jump-table  con-
 tains  the  entry  point  addresses of all procedures local to the
 block, and also the code addresses of  statement  and  code-labels
 situated  within the block.  Each jump-table also contains a five-
 word header which is indexed negatively from the code  base.   The
 contents of the header are as follows:

 (1)  Word 1 contains the number of words of  storage  occupied  by
      formal parameters.

 (2)  Word 2 contains the number of words of  storage  occupied  by
      local variables and temporaries.

 (3)  Word 3 contains the number of words of  storage  occupied  by
      the lock stack.

 (4)  Word 4 contains a serial number identifying the block,  which
      is  used in analysis of the program 'corpse' by the Post Mor-
      tem Generator.

 (5)  Word 5 contains the code address of an internal label marking
      the start of finalisation code for the block.

 When a block is activated, the contents of the header are used  to
 construct the new stack frame.  The code base address is placed in
 word 2 of the new stack frame header enabling access to the  jump-
 table  throughout  the lifetime of the block.  Non-local jumps are
 therefore implemented by locating the non-local  stack  frame  via
 the  static  chain, and using word 2 of the frame header to locate
 the corresponding non-local jump-table.

 The last segment of the file is degenerate and  contains  a  jump-
 table  corresponding  to the pseudo-block surrounding the program.
 This jump-table contains the entry point addresses  of  internally
 compiled procedures to preset and postset text files, and the main
 program entry point addresses.  The last word on the code file  is
 an index into the jump-table location containing the program entry
 point.

 The P-machine loader reads the P-code file into low address memory
 and initialises the runtime stack by creating stack frames for the
 pseudo-block and the global program block.  For a P-code file con-
 taining  n segments, the P-machine memory map immediately prior to
 execution is as follows:


        Word 0 .--------------.
               |              |
               |              |
               | Jump-Table 1 |
               |              |
               |--------------|
               |              |
               |              |
               |              |
               | Code-Table 1 |
               |              |
               |              |
               |              |
               |              |
               |--------------|

                      .
                      .
                      .
                      .


               |--------------|
               |              |
               |Jump-Table n-1|
               |              |
      .------->|--------------|<----- Main program Code Base
      |   |    |              |       and Entry Point
      |   |    |              |
      |   |    |Code-Table n-1|
      |   |    |              |
      |   |    |              |
      |   |    |              |
      |   |    |              |
      |   |    |--------------|
      |   .____|  Program EP  |
      |   ---->|--------------|
      |   |    |              |
      |   |    |              |
      |   |    | Jump-Table n |
      |   |    |              |
      |   |    |              |
      |   |    |              |
      |   |    |--------------|
      |   .____|   Index EP   |<------ Last code file word
      |        |--------------|  ^
      |        |              |  |     Stack Frame Header for
      |        |    Header    |  |     Pseudo-Block
      |        |              |  v
      |        |--------------|  ^
      |        |              |  |     Stack Frame Header for
      |        |              |  |     Global Block
      |        |--------------|  |
      .________|Code-base Link|  v
               |--------------|<------ Global Frame for
               |              |        program data
               |              |
               |              |
               |              |        Global Data
               |              |            +
               |              |        Temporaries
               |              |            +
               |              |        Lock Stack
               |              |
               |              |
               |--------------|<------ Stack Pointer
                      .
                      .
                      .

                    Free
                   Memory

                      .
                      .
                      .





                                                                     }

procedure LoadProgram;

  var GlobalData, LockData, ProgramBase: WordRange;
      CodeWord: MCWordForm;
      Amount: Scalar;
      i: MCByteIndex;

  begin
    Amount := 0;
    while not eof(CodeFile) do
      begin
        Memory[Amount] := CodeFile^;
        Amount := Amount + 1;
        get(CodeFile)
      end;
    if Amount > MCMaxMemWord - MinBlock then FatalError(5);
    CodeBase := Amount - 1;
    ProgramBase := Memory[CodeBase - Memory[CodeBase]];
    with CodeWord do
      for i := 0 to MCMaxByteNum do WChars[i] := chr(haltp);

    { recast(codeword,asvalue) }
    ProgramCounter := CodeBase * MCBytesPerWord;
    Memory[CodeBase] := CodeWord.WValue;

    { set external environment frame }
    EnclosingFrame := 0;
    CurrentFrame := 0;
    StackTop := 0;
    NewFrame := CodeBase + FrameSize + 1;
    SetFrameHeader;
    CurrentFrame := NewFrame;
    EnclosingFrame := CurrentFrame;

    { set global frame }
    CodeBase := ProgramBase;
    NewFrame := NewFrame + FrameSize;
    GlobalFrame := NewFrame;
    StackTop := NewFrame - 1;
    GlobalData :=
      Memory[ProgramBase - LocalsOffset] +
      Memory[ProgramBase - ParamsOffset];
    LockData := Memory[ProgramBase - LocksOffset];
    ClaimStack(GlobalData);
    SetFrameHeader;
    ClaimStack(LockData);
    if ChecksOn and (GlobalData > 0)
    then PresetMemory(GlobalFrame, GlobalData);
    CurrentFrame := GlobalFrame;
    Overflowed := false;
    ProgramCounter := ProgramBase * MCBytesPerWord
  end { loadprogram };


{

                          CHAPTER 10




                      Program Execution





 The procedure RunProgram performs instruction fetch and  execution
 until the program terminates or halts in error.  Every instruction
 is denoted by a P-code  byte-value  that  additionally  serves  to
 identify  the number and type of arguments.  An instruction may be
 associated with more than one P-code value  if  variation  in  the
 type  of  arguments  is permitted.  The eight distinct instruction
 formats are:

                 P-code
                 P-code byte
                 P-code word
                 P-code byte byte
                 P-code word byte
                 P-code byte bytestring
                 P-code byte wordstring
                 P-code word wordstring

 To allow for more than 256 P-code values, instructions are  parti-
 tioned  into  two  classes.   There  are  seven  groups of Class 1
 instructions dealing with in-line constants, memory-access, opera-
 tions,  jumps,  procedure  entry  and exit, memory management, and
 run-time check support.  Class 2 instructions deal with the  stan-
 dard  procedures and functions, and the mapping of program parame-
 ters,  and are identified by a prefixed 'escape' byte.


                                                                     }

procedure RunProgram;

  var OpCode: MCByte;
      Finished: Boolean;
      Name: Alfa;

      { processor scratchpad }

      Datum: MCWord;
      Base, Base1, Base2, Start: MCWord;
      Size, Size1, Size2, ReducedSize, CheckSize: MCWord;
      ByteAddress, Offset, FrameOffset, IndexValue: MCWord;
      CAPBpBlock: MCWord;
      ThisFrame, NextFrame, LabelledFrame, HeapAddress: WordRange;
      LabelEntry: MCWord;
      Parameters, StackSpace, LockSpace: MCWord;
      RightMost, ElementsPerWord: MCWord;
      Upper, Lower: MCWord;
      FirstBit, LastBit, EmptyWords: MCWord;
      Logical1, Logical2: MCWordSet;
      SelectorValue, SelectedValue: MCWord;
      ActiveBits, LockBits, LockWord: MCWordForm;
      TrapCode: MCWord;
      i: MCWord;
      OpBuffer, IpBuffer: MCWordForm;
      Bits: MCWord;
      Bit: MCBit;
      Byte: MCByte;

{

 The functions NextByte and NextWord fetch the values of  byte  and
 word  arguments  respectively.   Word-string  arguments  are  word
 aligned and are fetched from the next word  boundary  computed  by
 the function WordBoundary.

                                                                     }

  function NextByte: MCByte;

    var CodeWord: MCWordForm;

    begin
      CodeWord.WValue := Memory[ProgramCounter div MCBytesPerWord];

      { recast(codeword,asbytes) }
      NextByte := ord(CodeWord.WChars[ProgramCounter mod MCBytesPerWord]);
      ProgramCounter := ProgramCounter + 1
    end { nextbyte };

  function NextWord: MCWord;

    var Buffer: MCWordForm;
        ByteIndex: MCByteIndex;

    begin
      for ByteIndex := 0 to MCMaxByteNum do
        Buffer.WChars[ByteIndex] := chr(NextByte);

      { recast(buffer,asvalue) }
      NextWord := Buffer.WValue
    end { nextword };

  function WordBoundary: WordRange;

    var Address: ByteRange;

    begin
      Address := ProgramCounter;
      while (Address mod MCBytesPerWord) <> 0 do
        Address := Address + 1;
      WordBoundary := Address div MCBytesPerWord
    end { wordboundary };

  begin
    { variant activebits := [2,5,8,11,14,17,20,23,26,29] }
    ActiveBits.WValue := 0;
    for i := 0 to MaxVariantLevel do MCSetBit(ActiveBits, 3 * i + 2);

    { variant lockbits := [0,3,6,9,12,15,18,21,24,27] }
    LockBits.WValue := 0;
    for i := 0 to MaxVariantLevel do MCSetBit(LockBits, 3 * i);

{

 Program execution commences at this point.   The  following  para-
 graphs  contain  brief  descriptions  of each instruction together
 with argument types and explanatory notes where  appropriate.  The
 notation TOS is as shorthand for 'top of stack'.

                                                                     }

    Finished := false;
    repeat
      OpCode := NextByte;
      case OpCode of

{

 10.1  Constant-Instructions

 The following instructions fetch and  load  in-line  constants  or
 their addresses onto the top of the stack.


         Instructions        Arguments                Notes

         ShortConstant                                  1
         Constant             byte|word                 2
         ConstMultiple       (byte|word) wordstring     3
         ConstRefMultiple    (byte|word) wordstring     4
         ConstSet            (byte|word) wordstring     5


 (1)  The first 31 P-code values are reserved to  denote  constants
      in the range 0..31.

 (2)  The argument specifies the constant value to be  pushed  onto
      the stack.

 (3)  The first argument specifies the  length  of  the  multi-word
      constant to be pushed onto the stack.  The second argument is
      the word-aligned multi-word constant.

 (4)  The code address of the multi-word constant  is  pushed  onto
      the stack.

 (5)  The set constant is pushed onto the stack in reduced format.


                                                                     }

        cons0, cons1, cons2, cons3, cons4, cons5, cons6, cons7,
        cons8, cons9, cons10, cons11, cons12, cons13, cons14, cons15,
        cons16, cons17, cons18, cons19, cons20, cons21, cons22,
        cons23, cons24, cons25, cons26, cons27, cons28, cons29,
        cons30, cons31 :
          { shortconstant         }
          begin Push(OpCode) end;

        { in-line constants }

        consb :
          { constant(b)           }
          begin Push(NextByte) end;
        consw :
          { constant(w)           }
          begin Push(NextWord) end;
        conmb :
          { constmultiple(b)      }
          begin
            Size := NextByte;
            CodeBase := WordBoundary;
            LoadWords(CodeBase, Size);
            ProgramCounter := (CodeBase + Size) * MCBytesPerWord
          end;
        conmw :
          { constmultiple(w)      }
          begin
            Size := NextWord;
            CodeBase := WordBoundary;
            LoadWords(CodeBase, Size);
            ProgramCounter := (CodeBase + Size) * MCBytesPerWord
          end;
        crefb :
          { constrefmultiple(b)   }
          begin
            Size := NextByte;
            CodeBase := WordBoundary;
            Push(CodeBase);
            ProgramCounter := (CodeBase + Size) * MCBytesPerWord
          end;
        crefw :
          { constrefmultiple(w)   }
          begin
            Size := NextWord;
            CodeBase := WordBoundary;
            Push(CodeBase);
            ProgramCounter := (CodeBase + Size) * MCBytesPerWord
          end;
        csetb :
          { constset(b)           }
          begin
            Size := NextByte;
            CodeBase := WordBoundary;
            LoadWords(CodeBase, Size);
            Push(-Size);
            ProgramCounter := (CodeBase + Size) * MCBytesPerWord;

          end;
        csetw :
          { constset(w)           }
          begin
            Size := NextWord;
            CodeBase := WordBoundary;
            LoadWords(CodeBase, Size);
            Push(-Size);
            ProgramCounter := (CodeBase + Size) * MCBytesPerWord
          end;

{

 10.2  Memory Access

 The following instructions provide access to memory for the  load-
 ing  of operands and storing of results.  The group is sub-divided
 into word access in the local, global, enclosing and  intermediate
 stack-frames,  word  access  by indexing and indirection, partword
 access, partword indexing, byte access, byte indexing, and  multi-
 word access.

 10.2.1 Direct Word Access


         Instructions          Arguments     Notes

         ShortLoadLocal                        1
         LoadLocal             byte|word       2
         LoadRefLocal          byte|word       3
         ShortStoreLocal                       1
         StoreLocal            byte|word       2
         LoadGlobal            byte|word       2
         LoadRefGlobal         byte|word       3
         StoreGlobal           byte|word       2
         LoadEnclosing         byte|word       2
         LoadRefEnclosing      byte|word       3
         StoreEnclosing        byte|word       2
         LoadIntermediate      byte|word       2
         LoadRefIntermediate   byte|word       3
         StoreIntermediate     byte|word       2


 (1)  These instructions perform  optimised  access  to  the  local
      stack-frame.  The  word-offset  from  the stack-frame base is
      implicit in the P-code value.

 (2)  These instructions perform access to  the  local,  enclosing,
      intermediate,  and  global  stack  frames.   In each case the
      argument specifies a word offset  in  the  appropriate  stack
      frame.

 (3)  These instructions push a word  address  in  the  appropriate
      stack-frame onto the stack.  In each case the argument speci-
      fies the word-offset in the frame.

                                                                     }

        lodl0, lodl1, lodl2, lodl3, lodl4, lodl5, lodl6, lodl7 :
          { shortloadlocal        }
          begin Push(Memory[CurrentFrame + OpCode - lodl0]) end;
        lodlb :
          { loadlocal(b)          }
          begin Push(Memory[CurrentFrame + NextByte]) end;
        lodlw :
          { loadlocal(w)          }
          begin Push(Memory[CurrentFrame + NextWord]) end;
        lrf0, lrf1, lrf2, lrf3, lrf4, lrf5, lrf6, lrf7 :
          { loadreflocal(0..7)    }
          begin Push(CurrentFrame + OpCode - lrf0) end;
        lrflb :
          { loadreflocal(b)       }
          begin Push(CurrentFrame + NextByte) end;
        lrflw :
          { loadreflocal(w)       }
          begin Push(CurrentFrame + NextWord) end;
        stol0, stol1, stol2, stol3, stol4, stol5, stol6, stol7 :
          { shortstorelocal       }
          begin Pop(Memory[CurrentFrame + OpCode - stol0]) end;
        stolb :
          { storelocal(b)         }
          begin Pop(Memory[CurrentFrame + NextByte]) end;
        stolw :
          { storelocal(w)         }
          begin Pop(Memory[CurrentFrame + NextWord]) end;

        { global block access }

        lodgb :
          { loadglobal(b)         }
          begin Push(Memory[GlobalFrame + NextByte]) end;
        lodgw :
          { loadglobal(w)         }
          begin Push(Memory[GlobalFrame + NextWord]) end;
        lrfgb :
          { loadrefglobal(b)      }
          begin Push(GlobalFrame + NextByte) end;
        lrfgw :
          { loadrefglobal(w)      }
          begin Push(GlobalFrame + NextWord) end;
        stogb :
          { storeglobal(b)        }
          begin Pop(Memory[GlobalFrame + NextByte]) end;
        stogw :
          { storeglobal(w)        }
          begin Pop(Memory[GlobalFrame + NextWord]) end;

        { enclosing block access }

        lodeb :
          { loadenclosing(b)      }
          begin
            Push(Memory[Memory[CurrentFrame - SLOffset] + NextByte])
          end;
        lodew :
          { loadenclosing(w)      }
          begin
            Push(Memory[Memory[CurrentFrame - SLOffset] + NextWord])
          end;
        lrfeb :
          { loadrefenclosing(b)   }
          begin Push(Memory[CurrentFrame - SLOffset] + NextByte) end;
        lrfew :
          { loadrefenclosing(w)   }
          begin Push(Memory[CurrentFrame - SLOffset] + NextWord) end;
        stoeb :
          { storeenclosing(b)     }
          begin
            Pop(Memory[Memory[CurrentFrame - SLOffset] + NextByte])
          end;
        stoew :
          { storeenclosing(w)     }
          begin
            Pop(Memory[Memory[CurrentFrame - SLOffset] + NextWord])
          end;

        { intermediate block access }

        lodib :
          { loadintermediate(b)   }
          begin
            Offset := NextByte;
            FrameOffset := NextByte;
            Push(Memory[OuterFrame(FrameOffset) + Offset])
          end;
        lodiw :
          { loadintermediate(w)   }
          begin
            Offset := NextWord;
            FrameOffset := NextByte;
            Push(Memory[OuterFrame(FrameOffset) + Offset])
          end;
        lrfib :
          { loadrefintermediate(b) }
          begin
            Offset := NextByte;
            FrameOffset := NextByte;
            Push(OuterFrame(FrameOffset) + Offset)
          end;
        lrfiw :
          { loadrefintermediate(w) }
          begin
            Offset := NextWord;
            FrameOffset := NextByte;
            Push(OuterFrame(FrameOffset) + Offset)
          end;
        stoib :
          { storeintermediate(b)   }
          begin
            Offset := NextByte;
            FrameOffset := NextByte;
            Pop(Memory[OuterFrame(FrameOffset) + Offset])
          end;
        stoiw :
          { storeintermediate(w)  }
          begin
            Offset := NextWord;
            FrameOffset := NextByte;
            Pop(Memory[OuterFrame(FrameOffset) + Offset])
          end;

{

 10.2.2  Indirect Word Access


         Instructions    Arguments     Notes

         LoadIndirect                    1
         ShortIndex                      2
         Index           byte|word       3
         IndexRef        byte|word       4
         IndexCAP                        5
         StoreIndirect                   1
         AdjustPlus      byte|word       6
         AdjustP1                        6
         AdjustMinus     byte|word       6
         AdjustM1                        6
         AddAddress                      6


 (1)  These instructions perform indirect load and  store  via  the
      word address on the top of the stack.

 (2)  These instructions perform indexed load from the base-address
      on  top  of  the stack.  The offset is implicit in the P-code
      byte value.

 (3)  These instructions perform an indexed  load  from  the  base-
      address on top of the stack using the argument as offset.

 (4)  These instructions perform indexed load using an index  value
      and  base  address on top of the stack and the argument as an
      index scale factor.

 (5)  This instruction is used to  index-load  a  conformant  array
      element  using the bound pair block to provide scaling infor-
      mation.

 (6)  These instructions are used to modify an address value on top
      of  the stack.  The adjustment is implicit if it is +1 or -1,
      and specified by the argument value otherwise.

                                                                     }

        lodin :
          { loadindirect          }
          begin
            Pop(Address);
            Push(Memory[Address])
          end;
        indx1, indx2, indx3, indx4, indx5, indx6, indx7 :
          { shortindex            }
          begin
            Pop(Base);

            { recastbyte(opcode,asbyte) }
            Push(Memory[Base + OpCode - lodin])
          end;
        indxb :
          { index(b)              }
          begin
            Pop(Base);
            Push(Memory[Base + NextByte])
          end;
        indxw :
          { index(w)              }
          begin
            Pop(Base);
            Push(Memory[Base + NextWord])
          end;
        ixrfb :
          { indexref(b)           }
          begin
            Pop(IndexValue);
            Pop(Base);
            Push(Base + IndexValue * NextByte)
          end;
        ixrfw :
          { indexref(w)           }
          begin
            Pop(IndexValue);
            Pop(Base);
            Push(Base + IndexValue * NextWord)
          end;
        ixcap :
          { indexcap              }
          begin
            Pop(IndexValue);
            Pop(CAPBpBlock);
            Pop(Base);
            Push
              (Base +
               (IndexValue - Memory[CAPBpBlock]) *
               Memory[CAPBpBlock + 2])
          end;
        stoin :
          { storeindirect         }
          begin
            Pop(Address);
            Pop(Datum);
            Memory[Address] := Datum
          end;
        adjpb :
          { adjustplus(b)         }
          begin
            Pop(Address);
            Push(Address + NextByte)
          end;
        adjpw :
          { adjustplus(w)         }
          begin
            Pop(Address);
            Push(Address + NextWord)
          end;
        adjp1 :
          { adjustp1              }
          begin
            Pop(Address);
            Push(Address + 1)
          end;
        adjmb :
          { adjustminus(b)        }
          begin
            Pop(Address);
            Push(Address - NextByte)
          end;
        adjmw :
          { adjustminus(w)        }
          begin
            Pop(Address);
            Push(Address - NextWord)
          end;
        adjm1 :
          { adjustm1              }
          begin
            Pop(Address);
            Push(Address - 1)
          end;
        adadr :
          { addaddress            }
          begin
            Pop(Offset);
            Pop(Address);
            Push(Address + Offset)
          end;

{

 10.2.3  Part-Word-Access

 The following instructions access part-words by means of a stacked
 part-word address that takes the form:

         TOS   : Right-most bit position of part-word field
         TOS-1 : Width of part-word field
         TOS-2 : Address of word containing the field


         Instructions      Arguments     Notes

         LoadBit             byte          1
         LoadPacked                        2
         StoreBit                          1
         StorePacked                       2
         IndexPackedRef      byte          3
         IndexPackedCAP                    6
         IndexSubWord        byte          4
         AdjustPackedRef                   5


 (1)  These instructions access single bit fields in the word whose
      address is on top of the stack.  The argument specifies a bit
      position measured from the least-significant end of the word.


 (2)  These instructions access part-word fields using  the  three-
      word part-word address on top of the stack.

 (3)  This instruction produces a part-word  address  corresponding
      to  an  indexed element of a packed array.  The byte argument
      specifies the number of elements per word.

 (4)  This instruction produces a  part-word  address  by  indexing
      from  a  part-word base address.  The byte argument specifies
      the number of elements per word.

 (5)  This instruction modifies a part-word address by altering the
      field-width and field bit-offset.

 (6)  This instruction normalises the index value for a packed con-
      formant array.

                                                                     }

        lodbt :
          { loadbit               }
          begin
            Pop(Base);
            IpBuffer.WValue := Memory[Base];
            MCGetBit(IpBuffer, NextByte, Bit);
            Push(Bit)
          end;
        lodpk :
          { loadpacked            }
          begin
            Pop(RightMost);
            Pop(Size);
            Pop(Base);
            IpBuffer.WValue := Memory[Base];
            OpBuffer.WValue := 0;
            for i := 0 to Size - 1 do
              begin
                MCGetBit(IpBuffer, RightMost + i, Bit);
                if Bit <> 0 then MCSetBit(OpBuffer, i)
              end;
            Push(OpBuffer.WValue)
          end;
        stobt :
          { storebit              }
          begin
            Pop(Base);
            OpBuffer.WValue := Memory[Base];
            Pop(Datum);
            if Datum = 0
            then MCClearBit(OpBuffer, NextByte)
            else MCSetBit(OpBuffer, NextByte);
            Memory[Base] := OpBuffer.WValue
          end;
        stopk :
          { storepacked           }
          begin
            Pop(RightMost);
            Pop(Size);
            Pop(Base);
            Pop(IpBuffer.WValue);
            OpBuffer.WValue := Memory[Base];
            for i := 0 to Size - 1 do
              begin
                MCGetBit(IpBuffer, i, Bit);
                if Bit = 0
                then MCClearBit(OpBuffer, RightMost + i)
                else MCSetBit(OpBuffer, RightMost + i)
              end;
            Memory[Base] := OpBuffer.WValue
          end;
        ixpkb :
          { indexpackedref        }
          begin
            ElementsPerWord := NextByte;
            Pop(IndexValue);
            Pop(Base);
            Size := MCBitsPerWord div ElementsPerWord;
            Push(Base + IndexValue div ElementsPerWord);
            Push(Size);
            Push((IndexValue mod ElementsPerWord) * Size)
          end;
        ixpcap :
          { indexpackedcap        }
          begin
            Pop(IndexValue);
            Pop(CAPBpBlock);
            Push(IndexValue - Memory[CAPBpBlock])
          end;
        ixswb :
          { indexsubword          }
          begin
            ElementsPerWord := NextByte;
            Pop(IndexValue);
            Pop(RightMost);
            Pop(Size);
            Size := MCBitsPerWord div ElementsPerWord;
            Push(Size);
            Push(RightMost + Size * IndexValue)
          end;
        adjpk :
          { adjustpackedref       }
          begin
            Pop(Base1);
            Pop(Size1);
            Pop(Base2);
            Pop(Size2);
            Push(Size1);
            Push(Base1 + Base2)
          end;

{

 10.2.4  Byte Access

 The following instructions cater for  character  byte  access  and
 indexing.


         Instructions         Arguments     Notes

         LoadByte                             1
         StoreByte                            1
         LoadByteReference                    2
         IndexByteReference                   2


 (1)  These instructions perform byte access using the byte address
      on top of the stack.

 (2)  These instructions generate a byte address from the word base
      address on top of the stack.

                                                                     }

        lodby :
          { loadbyte              }
          begin
            Pop(ByteAddress);
            Element.WValue := Memory[ByteAddress div MCBytesPerWord];
            MCGetByte(Element, ByteAddress mod MCBytesPerWord, Byte);
            Push(Byte)
          end;
        stoby :
          { storebyte             }
          begin
            Pop(ByteAddress);
            Pop(Datum);
            Element.WValue := Memory[ByteAddress div MCBytesPerWord];
            MCSetByte(Element, ByteAddress mod MCBytesPerWord, Datum);
            Memory[ByteAddress div MCBytesPerWord] := Element.WValue
          end;
        lbyrf :
          { loadbytereference     }
          begin
            Pop(Address);
            Push(Address * MCBytesPerWord)
          end;
        ibyrf :
          { indexbytereference    }
          begin
            Pop(IndexValue);
            Pop(Base);
            Push(IndexValue + Base * MCBytesPerWord)
          end;

{

 10.2.5  Multiple Word Access

 The following instructions manipulate multi-word fields.


         Instructions     Arguments     Notes

         LoadMultiple     byte|word
         StoreMultiple    byte|word
         StoreRepeated    byte|word       1
         Move             byte|word
         SizeCAP                          2
         MoveCAP                          3
         DuplicateCAP                     4
         ReleaseCAP                       5


 (1)  This instruction repeatedly stores a value into an  array  of
      memory words.  The argument determines the size of the array.

 (2)  This instruction computes the word size of a conformant array
      parameter.

 (3)  This instruction moves the contents of one  conformant  array
      parameter into another.

 (4)  This instruction creates a copy of a value  conformant  array
      on the heap.

 (5)  This instruction releases the copy of  the  value  conformant
      array from the heap.

                                                                     }

        lodmb :
          { loadmultiple(b)       }
          begin
            Pop(Base);
            LoadWords(Base, NextByte)
          end;
        lodmw :
          { loadmultiple(w)       }
          begin
            Pop(Base);
            LoadWords(Base, NextWord)
          end;
        stomb :
          { storemultiple(b)      }
          begin
            Pop(Address);
            StoreWords(Address, NextByte)
          end;
        stomw :
          { storemultiple(w)      }
          begin
            Pop(Address);
            StoreWords(Address, NextWord)
          end;
        storb :
          { storerepeated(b)      }
          begin
            Pop(Base);
            Pop(Datum);
            Spray(Datum, Base, NextByte)
          end;
        storw :
          { storerepeated(w)     }
          begin
            Pop(Base);
            Pop(Datum);
            Spray(Datum, Base, NextWord)
          end;
        moveb :
          { move                  }
          begin
            Pop(Base1);
            Pop(Base2);
            MoveWords(Base2, Base1, NextByte)
          end;
        movew :
          { move(w)               }
          begin
            Pop(Base1);
            Pop(Base2);
            MoveWords(Base2, Base1, NextWord)
          end;
        szcap :
          { sizecap               }
          begin
            Pop(CAPBpBlock);
            Push
              ((Memory[CAPBpBlock + 1] - Memory[CAPBpBlock] + 1) *
               Memory[CAPBpBlock + 2])
          end;
        mvcap :
          { movecap               }
          begin
            Pop(Base1);
            Pop(Size);
            Pop(Base2);
            MoveWords(Base2, Base1, Size)
          end;
        dpcap :
          { duplicatecap          }
          begin
            Pop(Size);
            Pop(Address);
            Acquire(Size, HeapAddress);
            MoveWords(Memory[Address], HeapAddress, Size);
            Memory[Address] := HeapAddress
          end;
        rlcap :
          { releasecap          }
          begin
            Pop(Size);
            Pop(Address);
            Release(Size, Memory[Address])
          end;

{

 10.3  Operations

 The instructions in this major group perform  operations  for  the
 manipulation  of  integer, real, Boolean, pointer, string, and set
 data.  In each case the operand(s) are on top of the stack.

 10.3.1  Boolean Operations


         Instructions     Arguments     Notes

         AndOperation
         NotOperation

                                                                     }

        andop :
          { andoperation          }
          begin
            PopLogical(Logical1);
            PopLogical(Logical2);
            PushLogical(Logical1 * Logical2)
          end;
        notop :
          { notoperation          }
          begin
            Pop(Datum);
            if Datum = 0 then Push(1) else Push(0)
          end;

{

 10.3.2 Integer Operations

 The following instructions perform operations on integer  operands
 at TOS and TOS-1.  An integer result is pushed back on the stack.

         Instructions     Arguments     Notes

         Increment        byte|word       1,2
         Incr1                            2
         Decrement        byte|word       1,2
         Dec1                             2
         AbsInteger                       1
         NegateInteger                    1
         SquareInteger                    1
         AddInteger                       1
         SubInteger                       1
         MulInteger                       1
         DivInteger                       1
         ModInteger                       1


 (1)  The operands are checked against the operation requested.  If
      overflow is predicted and run-time checks have been selected,
      the error is flagged and execution terminates.   If  overflow
      is predicted and checks have not been selected, the result is
      set to the 'undefined' word value.

 (2)  The argument specifies the increment or decrement.

                                                                     }

        incrb :
          { increment(b)          }
          begin
            Pop(Operand2);
            IntArith(Plus, Operand2, NextByte, Result);
            Push(Result)
          end;
        incrw :
          { increment(w)          }
          begin
            Pop(Operand2);
            IntArith(Plus, Operand2, NextWord, Result);
            Push(Result)
          end;
        incr1 :
          { incr1                 }
          begin
            Pop(Operand);
            IntArith(Plus, Operand, 1, Result);
            Push(Result)
          end;
        decrb :
          { decrement(b)          }
          begin
            Pop(Operand2);
            IntArith(Minus, Operand2, NextByte, Result);
            Push(Result)
          end;
        decrw :
          { decrement(w)          }
          begin
            Pop(Operand2);
            IntArith(Minus, Operand2, NextWord, Result);
            Push(Result)
          end;
        decr1 :
          { dec1                  }
          begin
            Pop(Operand);
            IntArith(Minus, Operand, 1, Result);
            Push(Result)
          end;
        absint :
          { absinteger            }
          begin
            Pop(Operand);
            if Operand < 0 then Push(-Operand) else Push(Operand)
          end;
        negint :
          { negateinteger         }
          begin
            Pop(Operand);
            Push(-Operand)
          end;
        sqrint :
          { squareinteger         }
          begin
            Pop(Operand);
            IntArith(Mul, Operand, Operand, Result);
            Push(Result)
          end;
        addint :
          { addinteger            }
          begin
            PopOperands;
            IntArith(Plus, Operand2, Operand1, Result);
            Push(Result)
          end;
        subint :
          { subinteger            }
          begin
            PopOperands;
            IntArith(Minus, Operand2, Operand1, Result);
            Push(Result)
          end;
        mulint :
          { mulinteger            }
          begin
            PopOperands;
            IntArith(Mul, Operand2, Operand1, Result);
            Push(Result)
          end;
        divint :
          { divinteger            }
          begin
            PopOperands;
            IntArith(Idiv, Operand2, Operand1, Result);
            Push(Result)
          end;
        modint :
          { modinteger            }
          begin
            PopOperands;
            IntArith(Imod, Operand2, Operand1, Result);
            Push(Result)
          end;

{

 10.3.3 Integer Comparisons

 The  following  instructions  perform  comparisons  between  valid
 integer  values.   In  each  case, the operands are popped off the
 stack and replaced by a Boolean result value.


         Instructions     Arguments     Notes

         OddInteger
         TestIEqual
         TestIUnequal
         TestILess
         TestILessOrEqual

                                                                     }

        oddint :
          { oddinteger            }
          begin
            Pop(Operand);
            Push(abs(Operand) mod 2)
          end;
        eqint :
          { testiequal            }
          begin
            PopOperands;
            Push(ord(Operand2 = Operand1))
          end;
        neint :
          { testiunequal          }
          begin
            PopOperands;
            Push(ord(Operand2 <> Operand1))
          end;
        ltint :
          { testiless             }
          begin
            PopOperands;
            Push(ord(Operand2 < Operand1))
          end;
        leint :
          { testilessorequal      }
          begin
            PopOperands;
            Push(ord(Operand2 <= Operand1))
          end;

{

 10.3.4  Scalar Value Checks

 The following instructions check if the value on top of the  stack
 is 'undefined'.


         Instruction        Arguments           Notes

         CheckTopValue         byte              1,2
         CheckTopDefined       byte              1,3
         CheckCAP              byte              4


 (1)  The byte argument specifies an error code.

 (2)  This instruction  checks  the  value  at  TOS-1  against  the
      check-value at TOS.  The instruction is used to perform unde-
      fined checking for packed-fields where the check-value  is  a
      suitably chosen integer.

 (3)  This instruction is used to check TOS against the representa-
      tion of the machine 'undefined' word value.

 (4)  The elements of a conformant array whose word address  is  at
      TOS-1  are checked against the machine 'undefined' value. The
      size of the array is at TOS.

                                                                     }

        chktv :
          { checktopvalue         }
          begin
            Pop(Datum);
            TrapCode := NextByte;
            if Datum = Memory[StackTop] then Error(TrapCode)
          end;
        chktd :
          { checktopdefined       }
          begin
            TrapCode := NextByte;
            if MemoryUndefined(StackTop, 1) then Error(TrapCode)
          end;
        chkcap :
          { checkcap              }
          begin
            TrapCode := NextByte;
            if MemoryUndefined
                 (Memory[StackTop - 1], Memory[StackTop])
            then
              Error(TrapCode)
          end;

{

 10.3.5  Range-Checks

 The following instructions perform range-checking operations.   If
 the range is violated, an error is flagged, otherwise, the checked
 value is left at TOS.


                Instruction     Arguments     Notes

                CheckLimits        byte         1
                CheckLower         byte         2
                CheckUpper         byte         3
                CheckForLower      byte         4
                CheckForUpper      byte         4


 (1)  The value at TOS-2 is checked against upper and lower  limits
      at TOS and TOS-1.

 (2)  The value at TOS-1 is checked against a lower limit at TOS.

 (3)  The value at TOS-1 is checked against an upper limit at TOS.

 (4)  TOS contains the upper or lower bound of a  for-loop  control
      variable,  and  TOS-1  contains a computed limiting value for
      this variable. The value is checked against the bound and  an
      appropriate error is flagged if necessary.

                                                                     }

        chklm :
          { checklimits            }
          begin
            Pop(Upper);
            Pop(Lower);
            Datum := Memory[StackTop];
            TrapCode := NextByte;
            if (Datum < Lower) or (Datum > Upper) then Error(TrapCode)
          end;
        chklw :
          { checklower            }
          begin
            Pop(Lower);
            TrapCode := NextByte;
            if Memory[StackTop] < Lower then Error(TrapCode)
          end;
        chkup :
          { checkupper }
          begin
            Pop(Upper);
            TrapCode := NextByte;
            if Memory[StackTop] > Upper then Error(TrapCode)
          end;
        chkflw :
          { checkforlow }
          begin
            Pop(Lower);
            Pop(Datum);
            TrapCode := NextByte;
            if Datum < Lower then Error(TrapCode)
          end;
        chkfup :
          { checkforhigh }
          begin
            Pop(Upper);
            Pop(Datum);
            TrapCode := NextByte;
            if Datum > Upper then Error(TrapCode)
          end;
{



 10.3.6  Real Operations

 The following instructions perform arithmetic on real-operands  at
 TOS and TOS-1. A real result is pushed back on the stack.


                Instruction     Arguments     Notes

                Float
                AbsReal
                NegateReal
                SquareReal                      1
                AddReal
                SubtractReal
                MultiplyReal
                DivideReal                      2
                TruncateReal                    3
                RoundReal                       3


 (1)  If checks are requested the operand is  checked  against  the
      configurable  constant MCMaxRealRoot, and an error flagged if
      this value is exceeded.  Since the range of  reals  is  host-
      hardware  dependent,  MCMaxRealRoot is set in this version to
      the largest real whose square exceeds MCMaxint.

 (2)  If checks are requested an error is flagged if the divisor is
      zero.

 (3)  If checks are requested an error is flagged if the result  is
      not a valid integer.

                                                                     }

        float :
          { floatinteger          }
          begin
            Pop(Operand);
            PushReal(Operand)
          end;
        absrl :
          { absreal               }
          begin
            PopReal(RealOperand);
            if RealOperand < 0
            then PushReal(-RealOperand) else PushReal(RealOperand)
          end;
        negrl :
          { negatereal            }
          begin
            PopReal(RealOperand);
            PushReal(-RealOperand)
          end;
        sqrl :
          { squarereal            }
          begin
            PopReal(RealOperand);
            if abs(RealOperand) > MCMaxRealRoot
            then
              begin
                if ChecksOn
                then Error(32) else RealResult := Undefined
              end
            else RealResult := RealOperand * RealOperand;
            PushReal(RealResult)
          end;
        addrl :
          { addreal               }
          begin
            PopRealOperands;
            PushReal(RealOp2 + RealOp1)
          end;
        subrl :
          { subtractreal          }
          begin
            PopRealOperands;
            PushReal(RealOp2 - RealOp1)
          end;
        mulrl :
          { multiplyreal          }
          begin
            PopRealOperands;
            PushReal(RealOp2 * RealOp1)
          end;
        divrl :
          { dividereal            }
          begin
            PopRealOperands;
            if RealOp1 = 0
            then
              begin
                if ChecksOn
                then Error(44) else RealResult := Undefined
              end
            else RealResult := RealOp2 / RealOp1;
            PushReal(RealResult)
          end;
        trurl :
          { truncatereal          }
          begin
            PopReal(RealOperand);
            Fix(RealOperand, Result, Overflowed);
            if Overflowed
            then
              begin
                if ChecksOn then Error(35) else Result := Undefined
              end;
            Push(Result)
          end;
        rndrl :
          { roundreal             }
          begin
            PopReal(RealOperand);
            if RealOperand >= 0
            then Fix(RealOperand + 0.5, Result, Overflowed)
            else Fix(RealOperand - 0.5, Result, Overflowed);
            if Overflowed
            then
              begin
                if ChecksOn then Error(36) else Result := Undefined
              end;
            Push(Result)
          end;

{

 10.3.7  Real Comparisons

 The  following  instructions  perform  comparisons  between   real
 operands at TOS and TOS-1.  A Boolean result is pushed back on the
 stack.


               Instruction     Arguments     Notes

               TestREqual
               TestRUnequal
               TestRLess
               TestRLtOrEqual

                                                                     }

        eqrl :
          { testrequal            }
          begin
            PopRealOperands;
            Push(ord(RealOp2 = RealOp1))
          end;
        nerl :
          { testrunequal          }
          begin
            PopRealOperands;
            Push(ord(RealOp2 <> RealOp1))
          end;
        ltrl :
          { testrless            }
          begin
            PopRealOperands;
            Push(ord(RealOp2 < RealOp1))
          end;
        lerl :
          { testrltorequal        }
          begin
            PopRealOperands;
            Push(ord(RealOp2 <= RealOp1))
          end;

{

 10.3.8  Pointer Operations

 The following instructions perform operations on pointer operands.

                Instruction     Arguments      Notes

                LoadPointer                      1
                CheckPointer                     3
                TestPEqual                       2
                TestPUnequal                     2


 (1)  This instruction performs pointer dereference.  TOS  contains
      the  address  of the pointer.  If checks are requested errors
      are flagged  if  the  pointer  contents  are  'undefined'  or
      represent  the  value 'nil', or if key mismatch indicates the
      pointer is 'dangling'.  The result loaded is pushed  back  on
      the stack.

 (2)  These  instructions  perform  comparisons  of  the   pointers
      located  at TOS-1 and TOS-3.  A Boolean result is pushed back
      on the stack.

 (3)  This instruction checks a pointer value and flags an error if
      the pointer is found to be undefined or 'dangling'.

                                                                     }

        lodptr :
          { loadpointer           }
          begin
            Pop(Address);
            if MemoryUndefined(Address, PointerSize) then Error(4);
            Base := Memory[Address];
            if Base = 0 then Error(3);
            if Memory[Address + 1] <> Memory[Base - 1] then Error(4);
            Push(Base)
          end;
        chkptr :
          { checkpointer          }
          begin
            Address := Memory[StackTop];
            if MemoryUndefined(Address, PointerSize) then Error(43);
            Base := Memory[Address];
            if Base <> 0 then
              if Memory[Address + 1] <> Memory[Base - 1] then Error(43)
          end;
        eqptr :
          { testpequal            }
          begin
            StackTop := StackTop - 2 * PointerSize;
            Base1 := StackTop + 1;
            Base2 := Base1 + PointerSize;
            Result :=
              ord
                ((Memory[Base1] = Memory[Base2]) and
                 (Memory[Base1 + 1] = Memory[Base2 + 1]));
            Push(Result)
          end;
        neptr :
          { testpunequal          }
          begin
            StackTop := StackTop - 2 * PointerSize;
            Base1 := StackTop + 1;
            Base2 := Base1 + PointerSize;
            Result :=
              ord
                ((Memory[Base1] <> Memory[Base2]) or
                 (Memory[Base1 + 1] <> Memory[Base2 + 1]));
            Push(Result)
          end;

{

 10.3.9  String Comparisons

 The following  instructions  perform  comparisons  between  string
 operands.   In each case the word argument specifies the number of
 string bytes to be compared.  The operands  which  are  loaded  at
 word  aligned addresses on the stack, are popped and compared, and
 a Boolean result pushed back on the stack.


              Instruction     Argument     Notes

              TestSEqual        word
              TestSUnequal      word
              TestSLess         word
              TestLtOrEqual     word
                                                                     }

        eqstr :
          { testsequal            }
          begin
            Size := NextWord;
            Result := ord(StringEqual(Size));
            FreeStack(WordsFor(Size) * 2);
            Push(Result)
          end;
        nestr :
          { testsunequal          }
          begin
            Size := NextWord;
            Result := ord(not StringEqual(Size));
            FreeStack(WordsFor(Size) * 2);
            Push(Result)
          end;
        ltstr :
          { testsless             }
          begin
            Size := NextWord;
            Result := ord(StringLess(Size));
            FreeStack(WordsFor(Size) * 2);
            Push(Result)
          end;
        lestr :
          { testsltorequal        }
          begin
            Size := NextWord;
            Result := ord(StringLess(Size) or StringEqual(Size));
            FreeStack(WordsFor(Size) * 2);
            Push(Result)
          end;

{

 10.3.10 Set Operations

 The following instructions perform operations on set operands.


              Instruction     Argument     Notes

              LoadSet           byte         1
              StoreSet          byte         2
              SingletonSet                   3
              RangeSet                       4
              SetUnion                       5
              SetIntersection                5
              SetDifference                  5
              SetExpand         byte         6
              TestSetEqual                   7
              TestSetUnEqual                 7
              TestSubset                     7
              Inset                          8
              CheckSetUpper     byte         9
              CheckSetLimits    byte        10


 (1)  The set whose word address is at TOS is pushed onto the stack
      in  reduced format.  The byte argument gives the word-size of
      the expanded set.

 (2)  The set value in reduced format at TOS-1  is  expanded  to  a
      word  size  specified  by the byte argument and stored in the
      word address at TOS.

 (3)  The scalar value at TOS is popped and  used  to  construct  a
      singleton  set.   The  set  is pushed on the stack in reduced
      format.

 (4)  The instruction constructs  a  range-set  containing  numbers
      defined  by  the  limiting values at TOS and TOS-1.  The con-
      structed set is pushed on the stack in reduced format.

 (5)  These instructions perform  arithmetic  on  set  operands  in
      reduced-format.   The  result  set  is pushed on the stack in
      reduced format.

 (6)  This instruction expands the reduced-set at TOS to the  word-
      size specified by the byte argument.

 (7)  These instructions perform comparison between the reduced set
      operands  on  the stack.  A Boolean result is pushed onto the
      stack.

 (8)  This instruction tests membership in  a  reduced-set  on  the
      stack.   The  set  and member values are popped and a Boolean
      result value is pushed.

 (9)  This instruction performs a set range-check by  checking  all
      member  values  in  the reduced-set at TOS-1 are less than or
      equal to the limit value at TOS.  The byte  argument  identi-
      fies  the  error  to  be flagged if the condition is not met.
      The reduced-set is left at TOS.

 (10) This instruction performs a set range-check by  checking  all
      member values in the reduced-set at TOS-2 lie within the lim-
      iting values popped from TOS and TOS-1.

                                                                     }

        lodset :
          { loadset               }
          begin
            Size := NextByte;
            Pop(Base);
            ReducedSize := 0;
            if StackTop + Safety(Size + 1) >= HeapTop
            then SystemError(11);
            for i := 1 to Size do
              begin
                Operand := Memory[Base + i - 1];
                if Operand <> 0 then ReducedSize := i;
                Memory[StackTop + i] := Operand
              end;
            StackTop := StackTop + ReducedSize;
            PushSetSize(ReducedSize)
          end;
        stoset :
          { storeset              }
          begin
            Size := NextByte;
            Pop(Base);
            ExpandSet(Size);
            StoreWords(Base, Size)
          end;
        sglset :
          { singletonset          }
          begin
            Pop(Datum);
            Size := (Datum div MCWordSetBits) + 1;
            if StackTop + Safety(Size) >= HeapTop
            then SystemError(11);
            Start := StackTop + Size - 1;
            while StackTop < Start do Push(0);
            Element.WValue := 0;
            MCSetBit(Element, Datum mod MCWordSetBits);
            Push(Element.WValue);
            PushSetSize(Size)
          end;
        rngset :
          { rangeset              }
          begin
            Pop(Upper);
            Pop(Lower);
            if Lower > Upper
            then Size := 0
            else
              begin
                Size := (Upper div MCWordSetBits) + 1;
                if StackTop + Safety(Size) >= HeapTop
                then SystemError(11);
                Start := StackTop + (Lower div MCWordSetBits);
                while StackTop < Start do Push(0);
                Element.WValue := 0;
                for i := Lower to Upper do
                  begin
                    if i mod MCWordSetBits = 0
                    then
                      begin
                        Push(Element.WValue);
                        Element.WValue := 0
                      end;
                    MCSetBit(Element, i mod MCWordSetBits)
                  end;
                Push(Element.WValue)
              end;
            PushSetSize(Size)
          end;
        setuni :
          { setunion }
          begin
            GetSetDetails;
            with SetDetails do
              begin
                for i := 0 to Subset - 1 do
                  begin
                    SetElements(Base2 + i, Stride);
                    Element.WSet := Element2.WSet + Element1.WSet;
                    Push(Element.WValue)
                  end;
                if Size1 > Size2
                then
                  begin
                    LoadWords(Base1 + Subset, Size1 - Size2);
                    Size := Size1
                  end
                else Size := Size2;
                StackTop := Base2 + Size - 1;
                PushSetSize(Size)
              end
          end;
        setint :
          { setintersection       }
          begin
            GetSetDetails;
            with SetDetails do
              begin
                ReducedSize := 0;
                for i := 0 to Subset - 1 do
                  begin
                    SetElements(Base2 + i, Stride);
                    Element.WSet := Element2.WSet * Element1.WSet;
                    if Element.WSet <> [] then ReducedSize := i + 1;
                    Push(Element.WValue)
                  end;
                StackTop := Base2 + ReducedSize - 1;
                PushSetSize(ReducedSize)
              end
          end;
        setdif :
          { setdifference         }
          begin
            GetSetDetails;
            with SetDetails do
              begin
                ReducedSize := 0;
                for i := 0 to Subset - 1 do
                  begin
                    SetElements(Base2 + i, Stride);
                    Element.WSet := Element2.WSet - Element1.WSet;
                    if Element.WSet <> [] then ReducedSize := i + 1;
                    Push(Element.WValue)
                  end;
                if Size2 > Size1 then ReducedSize := Size2;
                StackTop := Base2 + ReducedSize - 1;
                PushSetSize(ReducedSize)
              end
          end;
        setexp :
          { setexpand             }
          begin ExpandSet(NextByte) end;

        { set comparisons }

        eqset :
          { testsetequal          }
          begin
            GetSetDetails;
            with SetDetails do
              if Size1 <> Size2
              then Result := ord(false)
              else
                begin
                  Result := ord(true);
                  for i := 0 to Subset - 1 do
                    begin
                      SetElements(Base2 + i, Stride);
                      if Element2.WSet <> Element1.WSet
                      then Result := ord(false)
                    end
                end;
            Push(Result)
          end;
        neset :
          { testsetunequal        }
          begin
            GetSetDetails;
            with SetDetails do
              if Size1 <> Size2
              then Result := ord(true)
              else
                begin
                  Result := ord(Subset <> 0);
                  for i := 0 to Subset - 1 do
                    begin
                      SetElements(Base2 + i, Stride);
                      if Element2.WSet = Element1.WSet
                      then Result := ord(false)
                    end
                end;
            Push(Result)
          end;
        leset :
          { testsubset            }
          begin
            GetSetDetails;
            with SetDetails do
              if Size2 > Size1
              then Result := ord(false)
              else
                begin
                  Result := ord(true);
                  for i := 0 to Subset - 1 do
                    begin
                      SetElements(Base2 + i, Stride);
                      if Element2.WSet - Element1.WSet <> []
                      then Result := ord(false)
                    end
                end;
            Push(Result)
          end;
        inset :
          { inset                 }
          begin
            Pop(Size);
            if Size < 0
            then Size := -Size
            else
              if Size > 0
              then
                begin
                  Push(Size);
                  Size := 1
                end;
            Base := StackTop - Size;
            Operand := Memory[Base];
            Offset := Operand div MCWordSetBits;
            Bits := Operand mod MCWordSetBits;
            if (Operand < 0) or (Offset >= Size)
            then Result := ord(false)
            else
              begin
                Element.WValue := Memory[Base + Offset + 1];
                MCGetBit(Element, Bits, Bit);
                Result := Bit
              end;
            StackTop := Base - 1;
            Push(Result)
          end;
        chksu :
          { checksetupper        }
          begin
            Pop(Upper);
            TrapCode := NextByte;
            Size := Memory[StackTop];
            if Size <> 0
            then
              begin
                if Size < 0 then Size := -Size else Size := 1;
                CheckSize := (Upper div MCWordSetBits) + 1;
                if Size > CheckSize
                then Error(TrapCode)
                else
                  if Size = CheckSize
                  then
                    begin
                      if Size = 1
                      then Base := StackTop
                      else Base := StackTop - Size;
                      LastBit := Upper mod MCWordSetBits;
                      Element.WValue := Memory[Base + Size - 1];
                      if not MCEmptyField
                               (Element, LastBit + 1, MCMaxSetBit)
                      then Error(TrapCode)
                    end
              end
          end;
        chksl :
          { checksetlimits       }
          begin
            Pop(Upper);
            Pop(Lower);
            TrapCode := NextByte;
            Size := Memory[StackTop];
            if Size <> 0
            then
              begin
                if Size < 0 then Size := -Size else Size := 1;
                CheckSize := (Upper div MCWordSetBits) + 1;
                if Size > CheckSize
                then Error(TrapCode)
                else
                  begin
                    EmptyWords := Lower div MCWordSetBits;
                    FirstBit := Lower mod MCWordSetBits;
                    LastBit := Upper mod MCWordSetBits;
                    if Size = 1
                    then Base := StackTop
                    else Base := StackTop - Size;
                    for i := 1 to EmptyWords do
                      if Memory[Base + i - 1] <> 0
                      then Error(TrapCode);
                    Element.WValue := Memory[Base + EmptyWords];
                    if not MCEmptyField(Element, 0, FirstBit - 1)
                    then Error(TrapCode);
                    if Size = CheckSize
                    then
                      begin
                        Element.WValue := Memory[Base + Size - 1];
                        if not MCEmptyField
                                 (Element, LastBit + 1, MCMaxSetBit)
                        then Error(TrapCode)
                      end
                  end
              end
          end;

{

 10.4  Jump Instructions

 The instructions in  this  group  implement  jumps  by  specifying
 either  a  relative  byte  displacement,  or by indirection via an
 entry in the local jump-table.


            Instructions         Arguments       Notes

            JumpForward           byte             1
            JumpBack              byte             1
            JumpVia               byte             2
            FJumpForward          byte             3
            FJumpBack             byte             3
            FJumpVia              byte             3
            JumpOut               byte byte        4
            FJumpConditional      byte             5
            TJumpConditional      byte             5
            CaseJump              WordString       6


 (1)  The byte argument  specifies  the  destination  byte  address
      relative to the current value of the ProgramCounter.

 (2)  The byte argument is an index to a jump-table  location  con-
      taining the destination byte address.

 (3)  The Boolean value at TOS is popped, and the jump is taken  if
      false.  The byte argument specifies a relative byte-offset or
      an index to a jump-table location.

 (4)  This instruction implements non-local jumps.  The first  byte
      argument specifies the nesting level displacement between the
      current and destination stack  frames  and  the  second  byte
      argument  is  an  index  into  the  non-local jump-table.  To
      ensure all intervening active blocks are correctly finalised,
      the  return addresses are reset to trigger execution of their
      respective finalisation code sequences.

 (5)  These instructions are used in the short  circuit  evaluation
      of  Boolean  expressions.  TOS contains a Boolean value which
      is popped and compared against the condition implied  by  the
      instruction  name.   If these match, the value is pushed as a
      result and the jump taken.  The  byte  argument  specifies  a
      relative forward displacement.

 (6)  This instruction implements a table-driven  case  jump.   The
      word  string  comprises  a  word-aligned  jump-table which is
      indexed by the index value TOS.  The first two  locations  of
      the  table contain upper and lower limits for the index value
      and an  error  is  flagged  if  either  of  these  limits  is
      exceeded.  The remaining elements of the table contain desti-
      nation  byte  addresses  as  defined  by  the  case-statement
      labels.  An error is flagged if no such address is associated
      with the index value.

                                                                     }

        jmpfw :
          { jumpforward           }
          begin
            Offset := NextByte;
            ProgramCounter := ProgramCounter + Offset
          end;
        jmpbk :
          { jumpback              }
          begin
            Offset := NextByte;
            ProgramCounter := ProgramCounter - Offset
          end;
        jmpvia :
          { jumpvia               }
          begin
            CodeBase := Memory[CurrentFrame - CBOffset];
            ProgramCounter := Memory[CodeBase - NextByte]
          end;

        { false jumps }

        fjpfw :
          { fjumpforward          }
          begin
            Offset := NextByte;
            Pop(Result);
            if Result = ord(false)
            then ProgramCounter := ProgramCounter + Offset
          end;
        fjpbk :
          { fjumpback             }
          begin
            Offset := NextByte;
            Pop(Result);
            if Result = ord(false)
            then ProgramCounter := ProgramCounter - Offset
          end;
        fjpvia :
          { fjumpvia              }
          begin
            LabelEntry := NextByte;
            Pop(Result);
            if Result = ord(false)
            then
              begin
                CodeBase := Memory[CurrentFrame - CBOffset];
                ProgramCounter := Memory[CodeBase - LabelEntry]
              end
          end;
        jmpout :
          { jumpout               }
          begin
            FrameOffset := NextByte;
            LabelEntry := NextByte;
            ThisFrame := CurrentFrame;
            LabelledFrame := OuterFrame(FrameOffset);
            NextFrame := Memory[ThisFrame - DLOffset];
            while NextFrame <> LabelledFrame do
              begin
                Memory[ThisFrame - RAOffset] :=
                  Memory[Memory[NextFrame - CBOffset] - FinalsOffset];
                ThisFrame := NextFrame;
                NextFrame := Memory[ThisFrame - DLOffset]
              end;
            Memory[ThisFrame - RAOffset] :=
              Memory[Memory[NextFrame - CBOffset] - LabelEntry];
            ProgramCounter :=
              Memory[Memory[CurrentFrame - CBOffset] - FinalsOffset]
          end;

        { special jumps }

        fjpcnd :
          { fjumpconditional      }
          begin
            Offset := NextByte;
            Pop(Result);
            if Result = ord(false)
            then
              begin
                Push(Result);
                ProgramCounter := ProgramCounter + Offset
              end
          end;
        tjpcnd :
          { tjumpconditional      }
          begin
            Offset := NextByte;
            Pop(Result);
            if Result = ord(true)
            then
              begin
                Push(Result);
                ProgramCounter := ProgramCounter + Offset
              end
          end;
        casejp :
          { casejump              }
          begin
            Base := WordBoundary;
            Pop(IndexValue);
            Lower := Memory[Base];
            Upper := Memory[Base + 1];
            if (IndexValue < Lower) or (IndexValue > Upper)
            then Error(51);
            Address := (Base + 2) + (IndexValue - Lower);
            Offset := Memory[Address];
            if Offset = 0 then Error(51);
            ProgramCounter := Offset + Address * MCBytesPerWord
          end;

{

 10.5  Procedure Entry and Exit

 The  following  instructions  implement  procedure  and   function
 entry/exit.  The entry sequence consists of the MarkStack instruc-
 tion to create a hole for the  incipient  stack  frame  header,  a
 series of instructions to evaluate or reference the actual parame-
 ters, and a block call instruction to enter the procedure or func-
 tion.   The block exit sequence consists of a single EndProcedure,
 EndFunction or EndMultiFunction instruction.


               Instructions    Arguments        Notes

               MarkStack                          1
               CallGlobal        byte             2
               CallLocal         byte             2
               CallLevel         byte             2
               CallOuter         byte byte        2
               CallFormal                         4
               EndProcedure                       5
               EndFunction                        6
               EndMultiFunction


 (1)  This  instruction  creates  a   six-word   'hole'    on   the
      expression-stack by advancing the stack-pointer.  The hole is
      later used to hold header information  when  the  new  stack-
      frame is created.

 (2)  These instructions call a procedure or function whose nesting
      level  displacement  from the calling procedure is known.  In
      the case of CallOuter, the displacement is given by the first
      byte  argument.  The remaining byte argument is an index into
      the jump-table of the block containing  the  procedure.   The
      jump-table entry holds the code-base of the called procedure.
      The new stack frame is created using information embedded  in
      the jump-table header of the called procedure.  If checks are
      requested, that part of the  new  stack  frame  reserved  for
      local variables is preset to the 'undefined' value.

 (4)  This instruction is used to call a formal procedure or  func-
      tion.   In  this  case TOS contains the address of the formal
      parameter which is a double word field containing the address
      of  stack-frame  enclosing the actual procedure, and the jump
      table index for the actual procedure.

 (5)  This instruction exits from a procedure to the return address
      in the calling block.

 (6)  These instructions exit from a function by popping the  func-
      tion  result  off the stack, popping the current stack-frame,
      pushing the result back on the  stack,  and  exiting  to  the
      return  address  in  the  calling block.  EndMultiFunction is
      used for multi-word function results.

                                                                     }

        mark :
          { markstack             }
          begin ClaimStack(FrameSize) end;
        calglo, calloc, callev, calotr, calfml :
          { callglobal            }
          { calllocal             }
          { calllevel             }
          { callouter             }
          { callformal            }
          begin
            case OpCode of
              calglo : EnclosingFrame := GlobalFrame;
              calloc : EnclosingFrame := CurrentFrame;
              callev :
                EnclosingFrame := Memory[CurrentFrame - SLOffset];
              calotr :
                begin
                  FrameOffset := NextByte;
                  EnclosingFrame := OuterFrame(FrameOffset)
                end;
              calfml :
                begin
                  Pop(Address);
                  EnclosingFrame := Memory[Address];
                  LabelEntry := Memory[Address + 1]
                end
            end;
            if OpCode <> calfml then LabelEntry := NextByte;
            CodeBase :=
              Memory[Memory[EnclosingFrame - CBOffset] - LabelEntry];
            Parameters := Memory[CodeBase - ParamsOffset];
            StackSpace := Memory[CodeBase - LocalsOffset];
            LockSpace := Memory[CodeBase - LocksOffset];
            NewFrame := StackTop - Parameters + 1;
            ClaimStack(StackSpace);
            SetFrameHeader;
            if ChecksOn and (StackSpace > 0)
            then PresetMemory(NewFrame + Parameters, StackSpace);
            ClaimStack(LockSpace);
            CurrentFrame := NewFrame;
            ProgramCounter := CodeBase * MCBytesPerWord
          end;
        endp :
          { endprocedure         }
          begin
            ProgramCounter := Memory[CurrentFrame - RAOffset];
            StackTop := CurrentFrame - FrameSize - 1;
            CurrentFrame := Memory[CurrentFrame - DLOffset]
          end;
        endf :
          { endfunction }
          begin
            Pop(Result);
            ProgramCounter := Memory[CurrentFrame - RAOffset];
            StackTop := CurrentFrame - FrameSize - 1;
            CurrentFrame := Memory[CurrentFrame - DLOffset];
            Push(Result)
          end;
        endmf :
          { endmultifuunction }
          begin
            Size := NextByte;
            Base := StackTop - Size + 1;
            ProgramCounter := Memory[CurrentFrame - RAOffset];
            StackTop := CurrentFrame - FrameSize - 1;
            CurrentFrame := Memory[CurrentFrame - DLOffset];
            LoadWords(Base, Size)
          end;

{

 10.6  Dynamic Variables

 The instructions in this group support the implementation  of  the
 standard procedures new and dispose.

             Instructions      Arguments      Notes

             New1              byte|word        1
             Dispose1          byte|word        2
             CheckNew1                          2
             CheckDisp1                         3
             CheckDisp2        byte             3


 (1)  These instructions implement calls to new and  dispose.   TOS
      contains  the  address of a pointer variable and the argument
      specifies the number of words of storage required.

 (2)  This  instruction  performs  the  check  for  a  dereferenced
      pointer  that  is used as a factor.  A non-zero level-byte in
      the pointer header indicates that the  variable  was  created
      with the extended form of new, and must be flagged in error.

 (3)  These instructions perform the run-time  checks  on  dispose.
      Errors  are  reported if the dynamic variable is locked by an
      extended reference, or if the first form of dispose  is  used
      and  a  previous  extended  form of new specified a different
      number of case labels.  The checks  are  detected  using  the
      lock  bit  and level byte count embedded in the dynamic vari-
      able header.

                                                                     }

        newp1b :
          { new1                  }
          begin Newp(NextByte) end;
        newp1w :
          { new1                  }
          begin Newp(NextWord) end;
        dspp1b :
          { dispose1              }
          begin Disposep(NextByte) end;
        dspp1w :
          { dispose1              }
          begin Disposep(NextWord) end;
        chkn1 :
          { checknew1             }
          begin
            Pop(Address);
            IpBuffer.WValue := Memory[Address - HeapLockOffset];
            MCGetByte(IpBuffer, HeapLevelByte, Byte);
            if Byte <> 0 then Error(25)
          end;
        chkd1 :
          { checkdsp1             }
          begin
            Pop(Address);
            IpBuffer.WValue := Memory[Address - HeapLockOffset];
            MCGetBit(IpBuffer, HeapLockBit, Bit);
            if Bit <> 0 then Error(5);
            MCGetByte(IpBuffer, HeapLevelByte, Byte);
            if Byte <> 0 then Error(20)
          end;
        chkd2 :
          { checkdsp2             }
          begin
            Pop(Address);
            Datum := NextByte;
            IpBuffer.WValue := Memory[Address - HeapLockOffset];
            MCGetBit(IpBuffer, HeapLockBit, Bit);
            if Bit <> 0 then Error(5);
            MCGetByte(IpBuffer, HeapLevelByte, Byte);
            if (Byte <> 0) and (Byte <> Datum) then Error(21)
          end;

{

 10.7 Special Instructions

 The instructions in this group are largely concerned  with  imple-
 mentation of run-time checks.


           Instructions               Arguments          Notes

           NoOperation                                     1
           Preset                                          2
           PresetRepeated             byte|word            3
           LoadStack                  byte                 4
           DuplicateStack                                  5
           PopStack                                        6
           CheckWord                  byte                 7
           CheckRepeated              byte                 8
           CheckVntField                                   9
           TrapProgramError           byte                 10
           TrapIfTrue                 byte                 10
           TrapIfFalse                byte                 10
           SetLock                    byte                 11
           DiscardLocks               byte                 12
           PurgeLocks                                      13
           SampleAndPurgeLocks                             14
           HaltProgram                                     15


 (1)  The explicit NoOperation instruction is used  as  padding  to
      ensure  wordstring  arguments  and  jump tables are correctly
      aligned on word boundaries.

 (2)  This instruction is used to  preset  the  memory  word  whose
      address is at TOS to the 'undefined' value.

 (3)  This instruction is used to preset an array of  memory  words
      to  the undefined value.  The base address of the array is at
      TOS, and the size of the array is given by the byte  or  word
      argument.

 (4)  This instruction pushes an element on  the  expression  stack
      onto  the  top  of  the  expression stack.  The byte argument
      indexes the required stack element.

 (5)  This instruction duplicates the top of the stack element.

 (6)  This instruction pops the expression stack and  discards  the
      element at TOS.

 (7)  This instruction flags an error specified by the  byte  argu-
      ment  if  the memory word addressed by the contents of TOS is
      undefined.

 (8)  This instruction flags an error specified by the  byte  argu-
      ment  if an element of the memory array addressed by the con-
      tents of TOS-1 is undefined.  The size of the array is  given
      by the contents of TOS.

 (9)  This instruction checks accesses to  variant  record  fields.
      TOS  contains  the current selector value, TOS-1 contains the
      selector value required with the access, and  TOS-2  contains
      the  record base address. An error is flagged if the selector
      field is not defined, or the current selector  value  differs
      from the required selector value.

 (10) These three instructions are  concerned  with  mandatory  and
      conditional trapping of an error.  The error is identified by
      the byte argument.

 (11) This  instruction  sets  a  reference  lock  by  pushing  its
      current-value  on the lock-stack before setting it.  The byte
      argument specifies the position of the lock-bit.

 (12) This instruction pops and discards locks from the lock-stack.
      The  byte  argument  specifies  the  number  of entries to be
      retained in the lock-stack.

 (13) This instruction purges the lock-bits from the  control  word
      of  a  variant  record,  when a variant record is passed as a
      value parameter.  TOS contains the  address  of  the  control
      word.

 (14) This instruction is only used in the context of assigning one
      variant  record  to  another.   If  the  left hand record has
      references extending to  any  of  its  active  variants,  the
      assignment  is  permitted  only if the selector values remain
      unchanged.  Such a check would be extremely  inefficient  and
      the  P-machine  employs  a  simpler  but stronger check which
      prohibits assignment if any selector of the left  hand  vari-
      able  is  locked.   If  the assignment is permitted, the lock
      bits of the source record are purged  prior  to  copying  the
      control word.

 (15) This instruction provides normal program termination.

                                                                     }

        noop :
          { explicit no-operation }
          ;
        preset :
          { presetword            }
          begin
            Pop(Base);
            PresetMemory(Base, 1)
          end;
        presrb :
          { presetrepeated(b)     }
          begin
            Pop(Base);
            PresetMemory(Base, NextByte)
          end;
        presrw :
          { presetrepeated(w)    }
          begin
            Pop(Base);
            PresetMemory(Base, NextWord)
          end;
        ldstk :
          { loadstack           }
          begin Push(Memory[StackTop - NextByte]) end;
        dpstk :
          { duplicatestack      }
          begin Push(Memory[StackTop]) end;
        popstk :
          { popstack            }
          begin StackTop := StackTop - 1 end;
        chkwd :
          { checkword           }
          begin
            TrapCode := NextByte;
            Pop(Base);
            if MemoryUndefined(Base, 1) then Error(TrapCode)
          end;
        chkrp :
          { checkrepeated       }
          begin
            TrapCode := NextByte;
            Pop(Size);
            Pop(Base);
            if MemoryUndefined(Base, Size) then Error(TrapCode)
          end;
        chkvf :
          { checkvntfield       }
          begin
            Pop(SelectorValue);
            Pop(SelectedValue);
            LockWord.WValue := Memory[Memory[StackTop]];
            Datum := NextByte;
            MCGetBit(LockWord, 3 * Datum + 2, Bit);
            if (Bit = 0) or
               (SelectorValue <> SelectedValue)
            then Error(2)
          end;
        trapp :
          { trapprogramerror    }
          begin Error(NextByte) end;
        trapt :
          { trapiftrue          }
          begin
            TrapCode := NextByte;
            Pop(Datum);
            if Datum = ord(true) then Error(TrapCode)
          end;
        trapf :
          { trapiffalse         }
          begin
            TrapCode := NextByte;
            Pop(Datum);
            if Datum = ord(false) then Error(TrapCode)
          end;
        setlk :
          { setlock             }
          begin
            Bits := NextByte;
            Pop(Base);
            OpBuffer.WValue := Memory[Base];
            MCGetBit(OpBuffer, Bits, Bit);
            PushLock(Bit, Base, Bits);
            MCSetBit(OpBuffer, Bits);
            Memory[Base] := OpBuffer.WValue
          end;
        dislk :
          { discardlocks        }
          begin
            Datum := NextByte;
            while Memory[CurrentFrame - LDOffset] > Datum do
              begin
                PopLock(Bit, Base, Bits);
                OpBuffer.WValue := Memory[Base];
                if Bit = 0
                then MCClearBit(OpBuffer, Bits)
                else MCSetBit(OpBuffer, Bits);
                Memory[Base] := OpBuffer.WValue
              end
          end;
        pglk :
          { purgelocks          }
          begin
            Pop(Base);
            LockWord.WValue := Memory[Base];
            LockWord.WSet := LockWord.WSet - LockBits.WSet;
            Push(LockWord.WValue);
            Push(Base + 1)
          end;
        smpglk :
          { sampleandpurgelocks }
          begin
            Pop(Base1);
            LockWord.WValue := Memory[Base1];
            LockWord.WSet := LockWord.WSet - ActiveBits.WSet;
            if LockWord.WSet <> [] then Error(2);
            Pop(Base2);
            LockWord.WValue := Memory[Base2];
            LockWord.WSet := LockWord.WSet - LockBits.WSet;
            Memory[Base1] := LockWord.WValue;
            Push(Base2 + 1);
            Push(Base1 + 1)
          end;
        haltp :
          { haltprogram          }
          begin
            Finished := true;
            Termination := Normal
          end;

{

 10.8  Class 2 Instructions

 The class 2 instructions are distinguished  by  a  preceding  255
 'escape'  byte-value.   They  comprise  instructions  for I/O, the
 transfer functions pack and unpack, the arithmetic functions,  and
 program parameter exchange.

                                                                     }

        escape :
          { escape to class 2                    }
          begin
            OpCode := NextByte;
            case OpCode of

{

 10.8.1

 The following instructions implement the standard I/O functions.


             Instructions        Arguments        Notes

             CheckReadMode                          1
             CheckWriteMode                         1
             CheckBuffer                           10
             RewriteFile                            2
             ResetFile                              2
             EndofLine                              3
             EndofFile                              3
             GetFile                                2
             ReadInteger                            4
             ReadReal                               4
             ReadLine                               5
             PutFile                                2
             WriteCharacter                         6
             WriteBoolean                           6
             WriteStrig                             6
             WriteInteger                           6
             WriteFloated                           7
             WriteFixed                             7
             PageFile                               8
             CloseFile                              9


 (1)  These instructions check the mode of the  file  and  flag  an
      error if the actual mode differs from the expected mode.

 (2)  These instructions perform the primitive  I/O  operations  of
      reset, rewrite, put and get.  For the first two, TOS contains
      the address of the file variable.  For the  second  two,  TOS
      contains an index-value identifying the file.

 (3)  These instructions implement the functions eof and eoln.

 (4)  These instructions handle real and integer  input  from  text
      files.

 (5)  This instruction implements the procedure readln for  a  text
      file.

 (6)  These instructions implement text output for integer, charac-
      ter, Boolean, and string data.

 (7)  These instructions implement fixed and  floating  point  text
      output for real data.

 (8)  This instruction implements the procedure page.

 (9)  This instruction  closes  a  file, flushes  the  buffer,  and
      releases the buffer storage and file descriptor.

 (10) This instruction performs an existence check on a file buffer
      and  flags  an  error  if  any  attempt is made to access the
      buffer-variable of a file prior  to  an  initial  rewrite  or
      reset operation.

                                                                     }

              chkrd :
                { checkreadmode         }
                begin
                  Pop(FileIndex);
                  with PascalFiles[FileIndex] do
                    if Mode <> Inspection then Error(14)
                end;
              chkwr :
                { checkwritemode        }
                begin
                  Pop(FileIndex);
                  with PascalFiles[FileIndex] do
                    if Mode <> Generation then Error(9)
                end;
              chkbuf :
                { checkbuffer           }
                begin
                  FileIndex := Memory[Memory[StackTop]];
                  if FileIndex = UndefinedIndex then Error(61)
                end;
              rewrtf :
                { rewritefile           }
                begin
                  Pop(Address);
                  RewriteFile
                end;
              resetf :
                { resetfile             }
                begin
                  Pop(Address);
                  ResetFile
                end;
              eolnf :
                { endofline             }
                begin
                  Pop(FileIndex);
                  with PascalFiles[FileIndex] do
                    Push(ord(BytePointer = LastByte))
                end;
              eoff :
                { endoffile             }
                begin
                  Pop(FileIndex);
                  with PascalFiles[FileIndex] do Push(ord(EndOfFile))
                end;
              getf, rdint, rdrl, rdln :
                { getfile              }
                { readinteger          }
                { readreal             }
                { readline             }
                begin
                  Pop(FileIndex);
                  case OpCode of
                    getf : GetFile;
                    rdint : ReadNumber(IntForm);
                    rdrl : ReadNumber(RealForm);
                    rdln : ReadLine
                  end
                end;
              putf, wrchar, wrbool, wrstr, wrint, wrflt, wrfix, wrln,
              pagef :
                { putfile              }
                { writecharacter       }
                { writeboolean         }
                { writestring          }
                { writeinteger         }
                { writefloated         }
                { writefixed           }
                { writeline            }
                { pagefile             }
                begin
                  Pop(FileIndex);
                  case OpCode of
                    putf : PutFile;
                    wrchar : WriteChar;
                    wrbool : WriteBoolean;
                    wrstr : WriteString;
                    wrint : WriteInteger;
                    wrflt : WriteReal(Floated);
                    wrfix : WriteReal(Fixed);
                    wrln : WriteLine;
                    pagef : PageFile
                  end
                end;
              closef :
                { closefile            }
                begin
                  Pop(Address);
                  CloseFile
                end;

{

 10.8.2  Transfer Functions

 The following instructions implement pack and unpack.


             Instructions      Arguments      Notes

             Pack                byte           1
             PackC               byte           1
             Unpack              byte           2
             UnpackC             byte           2


 (1)  These instructions implement the procedure  pack.   TOS  con-
      tains the size of the unpacked array, TOS-1 contains the base
      address of the packed array, and  TOS-2  contains  the  base-
      address  of  the unpacked array.  The byte argument specifies
      the number of bytes  per  word  in  the  packed  array.   The
      instruction  PackC  is used when runtime checks are requested
      to check that each unpacked array element is defined.

 (2)  These procedures implement the procedure  unpack.  The  stack
      contains  the  size  of  the unpacked array, the packed array
      base address, and  the  unpacked  array  base  address.   For
      UnpackC,   the   stack  additionally  holds  the  'undefined'
      CheckValue for elements of the packed array, and  every  such
      element  is  checked  against  this value.  The byte argument
      specifies the number of  elements  per  word  of  the  packed
      array.

                                                                     }

              pckpr :
                { pack                 }
                begin
                  Pop(Size);
                  Pop(Base1);
                  Pop(Base2);
                  Packp(Base2, Size, Base1, NextByte)
                end;
              pckprc :
                { packc                }
                begin
                  Pop(Size);
                  Pop(Base1);
                  Pop(Base2);
                  Packpc(Base2, Size, Base1, NextByte)
                end;
              Upkpr :
                { unpack               }
                begin
                  Pop(Size);
                  Pop(Base1);
                  Pop(Base2);
                  Unpackp(Base1, Base2, Size, NextByte)
                end;
              upkprc :
                { unpackc              }
                begin
                  Pop(Datum);
                  Pop(Size);
                  Pop(Base1);
                  Pop(Base2);
                  Unpackpc(Base1, Base2, Size, Datum, NextByte)
                end;

{

 10.8.3  Arithmetic Functions

 The following instructions implement the required arithmetic func-
 tions.   Each instruction pops the real argument off the stack and
 pushes a real result.


           Instructions        Arguments      Notes

           Sine
           Cosine
           NaturalExp
           NaturalLog                            1
           SquareRoot                            2
           ArcTangent


 (1)  If checks have been requested and the operand is <=0.0,  then
      an error is flagged.

 (2)  If checks have been requested and the operand is negative  an
      error is flagged.

                                                                     }

              sinf :
                { sine                  }
                begin
                  PopReal(RealOperand);
                  PushReal(sin(RealOperand))
                end;
              cosf :
                { cosine                }
                begin
                  PopReal(RealOperand);
                  PushReal(cos(RealOperand))
                end;
              expf :
                { naturalexp            }
                begin
                  PopReal(RealOperand);
                  PushReal(exp(RealOperand))
                end;
              logf :
                { naturallog            }
                begin
                  PopReal(RealOperand);
                  if RealOperand <= 0.0
                  then
                    begin
                      if ChecksOn
                      then Error(33) else RealResult := Undefined
                    end
                  else RealResult := ln(RealOperand);
                  PushReal(RealResult)
                end;
              sqrtf :
                { squareroot            }
                begin
                  PopReal(RealOperand);
                  if RealOperand < 0.0
                  then
                    begin
                      if ChecksOn
                      then Error(34) else RealResult := Undefined
                    end
                  else RealResult := sqrt(RealOperand);
                  PushReal(RealResult)
                end;
              arctnf :
                { arctangent             }
                begin
                  PopReal(RealOperand);
                  PushReal(arctan(RealOperand))
                end;

{

 1O.8.4  Program Parameter Exchange

 The instruction MapProgramParameter provides for  the  mapping  of
 program  parameters  to  external files.  The argument comprises a
 byte string containing the program parameter name and its position
 in  the program parameter list.  The position value is used to set
 the FileIndex and is used  as  an  identifying  value  thereafter.
 This  is  embedded  in  the  file  variable  memory  block,  and a
 corresponding descriptor is allocated for the file.  The variables
 InputId  and  OutputId  record the index values for the parameters
 Input and Output.
                                                                     }

              mapp :
                { mapprogramparameter   }
                begin
                  Size := NextByte;
                  FileIndex := NextByte;
                  Name := BlankName;
                  for i := 1 to Size - 1 do Name[i] := chr(NextByte);
                  Pop(Address);
                  Memory[Address] := FileIndex;
                  AllocatePascalFile;
                  if Name = 'INPUT       ' then InputId := FileIndex;
                  if Name = 'OUTPUT      ' then OutputId := FileIndex
                end
            end { class2 }
          end
      end { case }
    until Finished
  end { runprogram };

begin
  InitController;
  SetMCUndefined;
  InitMemory;
  InitSysFiles;
  InitPascIO;
  LoadProgram;
  RunProgram;
13:
  EndPascIO;
  EndController
end.










