{

                 A Post-mortem Dump Generator
                     for Standard Pascal





 This program generates post-mortem information following  abnormal
 program  termination  on  the  P-machine.  The program reports the
 nature of the error and supplements this with a trace-back of  all
 active program blocks together with a symbolic dump of their vari-
 ables. The interface between the post-mortem generator, the  model
 compiler,  and  the  P-machine  is  defined by the program heading
 which is as follows:-

                                                                     }

                           program

                       PostmortemDump

                  ( CorpseFile, ErrorFile,
            CodeMap, DataMap, NameFile, Output );

{

 The program parameters have the following nature and purpose:-

 CorpseFile  is an input file defining the state of  the  P-machine
             and the executing program when the error occurred, and
             is generated by the P-machine.

 ErrorFile   is an input textfile containing the texts of all error
             messages issued by the Model Implementation.

 CodeMap     is an input file holding the code map generated by the
             model compiler.

 DataMap     is an input file holding the data map generated by the
             model compiler.

 NameFile    is an input file holding the identifier  spellings  of
             all  named  objects  on the data map file, and is gen-
             erated by the model compiler.

 Output      is an output textfile on which the post-mortem dump is
             written.

 The principal steps involved in post-mortem analysis are  as  fol-
 lows.

 The following control information is read  from  the  first  eight
 words of the corpse file.

 (1)  A numeric code identifying the reason  for  program  termina-
      tion.

 (2)  The value of the P-machine program  counter  when  the  error
      occurred.

 (3)  The stack frame  base  for  the  block  in  which  the  error
      occurred.

 (4)  The stack frame base for the global program block.

 (5)  The top of the P-machine stack when the error occurred.

 (6)  The top of the P-machine heap when the error occurred.

 (7)  A Boolean value indicating whether the program  was  compiled
      with then generation of runtime checks selected.

 (8)  A Boolean value indicating  whether  a  post-mortem  dump  is
      required.

 The remainder of the corpse file contains a dump of the  P-machine
 memory  when  the error occurred, and is read into the array vari-
 able Corpse.

 Throughout post-mortem analysis, the code map, data map, and  name
 files  must  be  accessed in a non-sequential manner. Since Pascal
 possesses no indexed-sequential file access  method,  these  files
 are  simply  read  into  tables  implmented by the array variables
 Code, Data, and Names. The post-mortem generator will itself  ter-
 minate  with  a  fatal  error  if  the capacity of these tables is
 exceeded.

 Errors are classified as 'runtime', 'fatal' or 'system'  according
 to  whether  they  derive  from one of the insecurities in program
 execution defined in the  Pascal  Standard,  from  a  catastrophic
 failure  of  the  P-machine  itself,  or  from a system limitation
 inherent in the Model Implementation.  In each case, the classifi-
 cation  of the error is reported together with an explanatory mes-
 sage derived from the error file. Although the  Model  Implementa-
 tion is able to detect each of the distinct runtime errors defined
 in the Standard, it cannot always distinguish the precise  context
 in  which  they  occur.   Thus each runtime error message is post-
 fixed with one or more  'D'  numbers  corresponding  to  those  in
 Appendix D of the Standard.

 The location of the error is deduced by  using  the  code  map  to
 locate  the sequential program flow-point immediately prior to the
 point at which the error occurred. Word  4  of  the  corresponding
 block  jump-table header also holds a block-serial number which is
 used to locate the corresponding 'block-object' descriptor in  the
 data  map. This in turn allows the nature and name of the block to
 be reported. If the program was not compiled with  a  request  for
 post-mortem  information,  no attempt is made to locate the error,
 and the program terminates after printing the  error  message  and
 the value of the program counter.

 A full post-mortem analysis supplements the error  report  with  a
 trace-back  of  active  blocks  and  a symbolic variable dump. The
 trace-back gives the name and nature of each  block,  the  source-
 line  number  of  its  entry-point, and for procedure and function
 blocks, the source-line number of the  corresponding  call  state-
 ment.  For  each  such  block,  the  variable  dump consists of an
 alphabetically sorted list of local variables and parameter names,
 together with an analysis of the corresponding variable state.  In
 this version of the Model Implementation, all names are printed in
 upper case only.  Initially the variable state is determined to be
 'undefined', 'unknown', or 'defined', with the second value  aris-
 ing only when a program has been compiled without runtime checking
 code. Those variables classified as undefined are simply  reported
 as  such  using  the  notation <<undefined>> to denote this state.
 The rest are subjected to a further analysis  which  depends  upon
 the  variable's   type  information recorded in the data file, and
 which can be summarised as follows.

 (1)  Scalar variables are described in terms of their  values.  In
      each  case,  the  value is represented in a manner consistent
      with the variable's type-definition. Thus values belonging to
      an  enumerated-type  are  represented by the appropriate con-
      stant identifier, values of the standard type char, are given
      a graphic representation, and so on.

 (2)  Pointer variables are described in terms of an address  value
      when  the  pointer contents are non-zero, or as 'nil'. In the
      first case, the pointer is further qualified as 'bad' if  its
      contents  are  not  located on the heap, or 'dangling' if the
      dynamic variable has already been disposed.

 (3)  Set variables are described in terms of their values  printed
      as  set  constants.  Each  member is printed in a manner con-
      sistent with the base-type of the set.

 (4)  Array  variables,  and  conformant   array   parameters   are
      described in terms of their shape and their element types. No
      attempt is made to print the individual array elements,  how-
      ever  the  variable  state  is  further  analysed  as  either
      'totally defined' or 'partially undefined'. The element  type
      is described in source-language terms.

 (5)  Record variables are described in terms of  their  individual
      fields.  If  the  record has a variant part, then each nested
      variant is analysed in turn.  Variant selectors are qualified
      as  'locked'  if a reference has been extended to a component
      of the active variant, or  'preselected'  if  the  associated
      variant was specified in a call to the extended form of new.

 (6)  A file variable is described in terms of its mode, its  posi-
      tion,  and its buffer variable. In addition, the file-type as
      a whole is described in source-language terms.

 The source-language description  of  variable  types  is  complete
 except  for  the  inclusion of type-identifier names. For example,
 given the program declarations:

    type Score = 1..100;
         Colour = (Red,White,Blue);

    var  Result: array[Colour] of Score;

 the state of variable result will be reported as:

    RESULT = array[RED..BLUE] of 1..100
             and is  ...

 This version of the  post-mortem  generator  contains  no  special
 treatment of errors arising out of infinite or deep recursion, and
 will generate lengthy trace-backs  as  a  consequence.  Additional
 control   can  be achieved at no extra cost by simply limiting the
 number of times a recursive  block  is  dumped  as  the  stack  is
 unwound.


 Copyright  (C)     University of Manchester
                    Institute of Science and Technology
                    January 1984.

                                                                     }
label

   13 ;                                 { catastrophic pmd failure }




const



   {*************************************************************************}
   {*************************************************************************}
   {******                                                             ******}
   {******                                                             ******}
   {******                     Host Installation                       ******}
   {******                     -----------------                       ******}
   {******                                                             ******}
   {******                                                             ******}
   {****** 1. Investigate how your host Pascal processor implements    ******}
   {******    the variable:-                                           ******}
   {******                                                             ******}
   {******      HostWord : set of B0..Bn                               ******}
   {******                                                             ******}
   {******    If member B0 is mapped onto the most significant bit     ******}
   {******    position, then set the constant HostBitOrder to the      ******}
   {******    value LeftToRight.  Otherwise, set it to RightToLeft.    ******}
   {******                                                             ******}
   {****** 2. Investigate how your host Pascal processor implements    ******}
   {******    the packed variable:-                                    ******}
   {******                                                             ******}
   {******      HostWord : packed array [0..n] of char                 ******}
   {******                                                             ******}
   {******    If element HostWord[0] is mapped onto the most           ******}
   {******    significant byte position, then set the constant         ******}
   {******    HostByteOrder to the value LeftToRight. Otherwise, set   ******}
   {******    it to RightToLeft.                                       ******}
   {******                                                             ******}
   {****** 3. Set MaxHostChar to the largest host 'character value'.   ******}
   {******                                                             ******}
   {****** 4. Redefine the constants prefixed 'MC' for the host        ******}
   {******    machine.                                                 ******}
   {******                                                             ******}
   {*************************************************************************}
   {*************************************************************************}



   LeftToRight = 0 ;
   RightToLeft = 1 ;

   HostBitOrder  = RightToLeft ;        { for host Pascal compiler }
   HostByteOrder = LeftToRight ;
   MaxHostChar   = 255 ;                { for most character sets }



   { configuring and formatting constants }

   NilSerial      = 0 ;                 { unallocated serial number }
   MaxDataSerial  = 511 ;               { upper bound of data table }
   MaxNameIndex   = 8191 ;              { upper bound of names table }
   MaxCodeIndex   = 4095 ;              { upper bound of code table }

   Leftmost       = 2 ;                 { left-most print position }
   LeftMinus1     = 1 ;                 { leftmost - 1 }
   RightMost      = 80 ;                { right-most print position }
   RightPlus1     = 81 ;                { right-most + 1 }
   MaxField       = 78 ;                { rightmost-leftmost }
   Indent         = 3 ;                 { record-scope indentation }
   RealField      = 14 ;                { print field for reals }

   StartCode      = 0 ;                 { minimum error code }
   MaxCode        = 499 ;               { maximum error code }
   RuntimeBase    = 300 ;               { runtime errors:      301..400 }
   SystemBase     = 400 ;               { system limit errors: 401..420 }
   FatalBase      = 420 ;               { fatal errors:        421..450 }
   LinesPerText   = 2 ;                 { lines per error text }

   DispLimit      = 20 ;                { maximum address level }



   { constants describing the model P-machine }

   MCBitsPerWord  = 32 ;                { bits per memory-word }
   MCWordSetBits  = 31 ;                { bits per set-word }
   MCBitsPerByte  = 8 ;                 { bits per memory-byte }
   MCBytesPerWord = 4  ;                { mcbitsperword div mcbitsperbyte }
   MCMaxBitnum    = 31 ;                { mcbitsperword-1 }
   MCMaxSetBit    = 30 ;                { mcwordsetbits-1 }
   MCMaxByteNum   = 3 ;                 { mcbytesperword-1 }

   MCMaxint       = 2147483647 ;        { max mc-integer value }
   MCRealSize     = 2 ;                 { real size in words }
   MCMaxByte      = 255 ;               { max mc-byte value }
   MCMaxChar      = 255 ;               { max mc-char value }
   MCMaxLevel     = 32 ;                { max block nesting level }
   MCMaxSet       = 256 ;               { maximum members per set }

   MCMaxMemWord   = 16383 ;             { max memory-word address }
   MCMaxMemByte   = 65535 ;             { max memory-byte address }

   { P-machine stack-frame layout }

   SLOffset       = 1 ;                 { [framebase-1]=staticlink }
   CBOffset       = 2 ;                 { [framebase-2]=codebase   }
   DLOffset       = 3 ;                 { [framebase-3]=dynamiclink }
   RAOffset       = 4 ;                 { [framebase-4]=returnaddress }
   LDOffset       = 5 ;                 { [framebase-5]=lock-stack depth }
   LTOffset       = 6 ;                 { [framebase-6]=lock-stack top   }
   FrameSize      = 6 ;


   { P-machine jump-table layout }

   ParamsOffset   = 1 ;                 { [codebase-1] = parameter size }
   LocalsOffset   = 2 ;                 { [codebase-2] = stack size }
   LocksOffset    = 3 ;                 { [codebase-3] = lock-stack size }
   SerialOffset   = 4 ;                 { [codebase-4] = 0/pmd block serial }

   { file check-bits in file-control block }

   FileLockBit    = 0 ;                 { set when buffer is referenced }
   EofBit         = 1 ;                 { set when eof }
   EolBit         = 2 ;                 { set when eoln for a text-file }





type

   {-------------------------------------------------------------------------}
   {     My types                                                            }
   {-------------------------------------------------------------------------}
   TAMbit=( b0, b1, b2, b3, b4, b5, b6, b7, b8, b9,
            b10,b11,b12,b13,b14,b15,b16,b17,b18,b19,
            b20,b21,b22,b23,b24,b25,b26,b27,b28,b29,
            b30,b31 );

   TAMword= set of TAMbit;

   HostWord       = Integer ;
   Scalar         = 0..maxint ;

   DeclKind       = ( PreDefined,Declared ) ;

   TypeForm = (Scalars,SubRanges,Pointers,Arrays,CAPSchema,
               Records,Sets,Files,VariantPart,Variant) ;

   SetForms       = ( Unpacked,IsPacked,Constructed ) ;

   IdClass        = ( Domain,Types,Consts,Vars,Bound,Field,Proc,Func,Prog ) ;



   { types describing the model P-machine }

   MCBit          = 0..1 ;
   MCByte         = 0..MCMaxByte ;
   MCWord         = HostWord ;
   MCScalar       = 0..MCMaxint ;
   MCIntegerForm  = HostWord ;          { mc-integer-space: target=host }
   MCRealForm     = real ;              { mc-real-space:    target=host }
   MCCharSet      = 0..MCMaxChar ;
   MCBitRange     = 0..MCBitsPerWord ;
   MCBitIndex     = 0..MCMaxBitnum ;
   MCSetBits      = 0..MCMaxSetBit ;
   MCBitarray     = packed array[MCBitIndex] of MCBit ;
   MCByteIndex    = 0..MCMaxByteNum ;
   MCByteArray    = packed array[MCByteIndex] of MCByte ;
   MCCharArray    = packed array[MCByteIndex] of char ;
   MCWordSet      = set of MCSetBits ;

   TAMCastBit = record
                   case boolean of
                      false : (Index: MCBitIndex);
                      true  : (Bitt : TAMBit);
                end;

   WordCast       = ( AsSet,AsBits,AsBitSet,AsBytes,AsChars,AsReal,AsValue ) ;
   MCWordForm     = record
                       case WordCast of
                          AsSet   : ( WSet : MCWordSet ) ;
                          AsBits  : ( WBits : MCBitarray ) ;
                          AsBitSet: ( BitSet : TAMword );
                          AsBytes : ( WBytes : MCByteArray ) ;
                          AsChars : ( WChars : MCCharArray ) ;
                          AsReal  : ( WReal : MCRealForm ) ;
                          AsValue : ( WValue : MCWord )
                    end ;

   WordEntry      = ^ ListWord ;
   ListWord       = record
                       Word : MCWordForm ; Next : WordEntry
                    end ;

   ByteRange      = 0..MCMaxMemByte ;
   WordRange      = 0..MCMaxMemWord ;


   ValueKind      = ( Ordvalue,IntValue,BoolValue,CharValue,RealValue,
                      SetValue,StringValue,PntrValue ) ;

   ObjectValue    = record
                       WordSize : WordRange ;
                       case Kind : ValueKind of
                          Ordvalue,
                          IntValue,
                          BoolValue,
                          CharValue :
                             ( Ival : MCIntegerForm ) ;
                          RealValue :
                             ( Rval : MCRealForm ) ;
                          SetValue :
                             ( SetVal : WordEntry ) ;
                          StringValue :
                             ( Length : MCScalar ;
                               StringVal : WordEntry )   ;
                          PntrValue :
                             ( Pval : WordEntry )
                    end ;

   AddressLevel   = 0..DispLimit ;
   FieldLevel     = Scalar ;

   FieldOffset    = packed record
                       WordOffset : WordRange ;
                       Level : FieldLevel ;
                       case PartWord : Boolean of
                          false : ( WordSize : WordRange ) ;
                          true  : ( BitSize : MCBitRange ;
                                    BitOffset : MCBitIndex )
                    end ;

   BlockLabel     = packed record
                       BlockLevel : AddressLevel ; EntryOffset : MCByte
                    end ;

   RepKind        = ( ForScalar,ForSet,ForArray,ForARecord,ForVntRecord,
                      ForFile,ForPnter,ForVariant,ForSelector,ForCAP,ForReal,
                      ForString,ForOther );

   MachineValue   = packed record
                       Multiple : Scalar ;
                       case Defined : Boolean of
                          false : ( ) ;
                          true  : ( Magnitude : MCIntegerForm )
                    end ;

   TypeRepresentation = packed record
                           WordSize : WordRange ;
                           BitSize  : MCBitRange ;
                           Kind : RepKind ;
                           CheckValue : MachineValue ;
                           Min,Max : MCIntegerForm ;
                           PresetCode,PostSetCode,CheckCode : BlockLabel ;
                           Selector : FieldOffset
                        end ;

   StdTypes    = ( IntStd,RealStd,CharStd ) ;

   SerialRange = Scalar ;
   DataName    = Scalar ;
   MapIndex    = Scalar ;

   DataObject  = packed record
                   ObjSerial: SerialRange;
                   case ObjClass: IdClass of
                     Domain :
                        ( ) ;
                     Types:
                        (ObjRepresentation: TypeRepresentation;
                         case ObjForm: TypeForm of
                           Scalars:
                              (case ObjScalarKind: DeclKind of
                                 Declared:
                                    (ObjFirstConst: SerialRange) ;
                                 PreDefined:
                                    (StdType: StdTypes)) ;
                           SubRanges:
                              (ObjRangeType: SerialRange;
                               ObjMin, ObjMax: ObjectValue);
                           Sets:
                              (SetIsPacked: Boolean;
                               ObjBaseType: SerialRange);
                           Pointers:
                              ( ObjDomain : SerialRange ) ;
                           Arrays:
                              (ArrayIsPacked: Boolean;
                               ObjLowBound,ObjHighBound : ObjectValue ;
                               ObjAelType, ObjInxType: SerialRange);
                           CAPSchema:
                              (SchemaIsPacked: Boolean ;
                               ObjCompType,ObjInxSpec: SerialRange ;
                               ObjLowAddr,ObjHighAddr: WordRange) ;
                           Records:
                              (RecordIsPacked: Boolean;
                               ObjFixedPart, ObjVarPart: SerialRange);
                           Files:
                              (FileIsPacked, FileIsText: Boolean;
                               ObjFelType: SerialRange);
                           VariantPart:
                              (ObjTagField,ObjTagType,
                               ObjFstVariant: SerialRange;
                               ObjSelector : FieldOffset);
                           Variant:
                              (VariantIsDistinct: Boolean ;
                               ObjSubFixedPart, ObjSubVarPart,
                               ObjNextVariant: SerialRange;
                               ObjVariantValue: ObjectValue));
                     Consts :
                        (ConstName: DataName;
                         ConstValue: ObjectValue ;
                         NextConst: SerialRange);
                     Vars,
                     Bound:
                        (VarName: DataName;
                         VarType: SerialRange;
                         IsVarParam: Boolean;
                         LocalAddress: WordRange;
                         NextLocalVar: SerialRange);
                     Field:
                        (FieldName : DataName ;
                         FieldType: SerialRange;
                         ObjOffset: FieldOffset;
                         NextField: SerialRange);
                     Prog,Proc,Func:
                        { The serial no. of this data-table entry    }
                        { serves to identify the corresponding block }
                        { in the object program.                     }
                        (BlockName: DataName ;
                         BlockBody: MapIndex ;
                         FirstLocalVar: SerialRange)
                 end;

   TokenKinds = ( StartBody,FlowToken,EndBody ) ;
   MapToken  =  packed record
                   case Kind : TokenKinds of
                      StartBody :
                         ( BlockSerial : SerialRange ;
                           StartLine : Scalar ) ;
                      FlowToken :
                         ( CodeOffset,FlowLine : Scalar ) ;
                      EndBody :
                         ( JumpTableSize,CodeTableSize : Scalar )
                end ;

   CodeRange   = 0..MaxCodeIndex ;
   DataRange   = 0..MaxDataSerial ;
   NameRange   = 0..MaxNameIndex ;
   NameByte    = 0..MCMaxChar ;

   ErrorSpan      = StartCode..MaxCode ;
   ClassOfError   = ( RuntimeError,SystemError,FatalError,Other ) ;




var

   { The following global variable holds the representation of an  }
   { undefined P-machine memory word. This value cannot be repres- }
   { ented in Standard Pascal and is therefore constructed  on a   }
   { bit by bit basis by the procedure SetMCUndefined when the     }
   { Postmortem Dump Generator is initialised.                     }

   UndefinedValue : MCWord ;


   CorpseFile     : file of MCWord ;
   Corpse         : array[WordRange] of MCWord ;


   ErrorFile : text ;
   ErrorCode : ErrorSpan ;
   ErrorClass : ClassOfError ;


   ProgramCounter : ByteRange ;
   LocalFrame,GlobalFrame,StackTop,HeapTop : WordRange ;
   ChecksRequested,PMDRequested : Boolean ;


   CodeMap : packed file of MapToken ;
   DataMap : packed file of DataObject ;
   NameFile : packed file of NameByte ;


   { the code,data, and name tables }

   CodeIndex,CodeLimit : CodeRange ;
   Code : packed array[CodeRange] of MapToken ;
   Data : packed array[DataRange] of DataObject ;
   Names : packed array[NameRange] of NameByte ;




{ - - -              global procedures                   - - - }

procedure MCSetBit ( var Word : MCWordForm ; BitIndex : MCBitIndex ) ;

   var
      TheBit : TAMCastBit;

   begin
      if HostBitOrder = LeftToRight
      then
         TheBit.Index := MCMaxBitNum-BitIndex
      else
         TheBit.Index := BitIndex;
      Word.BitSet:=Word.BitSet + [TheBit.Bitt];
   end { mcsetbit } ;

procedure MCGetBit ( Word : MCWordForm ; BitIndex : MCBitIndex ;
                     var ResultBit : MCBit ) ;

   var
      TheBit : TAMCastBit;

   begin
      if HostBitOrder = LeftToRight
      then
         TheBit.Index := MCMaxBitNum-BitIndex
      else
         TheBit.Index := BitIndex;
      ResultBit := ord( TheBit.Bitt in Word.BitSet);
   end { mcgetbit } ;

procedure SetMCUndefined;

  var MachineWord: MCWordForm;

  begin
    MachineWord.WValue := 0;
    MCSetbit(MachineWord, MCMaxBitnum);
    UndefinedValue := MachineWord.WValue
  end { setmcundefined };

function MCUndefined(Word: MCWord): Boolean;

  begin MCUndefined := (Word = UndefinedValue) end;

function BlkSerialof(StackFrame: WordRange): SerialRange;

  var CodeBase: WordRange;

  begin
    CodeBase := Corpse[StackFrame - CBOffset];
    BlkSerialof := Corpse[CodeBase - SerialOffset]
  end { blkserialof };

procedure AbortPMD(AbortCode: Scalar);

  begin
    writeln;
    writeln(' ': LeftMinus1, '--- PMD error --- ');
    writeln;
    write(AbortCode: 4, ' :  ');
    case AbortCode of
      1 : writeln('Error file not found. Check call to PMD.');
      2 : writeln('Corpse file not found. Check call to PMD.');
      3 : writeln('Unexpected end of corpse file.');
      4 : writeln('System limit: corpse file too large.');
      5 : writeln('Code map file not found. Check call to PMD.');
      6 : writeln('Name file not found. Check call to PMD.');
      7 :
        writeln
          ('System limit: program too large. Expand name table.');
      8 : writeln('Data map file not found. Check call to PMD.');
      9 :
        writeln
          ('System limit: program too large. Expand data table.');
      10 :
        writeln('System limit: program too large. Expand code table.')
    end;
    goto 13
  end { abortpmd };

{ - - -              The corpse file handler               - - - }

function NextCorpseWord: MCWord;

  begin
    get(CorpseFile);
    if eof(CorpseFile) then AbortPMD(3);
    NextCorpseWord := CorpseFile^
  end { nextcorpseword };

procedure ReadCorpseFile;

  var Index: Scalar;

  begin
    reset(CorpseFile);
    if eof(CorpseFile) then AbortPMD(2);

    { read the 8-word control block }
    ErrorCode := CorpseFile^;
    ProgramCounter := NextCorpseWord;
    LocalFrame := NextCorpseWord;
    GlobalFrame := NextCorpseWord;
    StackTop := NextCorpseWord;
    HeapTop := NextCorpseWord;
    ChecksRequested := (NextCorpseWord = ord(true));
    PMDRequested := (NextCorpseWord = ord(true));

    { now read the remainder of the corpse }
    get(CorpseFile);
    Index := 0;
    while not eof(CorpseFile) do
      begin
        if Index > MCMaxMemWord
        then AbortPMD(4) else Corpse[Index] := CorpseFile^;
        Index := Index + 1;
        get(CorpseFile)
      end
  end { readcorpsefile };

{ - - -            The code-map handler               - - - }

procedure ReadCodemap;

  var Index: Scalar;

  begin
    if PMDRequested
    then
      begin
        reset(CodeMap);
        if eof(CodeMap) then AbortPMD(5);
        Index := 0;
        while not eof(CodeMap) do
          begin
            if Index > MaxCodeIndex
            then AbortPMD(10) else Code[Index] := CodeMap^;
            Index := Index + 1;
            get(CodeMap)
          end;
        CodeIndex := 0;
        CodeLimit := Index - 1
      end
  end { readcodemap };

procedure FindLine(FrameBase: WordRange; CodeAddress: ByteRange;
                   var SourceLine: Scalar);

  var CodeBase: WordRange;
      BlockSerial: SerialRange;
      CodeAdjustment: ByteRange;
      EndOfSearch: Boolean;
      PreviousToken: MapToken;

  begin
    CodeBase := Corpse[FrameBase - CBOffset];
    CodeAdjustment := CodeBase * MCBytesPerWord;
    BlockSerial := Corpse[CodeBase - SerialOffset];
    CodeIndex := Data[BlockSerial].BlockBody;
    EndOfSearch := false;
    repeat
      PreviousToken := Code[CodeIndex];
      CodeIndex := CodeIndex + 1;
      with Code[CodeIndex] do
        case Kind of
          FlowToken :
            EndOfSearch := CodeOffset + CodeAdjustment >= CodeAddress;
          EndBody : EndOfSearch := true
        end
    until EndOfSearch;
    with PreviousToken do
      if Kind = StartBody
      then SourceLine := StartLine else SourceLine := FlowLine
  end { findline };

{ - - -         The name-file handler            - - - }

procedure ReadNameFile;

  var Index: Scalar;

  begin
    if PMDRequested
    then
      begin
        reset(NameFile);
        if eof(NameFile) then AbortPMD(6);
        Index := 0;
        while not eof(NameFile) do
          begin
            if Index > MaxNameIndex
            then AbortPMD(7) else Names[Index] := NameFile^;
            Index := Index + 1;
            get(NameFile)
          end
      end
  end { readnamefile };

function NameLength(NameIndex: NameRange): Scalar;

  begin NameLength := Names[NameIndex] end;

procedure PrintName(NameIndex: NameRange; PrintField: Scalar);

  var i, Length: Scalar;

  begin
    Length := NameLength(NameIndex);
    for i := NameIndex + 1 to NameIndex + Length do
      write(chr(Names[i]));
    if PrintField > Length then write(' ': PrintField - Length)
  end { printname };

{ - - -            The data-map handler           - - - }

procedure ReadDataMap;

  var Index: SerialRange;

  begin
    if PMDRequested
    then
      begin
        reset(DataMap);
        if eof(DataMap) then AbortPMD(8);
        while not eof(DataMap) do
          begin
            Index := DataMap^.ObjSerial;
            if Index > MaxDataSerial
            then AbortPMD(9) else Data[Index] := DataMap^;
            get(DataMap)
          end
      end
  end { readdatamap };

{ - - -            Error diagnosis            - - - }

procedure SetErrorClass;

  begin
    if ErrorCode > FatalBase
    then ErrorClass := FatalError
    else
      if ErrorCode > SystemBase
      then ErrorClass := SystemError
      else
        if ErrorCode > RuntimeBase
        then ErrorClass := RuntimeError else ErrorClass := Other
  end { seterrorclass };

procedure PrintError;

  procedure PrintHeading;

    begin
      writeln;
      case ErrorClass of
        Other : ;
        RuntimeError : writeln(' ': 8, '--- Runtime Error ---');
        SystemError : writeln(' ': 8, '--- System Error ---');
        FatalError : writeln(' ': 8, '--- Fatal Error ---')
      end;
      writeln
    end { printheading };

  procedure FindText;

    var LineCount: Scalar;

    begin
      for LineCount := 1 to (ErrorCode - 1) * LinesPerText do
        readln(ErrorFile)
    end { findtext };

  procedure PrintText;

    var LineCount: 1..LinesPerText;

    begin
      for LineCount := 1 to LinesPerText do
        begin
          if not Eoln(ErrorFile)
          then
            begin
              repeat
                write(ErrorFile^);
                get(ErrorFile)
              until Eoln(ErrorFile);
              writeln
            end;
          readln(ErrorFile)
        end
    end { printtext };

  begin
    reset(ErrorFile);
    if eof(ErrorFile) then AbortPMD(1);
    PrintHeading;
    FindText;
    PrintText
  end { printerror };

procedure PrintLocation;

  var BlockSerial: SerialRange;
      ErrorLine: Scalar;

  begin
    writeln;
    write(' ': LeftMinus1);
    if PMDRequested
    then
      begin { return to declared block }
        BlockSerial := BlkSerialof(LocalFrame);
        while BlockSerial = NilSerial do
          begin
            ProgramCounter := Corpse[LocalFrame - RAOffset];
            StackTop := LocalFrame - FrameSize - 1;
            LocalFrame := Corpse[LocalFrame - DLOffset];
            BlockSerial := BlkSerialof(LocalFrame)
          end;

        { now locate the error line }
        FindLine(LocalFrame, ProgramCounter, ErrorLine);
        write('at or near line ', ErrorLine: 1, ' of ');
        with Data[BlockSerial] do
          begin
            case ObjClass of
              Prog : write('program ');
              Proc : write('procedure ');
              Func : write('function ')
            end;
            PrintName(BlockName, 1)
          end;
        writeln('.')
      end
    else
      writeln
        ('at address ', ProgramCounter - 1: 1, ' of the program.');
    writeln
  end { printlocation };

procedure PrintComments;

  begin
    write(' ': LeftMinus1);
    write('The program was compiled with ');
    if ChecksRequested then write('all ') else write('no ');
    writeln('run-time checks requested.');
    if PMDRequested
    then
      writeln
        (' ': LeftMinus1, 'Here is your post-mortem trace-back:-')
    else
      writeln
        (' ': LeftMinus1,
           'No post-mortem information was requested.');
    writeln
  end { printcomments };

{ - - -      variable dump and post-mortem trace      - - - }

procedure PrintVariables;

  var CallingFrame, FirstFrame: WordRange;
      StartLine, CallingLine: Scalar;
      BlockSerial: SerialRange;
      ReturnAddress: ByteRange;

  procedure PrintBlock;

    type FieldWidth = 0..MaxField;
         LinePosition = Leftmost..RightMost;
         DataField =
           packed
             record
               StartWord: WordRange;
               IsPacked: Boolean;
               StartBit, BitsField: MCBitRange
             end;

    var FirstField: DataField;

    procedure InitField(var Field: DataField);

      begin
        with Field do
          begin
            StartWord := 0;
            IsPacked := false;
            StartBit := 0;
            BitsField := MCBitsPerWord
          end
      end { initfield };

    procedure PrintScope(FirstItem: SerialRange;
                         ScopeBase: DataField;
                         StartAtPos: LinePosition);

      var Position, StartOfLine: LinePosition;
          NameField: FieldWidth;
          ThisItem: SerialRange;
          ItemField: DataField;
          Indirect: Boolean;

      procedure NewLine;

        begin
          writeln;
          Position := StartOfLine
        end { newline };

      procedure TabTo(NewPosition: LinePosition);

        begin
          write(' ': NewPosition - 1);
          Position := NewPosition
        end { tabto };

      procedure NextLine;

        begin
         writeln;
          TabTo(StartOfLine)
        end { nextline };

      procedure MakeSpaceFor(PrintField: FieldWidth);

        begin
          if Position + PrintField > RightMost
          then
            begin
              NewLine;
              if Position + PrintField > RightMost
              then TabTo(RightMost - PrintField)
              else TabTo(StartOfLine)
            end;
          Position := Position + PrintField
        end { makespacefor };

      procedure Printch(Ch: char);

        begin
          MakeSpaceFor(1);
          write(Ch)
        end { printch };

      procedure PrintString (String : packed array[l..u:Integer]
                                      of char);

        var Index, Length: FieldWidth;

        begin
          Length := u - l + 1;
          MakeSpaceFor(Length);
          for Index := 1 to Length do write(String[Index])
        end { printstring };

      function NextItem(Item: SerialRange): SerialRange;

        begin
          with Data[Item] do
            if ObjClass = Field
            then NextItem := NextField else NextItem := NextLocalVar
        end { nextitem };

      function ItemName(Item: SerialRange): DataName;

        begin
          with Data[Item] do
            if ObjClass = Field
            then ItemName := FieldName else ItemName := VarName
        end { itemname };

      procedure SetNameField;

        var Length: Scalar;

        begin
          NameField := 0;
          ThisItem := FirstItem;
          repeat
            Length := NameLength(ItemName(ThisItem));
            if Length > NameField then NameField := Length;
            ThisItem := NextItem(ThisItem)
          until ThisItem = NilSerial;
          if NameField > MaxField then NameField := MaxField
        end { setnamefield };

      procedure AddFields(Base: DataField; Offset: FieldOffset;
                          var Result: DataField);

        begin
          InitField(Result);
          with Result do
            begin
              StartWord := Base.StartWord + Offset.WordOffset;
              if Offset.PartWord
              then
                begin
                  IsPacked := true;
                  StartBit := Base.StartBit + Offset.BitOffset;
                  BitsField := Offset.BitSize
                end
            end
        end { addfields };

      procedure PrintValue(Field: DataField; TypeSerial: SerialRange);

        type DataStatus = (Undefined, Unknown, Defined);

        var Status: DataStatus;

        function DataValue(Field: DataField): MCWord;

          var Word: MCWord;
              OPBuffer, IPBuffer: MCWordForm;
              Index: MCBitIndex;
              BitValue: MCBit;

          begin
            with Field do
              begin
                Word := Corpse[StartWord];
                if IsPacked
                then
                  begin
                    IPBuffer.WValue := Word;
                    OPBuffer.WValue := 0;
                    for Index := 0 to BitsField - 1 do
                      begin
                        MCGetBit
                          (IPBuffer, StartBit + Index, BitValue);
                        if BitValue <> 0
                        then MCSetbit(OPBuffer, Index)
                      end;
                    Word := OPBuffer.WValue
                  end;
                DataValue := Word
              end
          end { datavalue };

        procedure FindStatus(Field: DataField;
                             TypeSerial: SerialRange;
                             var Status: DataStatus);

          var DefinedFields, UndefinedFields: Scalar;

          procedure InitStatus;

            begin
              DefinedFields := 0;
              UndefinedFields := 0;

            end { initstatus };

          procedure RecordStatus(LocalStatus: DataStatus);

            begin
              case LocalStatus of
                Defined : DefinedFields := DefinedFields + 1;
                Undefined : UndefinedFields := UndefinedFields + 1;
                Unknown :
              end
            end { recordstatus };

          procedure TotalStatus(FieldCount: Scalar;
                                var AsAWhole: DataStatus);

            begin
              AsAWhole := Unknown;
              if FieldCount = DefinedFields then AsAWhole := Defined;
              if FieldCount = UndefinedFields
              then AsAWhole := Undefined
            end { totalstatus };

          procedure LevelStatus(BaseField: DataField;
                                FixedPartSerial, VarPartSerial:
                                  SerialRange;
                                Level: Scalar;
                                var Status: DataStatus);

            var FieldCount: Scalar;
                FieldSerial, NextVariant: SerialRange;
                Field: DataField;
                FieldStatus, StateNow, VariantStatus: DataStatus;
                CheckWord: MCWordForm;
                ActiveBit: MCBit;
                Selector: MCWord;
                Found: Boolean;

            begin
              InitStatus;
              FieldSerial := FixedPartSerial;
              FieldCount := 0;
              while FieldSerial <> NilSerial do
                with Data[FieldSerial] do
                  begin
                    AddFields(BaseField, ObjOffset, Field);
                    FindStatus(Field, FieldType, FieldStatus);
                    RecordStatus(FieldStatus);
                    FieldCount := FieldCount + 1;
                    FieldSerial := NextField
                  end;
              TotalStatus(FieldCount, StateNow);
              if StateNow <> Unknown
              then
                if VarPartSerial <> NilSerial
                then
                  begin
                    CheckWord.WValue := Corpse[BaseField.StartWord];
                    MCGetBit(CheckWord, 3 * Level + 2, ActiveBit);
                    case StateNow of
                      Defined :
                        if ActiveBit = 0 then StateNow := Unknown;
                      Undefined :
                        if ActiveBit <> 0 then StateNow := Unknown
                    end;
                    if StateNow = Defined
                    then
                      with Data[VarPartSerial] do
                        begin
                          AddFields(BaseField, ObjSelector, Field);
                          Selector := DataValue(Field);
                          NextVariant := ObjFstVariant;
                          Found := false;
                          repeat
                            with Data[NextVariant] do
                              if Selector =
                                   ObjRepresentation.CheckValue.
                                     Magnitude
                              then Found := true
                              else NextVariant := ObjNextVariant
                          until Found or (NextVariant = NilSerial);
                          if Found
                          then
                            begin
                              with Data[NextVariant] do
                                LevelStatus
                                  (BaseField, ObjSubFixedPart,
                                     ObjSubVarPart, Level + 1,
                                       VariantStatus);
                              if VariantStatus <> Defined
                              then StateNow := Unknown
                            end
                        end
                  end;
              Status := StateNow
            end { levelstatus };

          procedure ArrayStatus(BaseField: DataField;
                                Elements: Scalar;
                                PackedArray: Boolean;
                                AeSerial: SerialRange;
                                var Status: DataStatus);

            var Index: Scalar;
                Element: DataField;
                ElsPerWord: 2..MCBitsPerWord;
                ElStatus: DataStatus;
                CanBePacked: Boolean;

            begin
              with Data[AeSerial] do
                begin
                  with ObjRepresentation do
                    if WordSize > 1
                    then CanBePacked := false
                    else
                      CanBePacked :=
                        PackedArray and
                        (BitSize <= MCBitsPerWord div 2);
                  Element := BaseField;
                  if CanBePacked
                  then
                    with Element do
                      begin
                        IsPacked := true;
                        ElsPerWord :=
                          MCBitsPerWord div ObjRepresentation.BitSize;
                        BitsField := MCBitsPerWord div ElsPerWord
                      end;
                  InitStatus;
                  for Index := 1 to Elements do
                    begin
                      FindStatus(Element, AeSerial, ElStatus);
                      RecordStatus(ElStatus);
                      with Element do
                        if IsPacked
                        then
                          if (Index mod ElsPerWord) = 0
                          then
                            begin
                              StartWord := StartWord + 1;
                              StartBit := 0
                            end
                          else StartBit := StartBit + BitsField
                        else
                          StartWord :=
                            StartWord + ObjRepresentation.WordSize
                    end;
                  TotalStatus(Elements, Status)
                end
            end { arraystatus };

          procedure CAPStatus(BaseField: DataField;
                              BPBlock: WordRange; PackedCAP: Boolean;
                              CompSerial: SerialRange;
                              var Status: DataStatus);

            var Index, Elements: Scalar;
                Component: DataField;
                ElsPerWord: 2..MCBitsPerWord;
                CompStatus: DataStatus;
                CanBePacked: Boolean;

            begin
              with Data[CompSerial] do
                begin
                  if PackedCAP
                  then
                    CanBePacked :=
                      PackedCAP and
                      (ObjRepresentation.BitSize <=
                         MCBitsPerWord div 2)
                  else CanBePacked := false;
                  Component := BaseField;
                  if CanBePacked
                  then
                    with Component do
                      begin
                        IsPacked := true;
                        ElsPerWord :=
                          MCBitsPerWord div ObjRepresentation.BitSize;
                        BitsField := MCBitsPerWord div ElsPerWord
                      end;
                  Elements :=
                    Corpse[BPBlock + 1] - Corpse[BPBlock] + 1;
                  InitStatus;
                  for Index := 1 to Elements do
                    begin
                      FindStatus(Component, CompSerial, CompStatus);
                      RecordStatus(CompStatus);
                      with Component do
                        if IsPacked
                        then
                          if (Index mod ElsPerWord) = 0
                          then
                            begin
                              StartWord := StartWord + 1;
                              StartBit := 0
                            end
                          else StartBit := StartBit + BitsField
                        else
                          StartWord := StartWord + Corpse[BPBlock + 2]
                    end;
                  TotalStatus(Elements, Status)
                end
            end { capstatus };

          begin { findstatus }
            Status := Unknown;
            if ChecksRequested
            then
              with Data[TypeSerial] do
                case ObjForm of
                  Scalars, SubRanges :
                    begin
                      Status := Defined;
                      if Field.IsPacked
                      then
                        begin
                          if DataValue(Field) =
                               ObjRepresentation.CheckValue.Magnitude
                          then Status := Undefined
                        end
                      else
                        if MCUndefined(Corpse[Field.StartWord])
                        then Status := Undefined
                    end;
                  Sets, Pointers :
                    if MCUndefined(Corpse[Field.StartWord])
                    then Status := Undefined else Status := Defined;
                  Records :
                    LevelStatus
                      (Field, ObjFixedPart, ObjVarPart, 0, Status);
                  Files :
                    if Corpse[Field.StartWord] = 0
                    then Status := Undefined else Status := Defined;
                  Arrays :
                    ArrayStatus
                      (Field,
                         ObjHighBound.Ival - ObjLowBound.Ival + 1,
                           ArrayIsPacked, ObjAelType, Status);
                  CAPSchema :
                    CAPStatus
                      (Field, ObjLowAddr + LocalFrame,
                         SchemaIsPacked, ObjCompType, Status)
                end
          end { findstatus };

        function IntDigits(IntValue: MCIntegerForm): Scalar;

          var Digits: Scalar;
              SignNeeded: Boolean;

          begin
            SignNeeded := IntValue < 0;
            Digits := 0;
            repeat
              IntValue := IntValue div 10;
              Digits := Digits + 1
            until IntValue = 0;
            if SignNeeded
            then IntDigits := Digits + 1 else IntDigits := Digits
          end { intdigits };

        procedure PrintScalar(ScalarValue: MCWord;
                              TypeSerial: SerialRange);

          procedure PrintConst(Ordinal: MCScalar;
                               FirstConst: SerialRange);

            var ConstSerial: SerialRange;
                Printed: Boolean;
                PrintField: FieldWidth;

            begin
              ConstSerial := FirstConst;
              Printed := false;
              repeat
                with Data[ConstSerial] do
                  if Ordinal = ConstValue.Ival
                  then
                    begin
                      PrintField := NameLength(ConstName);
                      MakeSpaceFor(PrintField);
                      PrintName(ConstName, PrintField);
                      Printed := true
                    end
                  else ConstSerial := NextConst
              until Printed or (ConstSerial = NilSerial);
              if not Printed then PrintString('(out of range)')
            end { printconst };

          procedure PrintStd(WordValue: MCWord; WhichType: StdTypes);

            var i: 1..MCRealSize;
                RealMap:
                  record
                    case Boolean of
                      false: (RealValue: MCRealForm);
                      true: (Words: array [1..MCRealSize] of MCWord)
                  end;
                PrintField: FieldWidth;

            begin
              case WhichType of
                IntStd :
                  begin
                    PrintField := IntDigits(WordValue);
                    MakeSpaceFor(PrintField);
                    write(WordValue: PrintField)
                  end;
                RealStd :
                  begin
                    PrintField := RealField;
                    MakeSpaceFor(PrintField);
                    with RealMap do
                      begin
                        for i := 1 to MCRealSize do
                          Words[i] := Corpse[Field.StartWord + i - 1];
                        write(RealValue: PrintField)
                      end
                  end;
                CharStd :
                  begin
                    MakeSpaceFor(3);
                    write('''');
                    if (WordValue >= 0) and
                       (WordValue <= MaxHostChar)
                    then
                      write(chr(WordValue));
                    write('''')
                  end
              end
            end { printstd };

          begin { printscalar }
            with Data[TypeSerial] do
              case ObjForm of
                Scalars :
                  case ObjScalarKind of
                    Declared : PrintConst(ScalarValue, ObjFirstConst);
                    PreDefined : PrintStd(ScalarValue, StdType)
                  end;
                SubRanges : PrintScalar(ScalarValue, ObjRangeType)
              end
          end { printscalar };

        procedure PrintPointer(Pvalue: MCWord);

          var PrintField: FieldWidth;

          begin
            if Pvalue = 0
            then PrintString('nil')
            else
              begin
                PrintField := IntDigits(Pvalue);
                MakeSpaceFor(PrintField);
                write(Pvalue: PrintField);
                if (Pvalue < StackTop) or (Pvalue > MCMaxMemWord)
                then PrintString(' (bad pointer)')
                else
                  if ChecksRequested
                  then
                    if Corpse[Pvalue - 1] <>
                         Corpse[Field.StartWord + 1]
                    then PrintString(' (dangling)')
              end
          end { printpointer };

        procedure PrintSet(BaseAddress: WordRange;
                           BaseSerial: SerialRange);

          var First: Boolean;
              SetWord: MCWordForm;
              SetBit: MCBit;
              Index: MCWord;

          begin
            First := true;
            Printch('[');
            with Data[BaseSerial] do
              for Index :=
                    ObjRepresentation.Min to ObjRepresentation.Max do
                begin
                  SetWord.WValue :=
                    Corpse[BaseAddress + Index div MCWordSetBits];
                  MCGetBit(SetWord, Index mod MCWordSetBits, SetBit);
                  if SetBit <> 0
                  then
                    begin
                      if not First then Printch(',');
                      PrintScalar(Index, BaseSerial);
                      First := false
                    end
                end;
            Printch(']')
          end { printset };

        procedure PrintRecord;

          procedure PrintLevel(FixedPartSerial, VarPartSerial:
                                 SerialRange;
                               Level: Scalar);

            var Locked, Preselected, Active, Found: Boolean;
                CheckWord: MCWordForm;
                SelectorField: DataField;
                Selector: MCWord;
                NextVariant: SerialRange;

            procedure TestBit(BitIndex: MCBitIndex;
                              var Bitset: Boolean);

              var CheckBit: MCBit;

              begin
                MCGetBit(CheckWord, BitIndex, CheckBit);
                Bitset := CheckBit <> 0
              end { testbit };

            procedure PrintTag(TagSerial: SerialRange);

              var TagField: DataField;
                  PrintField: FieldWidth;

              begin
                with Data[TagSerial] do
                  begin
                    PrintField := NameLength(FieldName);
                    MakeSpaceFor(PrintField);
                    PrintName(FieldName, PrintField);
                    PrintString(' = ');
                    AddFields(Field, ObjOffset, TagField);
                    PrintScalar(DataValue(TagField), FieldType)
                  end
              end { printtag };

            begin { printlevel }
              StartOfLine := StartOfLine + Indent;
              if FixedPartSerial <> NilSerial
              then PrintScope(FixedPartSerial, Field, StartOfLine);
              if VarPartSerial <> NilSerial
              then
                with Data[VarPartSerial] do
                  begin
                    TabTo(StartOfLine);
                    PrintString('(  ');
                    if ChecksRequested
                    then
                      begin
                        CheckWord.WValue := Corpse[Field.StartWord];
                        TestBit(3 * Level, Locked);
                        TestBit(3 * Level + 1, Preselected);
                        TestBit(3 * Level + 2, Active)
                      end
                    else
                      begin
                        Locked := false;
                        Preselected := false;
                        Active := true
                      end;
                    if Active and (ObjTagField <> NilSerial)
                    then PrintTag(ObjTagField);
                    if Locked then PrintString(' (locked)');
                    if Preselected then PrintString(' (preselected)');
                    NewLine;
                    if Active
                    then
                      begin
                        AddFields(Field, ObjSelector, SelectorField);
                        Selector := DataValue(SelectorField);
                        NextVariant := ObjFstVariant;
                        Found := false;
                        repeat
                          with Data[NextVariant] do
                            if Selector =
                                 ObjRepresentation.CheckValue.
                                   Magnitude
                            then Found := true
                            else NextVariant := ObjNextVariant
                        until Found or (NextVariant = NilSerial);
                        if Found
                        then
                          with Data[NextVariant] do
                            PrintLevel
                              (ObjSubFixedPart, ObjSubVarPart,
                                 Level + 1)
                      end;
                    TabTo(StartOfLine);
                    Printch(')');
                    NewLine
                  end;
              StartOfLine := StartOfLine - Indent
            end { printlevel };

          begin { printrecord }
            NewLine;
            StartOfLine := StartAtPos + Indent;
            TabTo(StartOfLine);
            PrintString('record');
            NewLine;
            with Data[TypeSerial] do
              PrintLevel(ObjFixedPart, ObjVarPart, 0);
            TabTo(StartOfLine);
            PrintString('end')
          end { printrecord };

        procedure PrintType(TypeSerial: SerialRange);

          var Bound, LastConst: MCWord;
              LastSerial, NextSerial: SerialRange;

          function Structured(TypeSerial: SerialRange): Boolean;

            begin
              with Data[TypeSerial] do
                Structured :=
                  ObjForm in [Arrays, Records, CAPSchema, Files]
            end { structured };

          procedure PrintParts(FixedPart, VarPart: SerialRange);

            var TagType, FieldSerial: SerialRange;

            procedure PrintField(FieldSerial: SerialRange);

              var NameField: FieldWidth;
                  NextType: SerialRange;

              begin
                with Data[FieldSerial] do
                  begin
                    NameField := NameLength(FieldName);
                    MakeSpaceFor(NameField);
                    PrintName(FieldName, NameField);
                    if NextField = NilSerial
                    then NextType := NilSerial
                    else NextType := Data[NextField].FieldType;
                    if FieldType = NextType
                    then Printch(',')
                    else
                      begin
                        PrintString(' : ');
                        PrintType(FieldType)
                      end
                  end
              end { printfield };

            procedure PrintVariant(ThisVariant: SerialRange);

              begin
                if ThisVariant <> NilSerial
                then
                  with Data[ThisVariant] do
                    begin
                      PrintVariant(ObjNextVariant);
                      TabTo(StartOfLine);
                      PrintScalar(ObjVariantValue.Ival, TagType);
                      if VariantIsDistinct
                      then
                        begin
                          PrintString(' : ');
                          NextLine;
                          PrintString('   (');
                          NewLine;
                          PrintParts(ObjSubFixedPart, ObjSubVarPart);
                          TabTo(StartOfLine);
                          PrintString('  ) ;')
                        end
                      else Printch(',');
                      NewLine
                    end
              end { printvariant };

            begin { printparts }
              StartOfLine := StartOfLine + Indent;
              FieldSerial := FixedPart;
              while FieldSerial <> NilSerial do
                with Data[FieldSerial] do
                  begin
                    TabTo(StartOfLine);
                    PrintField(FieldSerial);
                    if (NextField <> NilSerial) or
                       (VarPart <> NilSerial)
                    then PrintString(' ;');
                    FieldSerial := NextField;
                    NewLine
                  end;
              if VarPart <> NilSerial
              then
                with Data[VarPart] do
                  begin
                    TabTo(StartOfLine);
                    PrintString('case ');
                    if ObjTagField <> NilSerial
                    then PrintField(ObjTagField)
                    else PrintType(ObjTagType);
                    TagType := ObjTagType;
                    PrintString(' of');
                    NewLine;
                    StartOfLine := StartOfLine + Indent;
                    PrintVariant(ObjFstVariant);
                    StartOfLine := StartOfLine - Indent
                  end;
              StartOfLine := StartOfLine - Indent
            end { printparts };

          begin
            with Data[TypeSerial] do
              case ObjForm of
                Scalars :
                  case ObjScalarKind of
                    Declared :
                      begin
                        NextSerial := ObjFirstConst;
                        repeat
                          LastSerial := NextSerial;
                          NextSerial := Data[NextSerial].NextConst
                        until NextSerial = NilSerial;
                        LastConst := Data[LastSerial].ConstValue.Ival;
                        PrintScalar(0, TypeSerial);
                        PrintString(',...,');
                        PrintScalar(LastConst, TypeSerial)
                      end;
                    PreDefined :
                      case StdType of
                        IntStd : PrintString('integer');
                        RealStd : PrintString('real');
                        CharStd : PrintString('char')
                      end
                  end;
                SubRanges :
                  begin
                    PrintScalar(ObjMin.Ival, ObjRangeType);
                    PrintString('..');
                    PrintScalar(ObjMax.Ival, ObjRangeType)
                  end;
                Pointers :
                  begin
                    PrintString('pointer-type')
                  end;
                Sets :
                  begin
                    if SetIsPacked then PrintString('packed ');
                    PrintString('set of ');
                    PrintType(ObjBaseType)
                  end;
                Arrays :
                  begin
                    if ArrayIsPacked then PrintString('packed ');
                    PrintString('array [');
                    PrintType(ObjInxType);
                    PrintString('] of ');
                    if Structured(ObjAelType) then NextLine;
                    PrintType(ObjAelType)
                  end;
                CAPSchema :
                  begin
                    if SchemaIsPacked then PrintString('packed ');
                    PrintString('conformant array [');
                    Bound := Corpse[ScopeBase.StartWord + ObjLowAddr];
                    PrintScalar(Bound, ObjInxSpec);
                    PrintString('..');
                    Bound :=
                      Corpse[ScopeBase.StartWord + ObjHighAddr];
                    PrintScalar(Bound, ObjInxSpec);
                    PrintString('] of ');
                    if Structured(ObjCompType) then NextLine;
                    PrintType(ObjCompType)
                  end;
                Records :
                  begin
                    if RecordIsPacked then PrintString('packed ');
                    PrintString('record');
                    NewLine;
                    with Data[TypeSerial] do
                      PrintParts(ObjFixedPart, ObjVarPart);
                    TabTo(StartOfLine);
                    PrintString('end ')
                  end;
                Files :
                  if FileIsText
                  then PrintString('text-file')
                  else
                    begin
                      if FileIsPacked then PrintString('packed ');
                      PrintString('file of ');
                      if Structured(ObjFelType) then NextLine;
                      PrintType(ObjFelType)
                    end
              end
          end { printtype };

        procedure PrintArray;

          begin
            PrintType(TypeSerial);
            NextLine;
            PrintString('and is ');
            if Status = Defined
            then PrintString('totally defined')
            else PrintString('partially undefined')
          end { printarray };

        procedure PrintFile;

          var CheckWord: MCWordForm;
              FileBlock: WordRange;

          procedure PrintMode(FileMode: MCWord);

            begin
              case FileMode of
                1 :
                  {inspection}
                  PrintString('opened for reading by reset');
                2 :
                  {generation}
                  PrintString('opened for writing by rewrite');
                0 :
                  {notset}
                  PrintString('no previous reset or rewrite')
              end
            end { printmode };

          procedure PrintPosition(Word: MCWordForm);

            var BitValue: MCBit;

            begin
              MCGetBit(Word, FileLockBit, BitValue);
              if BitValue = 1 then PrintString('position is locked ');
              MCGetBit(Word, EofBit, BitValue);
              if BitValue = 1
              then PrintString('eof is true ')
              else PrintString('eof is false ');
              MCGetBit(Word, EolBit, BitValue);
              if BitValue = 1 then PrintString('eoln is true ');

            end { printposition };

          procedure PrintBuffer(FileName: NameRange;
                                FileType: SerialRange);

            var BufferField: DataField;
                PrintField: FieldWidth;

            begin
              with Data[FileType] do
                begin
                  InitField(BufferField);
                  with BufferField do
                    if FileIsText
                    then StartWord := FileBlock + 2
                    else StartWord := Corpse[FileBlock + 2];
                  PrintField := NameLength(FileName);
                  MakeSpaceFor(PrintField);
                  PrintName(FileName, PrintField);
                  PrintString('^ = ');
                  StartOfLine := StartOfLine + Indent;
                  PrintValue(BufferField, ObjFelType);
                  StartOfLine := StartOfLine - Indent
                end
            end { printbuffer };

          begin
            PrintType(TypeSerial);
            NextLine;
            FileBlock := Field.StartWord;
            PrintMode(Corpse[FileBlock + 3]);
            NextLine;
            CheckWord.WValue := Corpse[FileBlock + 1];
            PrintPosition(CheckWord);
            NextLine;
            PrintBuffer(ItemName(ThisItem), TypeSerial)
          end { printfile };

        begin { printvalue }
          FindStatus(Field, TypeSerial, Status);
          if Status <> Undefined
          then
            with Data[TypeSerial] do
              case ObjForm of
                Scalars : PrintScalar(DataValue(Field), TypeSerial);
                SubRanges :
                  PrintScalar(DataValue(Field), ObjRangeType);
                Pointers : PrintPointer(DataValue(Field));
                Sets : PrintSet(Field.StartWord, ObjBaseType);
                Arrays, CAPSchema : PrintArray;
                Records : PrintRecord;
                Files : PrintFile
              end
          else PrintString('<<undefined>>')
        end { printvalue };

      begin { printscope }
        SetNameField;
        ThisItem := FirstItem;
        repeat
          StartOfLine := StartAtPos;
          TabTo(StartOfLine);
          MakeSpaceFor(NameField);
          PrintName(ItemName(ThisItem), NameField);
          PrintString(' = ');
          with Data[ThisItem] do
            begin
              StartOfLine := StartOfLine + NameField + 3;
              case ObjClass of
                Vars, Bound :
                  begin
                    InitField(ItemField);
                    Indirect :=
                      IsVarParam or
                      (Data[VarType].ObjForm = CAPSchema);
                    with ItemField do
                      begin
                        StartWord :=
                          ScopeBase.StartWord + LocalAddress;
                        if Indirect
                        then StartWord := Corpse[StartWord]
                      end;
                    PrintValue(ItemField, VarType)
                  end;
                Field :
                  begin
                    AddFields(ScopeBase, ObjOffset, ItemField);
                    PrintValue(ItemField, FieldType)
                  end
              end
            end;
          NewLine;
          ThisItem := NextItem(ThisItem)
        until ThisItem = NilSerial
      end { printscope };

    begin { printblock }
      write(' ': LeftMinus1);
      with Data[BlockSerial] do
        begin
          case ObjClass of
            Prog : write('program ');
            Proc : write('procedure ');
            Func : write('function ')
          end;
          PrintName(BlockName, 1);
          if LocalFrame <> FirstFrame then write(' which');
          writeln(' was entered at line ', StartLine: 1);
          if FirstLocalVar <> NilSerial
          then
            begin
              InitField(FirstField);
              FirstField.StartWord := LocalFrame;
              writeln;
              PrintScope
                (FirstLocalVar, FirstField, Leftmost + Indent);
              writeln
            end;
          if CallingLine <> 0
          then
            writeln
              (' ': LeftMinus1, 'and was called from line ',
                 CallingLine: 1, ' of ...')
        end
    end { printblock };

  begin { printvariables }
    if PMDRequested
    then
      begin
        FirstFrame := LocalFrame;
        CallingFrame := LocalFrame;
        repeat
          LocalFrame := CallingFrame;
          BlockSerial := BlkSerialof(LocalFrame);
          CodeIndex := Data[BlockSerial].BlockBody;
          StartLine := Code[CodeIndex].StartLine;
          if LocalFrame <> GlobalFrame
          then
            begin
              CallingFrame := Corpse[LocalFrame - DLOffset];
              ReturnAddress := Corpse[LocalFrame - RAOffset];
              FindLine(CallingFrame, ReturnAddress - 1, CallingLine)
            end
          else CallingLine := 0;
          PrintBlock
        until LocalFrame = GlobalFrame
      end
  end { printvariables };

{ - - -                     main                       - - - }

begin
  SetMCUndefined ;
  ReadCorpseFile ;
  ReadCodemap ;
  ReadNameFile ;
  ReadDataMap ;
  SetErrorClass ;
  PrintError ;
  PrintLocation ;
  PrintComments ;
  PrintVariables ;
13:
end.
